{-# LANGUAGE DeriveDataTypeable,GeneralizedNewtypeDeriving #-}

module AST where

import Data.Typeable (Typeable)
import Data.Generics (Data)
import Data.IORef
import Control.Monad.Trans

--
-- Types
--

data TyPrim = TyInt | TyBool
    deriving (Eq,Show,Data,Typeable,Ord)

newtype TyvarName = TyvarName Int
    deriving (Eq,Show,Data,Typeable,Enum,Num,Ord)

type ClassName = String
type IfaceName = String

data TyG = TyN TyN | TyVar TyvarName
    deriving (Eq,Show,Data,Typeable,Ord)

data TyN = TyClass { tyc_name :: ClassName,
                     tyc_args :: [Type] }
    deriving (Eq,Show,Data,Typeable,Ord)

data TyK = TyIface { tyi_name :: IfaceName,
                     tyi_args :: [Type] }
    deriving (Eq,Show,Data,Typeable,Ord)

data TyNK = TyClassIface { tyci_name :: String,
                           tyci_args :: [Type] }
    deriving (Eq,Show,Data,Typeable,Ord)

data Type = TyG TyG | TyPrim TyPrim | TyK TyK | TyBot
          | TyParse TyNK -- only used during parsing
    deriving (Eq,Show,Data,Typeable,Ord)

--
-- Constraints
--

data ConR = Impl [TyG] TyK 
    deriving (Eq,Show,Data,Typeable,Ord)

data ConGR = ImplG [Type] TyK
    deriving (Eq,Show,Data,Typeable,Ord)

data ConP = ConR ConR | Ext TyvarName Type | Mono TyvarName
    deriving (Eq,Show,Data,Typeable)

data ConGP = ConGR ConGR | ExtG Type Type | MonoG Type
    deriving (Eq,Show,Data,Typeable)

data ConNil = ImplNil [Maybe Type] IfaceName [Maybe Type]
    deriving (Eq,Show)

--
-- Expressions and statements
--

newtype VarName = VarName Int
    deriving (Eq,Show,Data,Typeable,Enum,Num,Ord)

type FieldName = String
type MethodName = String

newtype Eq a => Ref a = Ref (IORef (Maybe a)) deriving (Eq,Data,Typeable)

instance Show (Ref a) where
    show _ = "<Ref>"

type TypeRef = Ref Type
type Coercion = Maybe TyK

data Expr = 
      Var VarName
    | FAccess Expr FieldName TypeRef
    | MCall Expr MethodName [Type] [Expr] 
            (Ref (-- return type
                  Maybe Type,
                  -- for a retroactive call: receiver type, argument types,
                  -- declaring interface
                  Maybe (Type, [Type], IfaceName),
                  -- argument coercions
                  [Coercion]))
    | SMCall TyK [Type] MethodName [Type] [Expr] (Ref (Type, [Coercion]))
    | New TyN [Expr] (Ref [Coercion])
    | Cast TyN Expr
    | PrimOp Op [Expr]
    | Lit Lit
    deriving (Eq,Show,Data,Typeable)

data Op = UnMinus | Add | Sub | Mult | Div | Mod | And | Or | Not
        | Leq | Lt | Geq | Gt | Equal | NotEqual
    deriving (Eq,Show,Data,Typeable)

data Lit = LitInt Int 
         | LitBool Bool 
         | LitString String 
         | LitNull
    deriving (Eq,Show,Data,Typeable)

data Stmt = 
      StmtExpr Expr
    | DefineLocal Type VarName Expr (Ref Coercion)
    | AssignLocal VarName Expr (Ref Coercion)
    | AssignField Expr FieldName Expr (Ref Coercion)
    | If Expr [Stmt]
    | IfElse Expr [Stmt] [Stmt]
    | While Expr [Stmt]
    | Return (Maybe Expr) (Ref Coercion)
    | Skip
    deriving (Eq,Show,Data,Typeable)

--
-- Class, interfaces, implementations, and programs
--

data Class = Class { class_name :: ClassName,
                     class_tyargs :: [TyvarName],
                     class_super :: Maybe TyN,
                     class_cons :: [ConP],
                     class_constructor :: Maybe Constructor,
                     class_fields :: [FieldDef],
                     class_methods :: [MethodDef] }
    deriving (Eq,Show,Data,Typeable)

newtype MethodArg = MethodArg { unMethodArg :: (Type,VarName) }
    deriving (Eq,Show,Data,Typeable)

data Constructor = Constructor { con_this :: VarName,
                                 con_args :: [MethodArg],
                                 con_body :: [Stmt] }
    deriving (Eq,Show,Data,Typeable)

data FieldDef = FieldDef { fd_type :: Type,
                           fd_name :: FieldName }
    deriving (Eq,Show,Data,Typeable)

data MethodSigGeneric a = MethodSig { ms_tyargs :: [TyvarName],
                                      ms_return :: Maybe Type,
                                      ms_name :: MethodName,
                                      ms_args :: [MethodArg],
                                      ms_cons :: [a] }
    deriving (Eq,Show,Data,Typeable)

type MethodSig = MethodSigGeneric ConP
type MethodSigG = MethodSigGeneric ConGP

data MethodDefGeneric a = MethodDef { 
      md_sig :: MethodSigGeneric a,
      md_this :: VarName, -- ignored if static method
      md_body :: [Stmt] 
} deriving (Eq,Show,Data,Typeable)

type MethodDef = MethodDefGeneric ConP
type MethodDefG = MethodDefGeneric ConGP

data Interface = Interface { iface_name :: IfaceName,
                             iface_tyargs :: [TyvarName],
                             iface_impl_types :: [TyvarName],
                             iface_impl_cons :: [ConR],
                             iface_cons :: [ConP],
                             iface_static_sigs :: [MethodSig],
                             iface_receiver_sigs :: [ReceiverSig] }
    deriving (Eq,Show,Data,Typeable)

data ReceiverSig = ReceiverSig { rcsig_name :: TyvarName,
                                 rcsig_sigs :: [MethodSig] }
    deriving (Eq,Show,Data,Typeable)

type ImplId = Int

data Implementation = Implementation { impl_tyargs :: [TyvarName],
                                       impl_iface :: TyK,
                                       impl_impl_types :: [TyN],
                                       impl_cons :: [ConP],
                                       impl_static_defs :: [MethodDef],
                                       impl_receiver_defs :: [ReceiverDef],
                                       impl_id :: ImplId }
    deriving (Eq,Show,Data,Typeable)

data ReceiverDef = ReceiverDef { rcdef_defs :: [MethodDef] }
    deriving (Eq,Show,Data,Typeable)

data Def = ImplDef Implementation | IfaceDef Interface | ClassDef Class
    deriving (Eq,Show,Data,Typeable)

data Program = Program { unProgram :: [Def] }
    deriving (Eq,Show,Data,Typeable)


--
-- Constants, predicates, and accessors
--

isObject :: Type -> Bool
isObject (TyG (TyN (TyClass "Object" []))) = True
isObject _ = False

isBot :: Type -> Bool
isBot TyBot = True
isBot _ = False

isPrim :: Type -> Bool
isPrim (TyPrim _) = True
isPrim _ = False

argType :: MethodArg -> Type
argType = fst . unMethodArg

argName :: MethodArg -> VarName
argName = snd . unMethodArg

tyInt :: Type
tyInt = TyPrim TyInt

tyBool :: Type
tyBool = TyPrim TyBool

tyString :: Type
tyString = TyG (TyN (TyClass "String" []))

tyStringN :: TyN
tyStringN = TyClass "String" []

tyObject :: Type
tyObject = TyG tyObjectG

tyObjectN :: TyN
tyObjectN = TyClass "Object" []

tyObjectG :: TyG
tyObjectG = TyN tyObjectN

mdefName :: MethodDefGeneric a -> MethodName
mdefName = ms_name . md_sig

mdefTyargs :: MethodDefGeneric a -> [TyvarName]
mdefTyargs = ms_tyargs . md_sig

mdefArgs :: MethodDefGeneric a -> [MethodArg]
mdefArgs = ms_args . md_sig

mdefReturn :: MethodDefGeneric a -> Maybe Type
mdefReturn = ms_return . md_sig

newRef :: (Eq a, MonadIO m) => m (Ref a)
newRef = 
    do r <- liftIO $ newIORef Nothing
       return (Ref r)

fillRef :: (Show a, Eq a, MonadIO m) => Ref a -> a -> m ()
fillRef (Ref r) x = 
    liftIO $ do old <- readIORef r
                case old of
                  Nothing -> writeIORef r (Just x)
                  _ -> return ()

fillTypeRef :: MonadIO m => TypeRef -> Type -> m ()
fillTypeRef _ t@(TyParse _) = error ("cannot fill TypeRef with " ++ show t)
fillTypeRef _ TyBot = error ("cannot fill TypeRef with TyBot")
fillTypeRef r t = fillRef r t

readRef :: (Eq a, MonadIO m) => Ref a -> m a
readRef (Ref r) = 
    liftIO $ do mx <- readIORef r
                case mx of
                  Nothing -> 
                      error ("cannot read Ref, it still contains Nothing")
                  Just x -> return x