import System.Console.GetOpt
import System.Exit
import System.Environment
import Control.Monad
import System.FilePath
import System.Cmd
import Control.Monad.Trans

import AST
import Parser
import Logging
import Lexer
import Pretty
import Symtab
import GIMonad
import ProgramTyping
import Java

version :: String
version = "development version"

data Flag   
    = Version
    | Help 
    | LogLevel LogLevel
    | LogFile FilePath
    | ClassPath FilePath
    | OutFile FilePath
    | Dump Phase
  deriving Eq

data Phase 
    = Lex
    | Parse
    | RawParse
  deriving Eq

programName :: String
programName = "core-javagic"

-- ---------------------------------------------------------------------
-- Handling command line flags

--
-- associate concrete flags with their abstract representation
--
opts :: [OptDescr Flag]
opts = 
    [ Option ['v'] ["version"] (NoArg Version) "version string"
    , Option ['?'] ["help"] (NoArg Help) "this message"
    , Option []    ["debug"] (NoArg (LogLevel Debug)) "enable debug messages"
    , Option []    ["quiet"] (NoArg (LogLevel Warn)) 
                   "display only warnings and errors"
    , Option []    ["log-file"] (ReqArg LogFile "FILE") "log file"
    , Option []    ["dump-tokens"]   (NoArg (Dump Lex)) 
            "dump lexed tokens"
    , Option []    ["dump-ast"]  (NoArg (Dump Parse)) 
            "print abstract syntax tree"
    , Option []    ["dump-rawast"]   (NoArg (Dump RawParse)) 
            "dump raw data type of abstract syntax tree"
    , Option ['o'] ["out-file"] (ReqArg OutFile "FILE") 
                 "output file (default: derived from first input file)"
    , Option [] ["classpath"] (ReqArg ClassPath "PATH") 
                 "Location of JavaGI runtime system"
    ]


--
-- parse the args.
--
-- we also short circuit here, for the known cases of --version or --help
--
parseArgs :: [String] -> IO ([Flag], [FilePath])
parseArgs argv = 
    case (getOpt Permute opts argv) of
      (flags,fs,[]) -> 
          case () of 
            _| Version `elem` flags -> putStrLn version >> exitWith ExitSuccess
             | Help `elem` flags -> putStrLn usage   >> exitWith ExitSuccess
             | otherwise         -> return (flags, fs)
      (_,_,err) -> fatal $ concat err ++ usage

usage :: String
usage = usageInfo ("Usage: " ++ programName  ++ " [OPTION...] FILES") opts

logLevel :: [Flag] -> LogLevel
logLevel [] = Info
logLevel (LogLevel l:_) = l
logLevel (_:rest) = logLevel rest

logTarget :: [Flag] -> LogTarget
logTarget [] = Stderr
logTarget (LogFile f:_) = File f
logTarget (_:rest) = logTarget rest

outFile :: [Flag] -> [FilePath] -> FilePath
outFile [] [] = bug "empty list of input files"
outFile [] (x:_) = replaceExtension x ".java"
outFile (OutFile f : _) _ = f
outFile (_ : rest) l = outFile rest l

classpath :: [Flag] -> Maybe String
classpath [] = Nothing
classpath (ClassPath p : _) = Just p
classpath (_ : rest) = classpath rest

parse :: [Flag] -> FilePath -> IO Program
parse flags f = 
    do toks <- scanFile f
       when (Dump Lex `elem` flags) $
            info ("Tokens from " ++ f ++ ":\n" ++ show toks)
       prog <- parseProgram toks
       when (Dump Parse `elem` flags) $
            info ("AST from " ++ f ++ ":\n" ++ showPP prog)
       when (Dump RawParse `elem` flags) $
            info ("Raw AST from " ++ f ++ ":\n" ++ show prog)
       return prog

mergeProgs :: [Program] -> Program
mergeProgs l = Program (concat (map unProgram l))

runJavac :: [Flag] -> FilePath -> GI ()
runJavac flags javaFile = 
    do let cmdLine = ("javac" ++ (case classpath flags of
                                    Nothing -> ""
                                    Just p -> " -classpath " ++ show p)
                      ++ " " ++ javaFile)
       info $ "Compiling to bytecode using command ``" ++ cmdLine ++ "''"
       e <- liftIO $ system cmdLine
       case e of
         ExitSuccess -> info ("Successfully compiled " ++ javaFile)
         ExitFailure i -> failGI ("javac failed with exit code " ++ show i)

main :: IO ()
main =
    do args <- getArgs
       (flags,files) <- parseArgs args
       initLogging (logLevel flags) (logTarget flags)
       if null files
          then fatal usage
          else do enterPhase "lex+parse"
                  progs <- mapM (parse flags) files
                  let prog = mergeProgs progs
                      symtab = buildSymtab prog
                  leavePhase
                  res <- withSymtab symtab $
                    do enterPhase "typechecking"
                       assertOKProg prog
                       info "program type checks!"
                       leavePhase
                       enterPhase "translation"
                       let out = outFile flags files
                       translateToJava prog out
                       leavePhase
                       enterPhase "java-compilation"
                       runJavac flags out
                       leavePhase
                  case res of
                    Right _ -> return ()
                    Left s -> fatal s
