module Environment (

   TEnv, mkTEnv, extendTEnv, emptyTEnv, addTyvars,
   upperBounds, transBounds, supImplCons, monoVars, definedTyvars,
   sup, sups,

   VEnv, mkVEnv, emptyVEnv, extendVEnv, typeOfVar
) where

import Control.Monad
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Set (Set)
import qualified Data.Set as Set
import Text.PrettyPrint
import Data.Maybe
import Data.List

import AST
import GIMonad
import Types
import Pretty
import Graph
import Subtyping
import Utils
import Positions
import Logging
import Dispatch

--
-- Type environment
--

data TEnv = TEnv { tenv_implCons :: Set ConGR,
                   tenv_upperBounds :: Map TyvarName [Type],
                   tenv_monoVars :: Set TyvarName,
                   tenv_tyvars :: Set TyvarName,
                   tenv_orig :: ([ConP], [TyvarName]) -- for debugging
                 }

instance Pretty TEnv where
    pp tenv = text "TEnv" <> braces (pp (tenv_orig tenv))

emptyTEnv :: TEnv
emptyTEnv = TEnv Set.empty Map.empty Set.empty Set.empty ([],[])

mkTEnv :: [ConP] -> [TyvarName] -> GI TEnv
mkTEnv cons tyvars = 
    do tenv <- extendTEnv emptyTEnv cons
       return $ addTyvars tenv tyvars

addToTEnv :: TEnv -> ConP -> GI TEnv
addToTEnv tenv c =
    let orig = (fst (tenv_orig tenv) ++ [c], snd (tenv_orig tenv)) in
    case c of
      ConR (Impl gs k) ->
          do cons <- sup (ImplG (map TyG gs) k)
             return $ tenv { tenv_implCons = Set.union cons (tenv_implCons tenv),
                             tenv_orig = orig }
      Ext x t -> 
          return $ tenv { tenv_upperBounds =  Map.insertWith (++) x [t] 
                                                (tenv_upperBounds tenv),
                          tenv_orig = orig }
      Mono x -> 
          return $ tenv { tenv_monoVars = Set.insert x (tenv_monoVars tenv),
                          tenv_orig = orig }

extendTEnv :: TEnv -> [ConP] -> GI TEnv
extendTEnv old ps = 
    do tenv <- foldM addToTEnv old ps
       checkTEnv tenv
       return tenv

addTyvars :: TEnv -> [TyvarName] -> TEnv
addTyvars tenv tyvars =
    tenv { tenv_tyvars = Set.fromList tyvars `Set.union` tenv_tyvars tenv,
           tenv_orig = (fst (tenv_orig tenv), 
                        snd (tenv_orig tenv) ++ tyvars) }

sups :: [ConGR] -> GI (Set ConGR)
sups cons = 
    do l <- mapM sup cons
       return $ Set.unions l

sup :: ConGR -> GI (Set ConGR)
sup p@(ImplG us (TyIface i vs)) =
    do iface <- findIface i
       let ss = iface_impl_cons iface
           ys = iface_impl_types iface
           xs = iface_tyargs iface
           subst = mkSubst (ys ++ xs) (us ++ vs)
       l <- mapM sup (apply subst ss)
       return $ Set.insert p $ Set.unions l

upperBounds :: TyvarName -> TEnv -> [Type]
upperBounds x tenv = 
    Map.findWithDefault [] x (tenv_upperBounds tenv)

transBounds :: TyvarName -> TEnv -> [Type]
transBounds x tenv = loop (upperBounds x tenv) Set.empty
    where loop [] acc = Set.toList acc
          loop (t@(TyG (TyVar y)) : rest) acc = 
              loop (upperBounds y tenv ++ rest) (Set.insert t acc)
          loop (t:rest) acc = loop rest (Set.insert t acc)
    
supImplCons :: TEnv -> Set ConGR
supImplCons tenv = tenv_implCons tenv

monoVars :: TEnv -> Set TyvarName
monoVars = tenv_monoVars

definedTyvars :: TEnv -> Set TyvarName
definedTyvars = tenv_tyvars

checkTEnv :: TEnv -> GI ()
checkTEnv tenv =
    do -- R-TEnv-1 holds trivially
       let (ps,_) = tenv_orig tenv
       checkRTenv2 ps
       -- R-TEnv-3 is checked in module Restrictions
       checkRTenv4 ps
       checkRTenv5 ps
       checkRTenv6 ps
       checkRTenv7 ps
    where
      checkRTenv2 ps =
          let edges = mapMaybe (\p -> case p of
                                        Ext x (TyG (TyVar y)) -> Just (x,y)
                                        _ -> Nothing) ps
              nodes = concatMap (\ (n1,n2) -> [n1,n2]) edges
              graph = buildUnlabelledGraph nodes edges
          in case enumAllCycles graph of
               [] -> return ()
               _ -> failGI ("Type environment " ++ showPP tenv ++ 
                            " is not contractive (R-TEnv-2)")
      checkRTenv4 ps =
          do let bounds = foldl (\m p -> case p of
                                           Ext x t@(TyG _) -> 
                                               Map.insertWith (++) x [t] m
                                           _ -> m) Map.empty ps
                 pairs = [ (t1,t2) | x <- Map.keys bounds,
                                     t1 <- fromJust (Map.lookup x bounds),
                                     t2 <- fromJust (Map.lookup x bounds),
                                     t1 < t2 ]
             b <- allM (\ (t1,t2) -> do b1 <- sub tenv t1 t2
                                        b2 <- sub tenv t2 t1
                                        return (b1 || b2)) pairs
             unless b $ failGI ("Type environment " ++ showPP tenv ++ 
                                " violates restriction R-TEnv-4")
      checkRTenv5 ps =
          mapM_ (\x -> do supers <- allSuperTypes tenv (TyG (TyVar x))
                          -- supers contains no duplicates, so it cannot 
                          -- contain more than one type using the same 
                          -- interface name
                          let superNames = [ i | TyK (TyIface i _) <- supers ]
                          if any (\n -> count n superNames >= 2) superNames
                             then failGI ("Type environment " ++ showPP tenv 
                                          ++ " violates restriction R-TEnv-5")
                             else return ())
                [ x | Ext x _ <- ps]
      checkRTenv6 ps =
          do b <- allM (\ i -> posMinus i 1) 
                       [ i | Ext _ (TyK (TyIface i _)) <- ps ]
             unless b $ failGI ("Type environment " ++ showPP tenv 
                                ++ " violates restriction R-TEnv-6")
      checkRTenv7 ps =
          do supPs' <- sups [ ImplG (map TyG gs) k | ConR (Impl gs k) <- ps ]
             let supPs = map (\ (ImplG ts k) -> Impl (map asG ts) k)
                             (Set.toList supPs')
             interfaces <- allInterfaces
             let ifaceNames = map iface_name interfaces
                 supPsByIface = map (\i -> (i, filter 
                                                 (\ (Impl _ (TyIface j _))
                                                      -> i == j) 
                                                 supPs))
                                    ifaceNames
             mapM_ checkRTenv71 supPsByIface
             mapM_ checkRTenv72 supPsByIface
      asG (TyG g) = g
      asG t = bug ("Environment.checkTEnv.asG: " ++ showPPQ t)
      checkRTenv71 (i, rs) =
          do let pairs = [ (r1, r2) | r1 <- rs, r2 <- rs ]
             dispIndices <- dispList i
             nDispIndices <- ndispList i
             nMinusIndices <- notPositions i Minus
             b <- allM (\ (Impl gs k, Impl hs k') -> 
                            do let dispGs = [ TyG $ gs!!(k-1) 
                                              | k <- dispIndices ]
                                   dispHs = [ TyG $ hs!!(k-1) 
                                              | k <- dispIndices ]
                               mts <- glbs tenv dispGs dispHs
                               case mts of
                                 Nothing -> return True
                                 Just _ ->
                                     return $ [ gs!!(k-1) | k <- nDispIndices ] 
                                                ==
                                              [ hs!!(k-1) | k <- nDispIndices ]
                                              &&
                                              [ gs!!(k-1) | k <- nMinusIndices ] 
                                                ==
                                              [ hs!!(k-1) | k <- nMinusIndices ]
                                              && 
                                              k == k')
                       pairs
             unless b $ failGI ("Type environment " ++ showPP tenv 
                                ++ " violates restriction R-TEnv-7(1)")

      checkRTenv72 (i, rs) =
          do impls <- findAllImpls i
             let rsxsImpls = map (\impl -> (impl_tyargs impl,
                                            Impl (map TyN (impl_impl_types impl))
                                                 (impl_iface impl)))
                                 impls
                 pairs = [ (r1, rx) | r1 <- rs, rx <- rsxsImpls ]
             dispIndices <- dispList i
             nDispIndices <- ndispList i
             nMinusIndices <- notPositions i Minus
             b <- allM (\ (Impl gs k, (xs, Impl hs k')) -> 
                            do let dispGs = [ gs!!(k-1) | k <- dispIndices ]
                                   dispHs = [ hs!!(k-1) | k <- dispIndices ]
                               msubst <- unifyModGLB tenv xs (zip dispGs dispHs)
                               case msubst of
                                 Nothing -> return True
                                 Just subst ->
                                     return $ [ TyG (gs!!(k-1)) 
                                                | k <- nDispIndices ] ==
                                              [ apply subst (hs!!(k-1)) 
                                                | k <- nDispIndices ]
                                              &&
                                              [ TyG (gs!!(k-1)) 
                                                | k <- nMinusIndices ] ==
                                              [ apply subst (hs!!(k-1)) 
                                                | k <- nMinusIndices ]
                                              && k == apply subst k')
                       pairs
             unless b $ failGI ("Type environment " ++ showPP tenv 
                                ++ " violates restriction R-TEnv-7(2)")
                            
--
-- Value environment
--

newtype VEnv = VEnv (Map VarName Type)

emptyVEnv :: VEnv
emptyVEnv = VEnv Map.empty

mkVEnv :: [VarName] -> [Type] -> VEnv
mkVEnv xs ts = extendVEnv emptyVEnv xs ts

extendVEnv :: VEnv -> [VarName] -> [Type] -> VEnv
extendVEnv (VEnv venv) xs ts =
    VEnv $ foldl (\venv (x,t) -> Map.insert x t venv) venv (zip xs ts)

typeOfVar :: VarName -> VEnv -> Maybe Type
typeOfVar x (VEnv m) = Map.lookup x m
