{-# LANGUAGE GeneralizedNewtypeDeriving #-}

module Execution ( execProgram ) where

{-

NOTES:

* Strings are treated specially. The interpreter does not instantiated object
  for string literals but uses the alternative "StringVal" of the "Value"
  data type. This implies that strings cannot be instantiated explicitly
  and that comparison of strings is done by value (and not by reference
  as in Java).                                                   

* We use a custom print routine for instances of Integer
-}

import Control.Monad.State
import Control.Monad.Reader
import qualified Data.Map as Map
import Data.Map (Map)
import Data.List
import Text.PrettyPrint

import AST
import GIMonad
import Inheritance
import Logging
import Pretty
import MethodDefinitions
import Types
import Builtins
import Utils
import Pretty
import WellFormedness

--
-- General definitions
--

type Store = Map Location Object

data Object = Object { obj_type :: TyN, obj_fields :: FieldMap }

newtype FieldMap = FieldMap { unFieldMap :: Map FieldName Value }

data Value = NonNull Location | Null | StringVal String 
           | IntVal Int | BoolVal Bool
    deriving (Eq,Show)

newtype Location = Location Int
    deriving (Eq,Show,Ord,Num)

type Env = Map VarName Value

data EXState = EXState { ex_env :: Env,
                         ex_store :: Store,
                         ex_nextFresh :: Location }

emptyEXState :: EXState
emptyEXState = EXState { ex_env = Map.empty,
                         ex_store = Map.empty,
                         ex_nextFresh = 0 }

type EX = StateT EXState GI

liftEX :: GI a -> EX a
liftEX = lift

failEX :: String -> EX a
failEX = liftEX . failGI

getEnv :: EX Env
getEnv = gets ex_env

setEnv :: Env -> EX Env
setEnv env = 
    do old <- getEnv
       modifyEnv (\_ -> env)
       return old

modifyEnv :: (Env -> Env) -> EX ()
modifyEnv f = modify  (\state -> state { ex_env = f (ex_env state) })

getStore :: EX Store
getStore = gets ex_store

modifyStore :: (Store -> Store) -> EX ()
modifyStore f = modify  (\state -> state { ex_store = f (ex_store state) })

freshLocation :: EX Location
freshLocation =
    do l <- gets ex_nextFresh
       modify (\state -> state { ex_nextFresh = l+1 })
       return l

lookupVar :: VarName -> EX Value
lookupVar x = 
    do env <- getEnv
       case Map.lookup x env of
         Just v -> return v
         Nothing -> bug ("Undefined variable " ++ showPP x)

lookupObject :: Location -> EX Object
lookupObject l = 
    do store <- getStore
       case Map.lookup l store of
         Just obj -> return obj
         Nothing -> bug ("Location " ++ show l ++ " not in store")

lookupField :: Location -> FieldName -> EX Value
lookupField l f =
    do obj <- lookupObject l
       case Map.lookup f (unFieldMap (obj_fields obj)) of
         Just v -> return v
         Nothing -> bug ("Field " ++ show f ++ " undefined for object of type "
                         ++ showPPQ (obj_type obj))

newObject :: TyN -> EX Location
newObject n = 
    do allFields <- liftEX $ findAllFields n
       let fieldMap = foldl (\ m fd -> let v = initialValue (fd_type fd)
                                          in Map.insert (fd_name fd) v m) 
                            Map.empty allFields
       let obj = Object n (FieldMap fieldMap)
       l <- freshLocation
       modifyStore (Map.insert l obj)
       return l
    where 
      findAllFields n@(TyClass c _) = 
          do cls <- findClass c
             super <- upOneClass n
             case super of
               Nothing -> return (class_fields cls)
               Just n' ->
                   do superFields <- findAllFields n'
                      return (class_fields cls ++ superFields)
      initialValue t = 
          case t of
            TyG (TyN _) -> Null
            TyK _ -> Null
            TyPrim TyInt -> IntVal 0
            TyPrim TyBool -> BoolVal False
            _ -> bug ("Detected invalid type " ++ showPPQ t ++ 
                      " during execution")

updateField :: Location -> FieldName -> Value -> EX ()
updateField l f v =
    do obj <- lookupObject l
       let obj' = obj { obj_fields = 
                            FieldMap $
                            Map.insert f v (unFieldMap (obj_fields obj)) }
       modifyStore (Map.insert l obj')
       

withEnv :: [VarName] -> [Value] -> EX a -> EX a
withEnv xs vs ex =
    do let env = Map.fromList (zip xs vs)
       old <- setEnv env
       x <- ex -- don't need to care about exceptions, 
               -- they abort the program anyway
       setEnv old
       return x

typeOfVal :: Value -> EX (Maybe TyN)
typeOfVal v =
    case v of
      NonNull l ->
          do obj <- lookupObject l
             return (Just $ obj_type obj)
      StringVal _ -> return (Just tyStringN)
      _ -> return Nothing

--
-- Evaluating expressions
--

evalExprNotVoid :: Expr -> EX Value
evalExprNotVoid expr = 
    do mres <- evalExpr expr
       case mres of
         Just res -> return res
         Nothing -> bug ("Evaluating expression " ++ showPPQ expr ++ 
                         " did not yield as result " ++
                         "(as required by the context)")

evalExpr :: Expr -> EX (Maybe Value)
evalExpr expr = 
    case expr of
      Var x -> 
          do v <- lookupVar x
             return (Just v)
      FAccess e f _ ->
          do v <- evalExprNotVoid e
             case v of
               Null -> failEX ("NullPointerException while evaluating " ++ 
                               showPPQ expr)
               NonNull l -> do v' <- lookupField l f
                               return (Just v')
               _ -> bug ("Cannot access field on value " ++ show v)
      MCall e m ts es _ ->
          do v <- evalExprNotVoid e
             vs <- mapM evalExprNotVoid es
             n <- typeOfVal v
             ns <- mapM typeOfVal vs
             mdef' <- liftEX $ getMDef m n ns
             case mdef' of
               Builtin id -> callBuiltinMethod id v vs
               NonBuiltin mdef ->
                   let xs = map argName (mdefArgs mdef)
                       this = md_this mdef
                       body = apply (mkSubst (mdefTyargs mdef) ts) 
                                    (md_body mdef)
                   in withEnv (this:xs) (v:vs) (execStmtList body)
      SMCall k ts m us es _ ->
          do vs <- mapM evalExprNotVoid es
             mdef <- liftEX $ getSMDef m k ts
             let xs = map argName (mdefArgs mdef)
                 body = apply (mkSubst (mdefTyargs mdef) us) (md_body mdef)
             withEnv xs vs (execStmtList body)
      New n@(TyClass c _) es _ ->
          do vs <- mapM evalExprNotVoid es
             cls <- liftEX $ findClass c
             l <- newObject n
             let v = NonNull l
             case (vs, class_constructor cls) of
               ([], Nothing) -> return ()
               (_, Just con) ->
                   let xs = map argName (con_args con)
                       this = con_this con
                       body = con_body con
                   in do withEnv (this:xs) (v:vs) (execStmtList body)
                         return ()
               (_, Nothing) -> bug ("Execution.evalExpr: class " ++ c ++ 
                                    " has no constructor but expression " ++
                                    showPPQ expr ++ 
                                    " instantiates the class " ++
                                    "with arguments")
             return (Just v)
      Cast n e ->
          do v <- evalExprNotVoid e
             vty <- typeOfVal v
             case vty of
               Nothing -> return (Just v)
               Just n' ->
                   do b <- liftEX $ n' `extendsClass` n
                      if b then return (Just v) 
                         else failEX ("ClassCastException while evaluating " ++ 
                                      showPPQ expr ++ ": " ++ showPPQ n' ++ 
                                      " is not a subclass of " ++ showPPQ n)
      PrimOp op es ->
          do vs <- mapM evalExprNotVoid es
             return $ Just $
               case (op, vs) of
                 (UnMinus, [IntVal i]) -> (IntVal (0-i))
                 (Add, [IntVal i, IntVal j]) -> (IntVal (i+j))
                 (Add, [StringVal s1, StringVal s2]) -> (StringVal (s1++s2))
                 (Sub, [IntVal i, IntVal j]) -> (IntVal (i-j))
                 (Mult, [IntVal i, IntVal j]) -> (IntVal (i*j))
                 (Div, [IntVal i, IntVal j]) -> (IntVal (i `div` j))
                 (Mod, [IntVal i, IntVal j]) -> (IntVal (i `mod` j))
                 (And, [BoolVal b1, BoolVal b2]) -> (BoolVal (b1 && b2))
                 (Or, [BoolVal b1, BoolVal b2]) -> (BoolVal (b1 || b2))
                 (Not, [BoolVal b]) -> (BoolVal (not b))
                 (Leq, [IntVal i, IntVal j]) -> (BoolVal (i <= j))
                 (Lt, [IntVal i, IntVal j]) -> (BoolVal (i < j))
                 (Geq, [IntVal i, IntVal j]) -> (BoolVal (i >= j))
                 (Gt, [IntVal i, IntVal j]) -> (BoolVal (i > j))
                 (Equal, _) -> BoolVal (equal vs)
                 (NotEqual, _) -> BoolVal (not $ equal vs)
                 _ -> mismatchedOp vs
          where
            mismatchedOp vs = bug ("Invalid use of primitive operator in " ++
                                   showPPQ expr ++ ", argument values: " ++ 
                                   show vs)
            equal vs =
                case vs of
                  [IntVal i, IntVal j] -> (i == j)
                  [BoolVal b1, BoolVal b2] -> (b1 == b2)
                  [StringVal s1, StringVal s2] -> (s1 == s2)
                  [StringVal _, Null] -> False
                  [Null, StringVal _] -> False
                  [NonNull l1, NonNull l2] -> (l1 == l2)
                  [NonNull _, Null] -> False
                  [Null, NonNull _] -> False
                  [Null, Null] -> True
                  _ -> mismatchedOp vs
      Lit l -> return $ Just $
          case l of
            LitInt i -> IntVal i
            LitBool b -> BoolVal b
            LitString s -> StringVal s
            LitNull -> Null

callBuiltinMethod :: BuiltinId -> Value -> [Value] -> EX (Maybe Value)
callBuiltinMethod id v vs = 
    case (id,v,vs) of
      (UtilsPrint, _, [StringVal s]) ->
          do putStrLnEX s
             return Nothing
      (UtilsPrint, _, [Null]) ->
          do putStrLnEX "null"
             return Nothing
      (UtilsPrint, _, [NonNull l]) ->
          do obj <- lookupObject l
             putStrLnEX (showPP obj)
             return Nothing
      (UtilsPrintInt, _, [IntVal i]) ->
          do putStrLnEX (show i)
             return Nothing
      (StringIntValue, StringVal s, []) ->
          do case mread s of
               Just i -> return (Just (IntVal i))
               _ -> failEX ("NumberFormatException: cannot convert " ++ 
                            show s ++ " to int")
      p -> bug ("Invalid call of builtin method: " ++ show p)

--
-- Executing statements
--

putStrLnEX :: String -> EX ()
putStrLnEX = liftEX . liftIO . putStrLn

execStmtList :: [Stmt] -> EX (Maybe Value)
execStmtList l = 
    do res <- findFirstJustM execStmt l
       case res of
         Nothing -> return Nothing
         Just x -> return x

-- The result "Nothing" indicates that the statement did not perform
-- a "return". The result "Just x" indicates that the statement performed
-- a "return" with result "x".
execStmt :: Stmt -> EX (Maybe (Maybe Value))
execStmt stmt =
    do debug ("Executing statement " ++ showPPQ stmt)
       dumpEnv
       dumpStore
       case stmt of
         StmtExpr expr -> 
             do evalExpr expr
                return Nothing
         DefineLocal _ x expr _ ->
             do v <- evalExprNotVoid expr
                modifyEnv (Map.insert x v)
                return Nothing
         AssignLocal x expr _ ->
             do v <- evalExprNotVoid expr
                modifyEnv (Map.insert x v)
                return Nothing
         AssignField expr f expr' _ ->
             do v <- evalExprNotVoid expr
                v' <- evalExprNotVoid expr'
                case v of
                  NonNull l -> updateField l f v'
                  Null -> failEX ("NullPointerException while executing " ++
                                  "statement " ++ showPPQ stmt)
                  _ -> bug ("Cannot updated field on value " ++ show v)
                return Nothing
         If expr body ->
             do v <- evalExprNotVoid expr
                case v of
                  BoolVal b ->
                      if b then findFirstJustM execStmt body else return Nothing
                  _ -> bug ("Illegal if-condition: " ++ show v)
         IfElse expr body body' ->
             do v <- evalExprNotVoid expr
                case v of
                  BoolVal b ->
                      if b then findFirstJustM execStmt body 
                           else findFirstJustM execStmt body'
                  _ -> bug ("Illegal if-condition: " ++ show v)
         While expr body ->
             let loop = do v <- evalExprNotVoid expr
                           case v of
                             BoolVal b ->
                                 if b then do res <- findFirstJustM execStmt 
                                                                    body 
                                              case res of
                                                Just _ -> return res
                                                Nothing -> loop
                                      else return Nothing
                             _ -> bug ("Illegal if-condition: " ++ show v)
             in loop
         Return Nothing _ ->
             return (Just Nothing)
         Return (Just expr) _ ->
             do v <- evalExprNotVoid expr
                return (Just (Just v))
         Skip -> return Nothing

--
-- Executing programs
--

execProgram' :: EX ()
execProgram' =
    do md <- liftEX $ checkMainOK "execute"
       l <- newObject (TyClass "Main" [])
       let this = md_this md
       info "Starting execution now ..."
       withEnv [this] [NonNull l] $ execStmtList (md_body md)
       return ()
                              
execProgram :: GI ()
execProgram = evalStateT execProgram' emptyEXState

--
-- Pretty instances
--

instance Pretty FieldMap where
    pp (FieldMap m) = 
        hsep $ punctuate comma (map (\ (name,typ) -> text name <> text "="
                                                     <> pp typ)
                                    (Map.toList m))

instance Pretty Value where
    pp (NonNull l) = pp l
    pp Null = text "null"
    pp (StringVal s) = text (show s) 
    pp (IntVal i) = text (show i) 
    pp (BoolVal b) = text (show b)

instance Pretty Object where
    pp (Object n fields) = 
        pp n <> parens (pp fields)

instance Pretty Location where
    pp (Location i) = text "l_" <> text (show i)


--
-- Debugging
--

dumpEnv :: EX ()
dumpEnv = 
    do env <- getEnv
       debug ("Env: " ++ showPP (Map.toList env))

dumpStore :: EX ()    
dumpStore =
    do store <- getStore
       debug ("Store: " ++ showPP (Map.toList store))

