{-# LANGUAGE PatternSignatures #-}
module ExpressionTyping where

import Control.Monad
-- import Debug.Trace

import AST
import Subtyping
import WellFormedness
import Types
import MethodTypes
import Environment
import GIMonad
import Pretty
import Utils
import Positions
import Logging

exprsTypes :: TEnv -> VEnv -> [Expr] -> GI [Type]
exprsTypes tenv venv es = mapM (exprType tenv venv) es

exprType :: TEnv -> VEnv -> Expr -> GI Type
exprType tenv venv expr = 
    addTrace ("Typing expression " ++ showPPQ expr ++ " failed")
             (do t <- exprType' tenv venv expr
                 debug ("Expression " ++ showPPQ expr ++ " has type " 
                        ++ showPPQ t)
                 return t)

exprType' :: TEnv -> VEnv -> Expr -> GI Type
exprType' tenv venv expr =
    case expr of
      Var x -> 
          case typeOfVar x venv of
            Just t -> return t
            Nothing -> failGI ("Type error: unbound variable " ++ showPP x)
      FAccess e f r -> 
          do t <- fieldType tenv venv e f
             fillTypeRef r t
             return t
      MCall e m vs es r ->
          do debug ("exprType', type checking method call " ++ showPPQ expr)
             t <- exprType tenv venv e
             ts <- exprsTypes tenv venv es
             debug ("exprType', t = " ++ showPPQ t ++ ", ts = " ++ 
                    showPPList ts)
             (msig', mRetroCall) <- mtype tenv m t ts
             debug ("exprType', msig' = " ++ showPPQ msig')
             let xs = ms_tyargs msig'
                 subst = mkSubst xs vs
                 us = apply subst $ map argType (ms_args msig')
                 mu :: Maybe Type = apply subst $ ms_return msig'
                 ps = apply subst $ ms_cons msig'
             debug ("exprType', xs = " ++ showPPList xs ++ ", vs = " ++ 
                    showPPList vs)
             debug ("exprType', us = " ++ showPPList us ++ ", mu = " ++ 
                    showPP mu ++ ", ps = " ++ showPPList ps)
             unless (length xs == length vs) $ 
                    failGI ("Type error: method " ++ m ++ " invoked with wrong "
                            ++ "number of type arguments")
             debug ("exprType', arity of type argument ok")
             b1 <- typesOK tenv vs
             unless b1 $ failGI ("Type error: method " ++ m ++ " invoked with "
                                 ++ "non-wellformed type arguments")
             debug ("exprType', type arguments well-formed")
             coercions <- assertSub tenv ts us
             debug ("exprType', actual argument types are subtypes of " ++
                    "formal argument types")
             b2 <- entailsAll tenv ps
             unless b2 $ failGI ("Type error: constraints " ++ showPPQ ps ++ 
                                 " of method " ++ m ++ " not satisfied")
             debug ("exprType', all constraints satisfied")
             fillRef r (mu, mRetroCall, coercions)
             case mu of
               Nothing -> throwGI VoidReturn ("Type error: method " ++ m ++ 
                                              " returns void")
               Just u -> return u
      SMCall k@(TyIface i _) ts m vs es r -> 
          do us' <- exprsTypes tenv venv es
             msig <- smtype tenv m k ts
             let xs = ms_tyargs msig
                 us = map argType (ms_args msig)
                 mu = ms_return msig
                 ps = ms_cons msig
                 subst = mkSubst xs vs
             unless (length xs == length vs) $ 
                    failGI ("Type error: static interface method " ++ m ++ 
                            " invoked with wrong number of type arguments")
             b1 <- typesOK tenv vs
             unless b1 $ failGI ("Type error: static interface method " ++ m ++ 
                                 " invoked with non-wellformed type arguments")
             b2 <- typesOK tenv ts
             unless b2 $ failGI ("Type error: static interface method " ++ m ++ 
                                 " invoked with non-wellformed implementing "
                                 ++ "types")
             coercions <- assertSub tenv us' (apply subst us)
             b3 <- entailsAll tenv (apply subst ps)
             unless b3 $ failGI ("Type error: constraints of static interface " 
                                 ++ "method" ++ m ++ " not satisfied")
             b4 <- do p <- posPlus i 1
                      anyMono <- anyM (\t -> entails tenv (MonoG t)) ts
                      return (not p || anyMono)
             unless b4 $ failGI ("Type error: cannot invoke static interface " 
                                 ++ "method " ++ m ++ "of interface " ++ 
                                 show i ++ ": interface " ++ i ++ 
                                 " is positive in its first implementing " ++
                                 "type and none of the types " ++ showPPQ ts ++
                                 " is mono")
             case mu of
               Nothing -> throwGI VoidReturn ("Type error: method " ++ m ++ 
                                              " returns void")
               Just u -> do let v = apply subst u 
                            fillRef r (v, coercions)
                            return v
      New (TyClass "String" _) _ _ ->
          failGI ("Type error: class string cannot be instantiated explicitly")
      New n@(TyClass c ts) es r ->
          do us <- exprsTypes tenv venv es
             b1 <- typeOK tenv (TyG (TyN n))
             unless b1 $ failGI ("Type error: type " ++ showPPQ n ++ 
                                 " not well-formed")
             cls <- findClass c
             let conArgTypes = 
                     case class_constructor cls of
                       Nothing -> []
                       Just constructor -> 
                           apply (mkSubst (class_tyargs cls) ts)
                                 (map argType (con_args constructor))
             coercions <- assertSub tenv us conArgTypes
             fillRef r coercions
             return (TyG (TyN n))
      Cast n e ->
          do _ <- exprType tenv venv e
             let t = TyG (TyN n)
             b <- typeOK tenv t
             if b then return t 
                else failGI ("Type error: type " ++ showPPQ n ++ 
                             " used in cast expression " ++ showPPQ e ++ 
                             " is not well-formed")
      PrimOp op es ->
          chooseGIs $
            map (\ (ts,t) ->
                      if (length ts /= length es)
                         then failGI ("Type error: arity missmatch for " ++
                                      "operator " ++ show op ++ 
                                      " used in expression " ++ 
                                      showPPQ expr)
                         else do us <- exprsTypes tenv venv es
                                 assertSub tenv us ts
                                 return t) 
                (opSigs op)
      Lit l -> return $
          case l of
            LitInt _ -> TyPrim TyInt
            LitBool _ -> TyPrim TyBool
            LitString _ -> TyG (TyN (TyClass "String" []))
            LitNull -> TyBot

fieldType :: TEnv -> VEnv -> Expr -> FieldName -> GI Type
fieldType tenv venv e f =
    do t <- exprType tenv venv e
       mn <- bound tenv t
       case mn of
         Nothing -> failGI ("Type error: cannot access field " ++ f ++
                            " on expression " ++ showPPQ e ++ 
                            ": expression has type " ++ showPPQ t)
         Just (TyClass c ts) ->
             do mfdef <- findField c f
                case mfdef of
                  Nothing -> failGI ("Type error: class " ++ show c ++
                                      " does not define a field " ++ f)
                  Just fdef ->
                      do cls <- findClass c
                         let u = fd_type fdef
                             subst = mkSubst (class_tyargs cls) ts
                         return (apply subst u)

assertSub :: TEnv -> [Type] -> [Type] -> GI [Coercion]
assertSub tenv ts us =
    do unless (length ts == length us) $
              failGI ("Type error: length of " ++ showPPList ts ++ " and " ++
                      showPPList us ++ " do not match")
       mapM (\ (t,u) -> do x <- subWithCoercion tenv t u
                           case x of
                             NoSubtype ->
                                 failGI ("Type error: " ++ showPPQ t ++ 
                                         " is not a subtype of " ++
                                         showPPQ u)
                             SubtypeWithoutCoercion ->
                                 return Nothing
                             SubtypeWithCoercion t ->
                                 return (Just t))
            (zip ts us)


opSigs :: Op -> [([Type], Type)]
opSigs op =
    case op of
      UnMinus -> [([tyInt], tyInt)]
      Not -> [([tyBool], tyBool)]
      Add -> [([tyInt,tyInt], tyInt), ([tyObject,tyObject], tyString)]
      _ | op `elem` [Sub, Mult, Div, Mod] -> [([tyInt,tyInt], tyInt)]
        | op `elem` [And, Or] -> [([tyBool,tyBool], tyBool)]
        | op `elem` [Leq, Lt, Geq , Gt] -> [([tyInt,tyInt], tyBool)]
        | op `elem` [Equal, NotEqual] -> [([tyInt,tyInt], tyBool),
                                          ([tyBool,tyBool], tyBool),
                                          ([tyObject,tyObject], tyBool)]
        | otherwise -> bug ("unsupported operator: " ++ show op)