{-# LANGUAGE GeneralizedNewtypeDeriving #-}

module GIMonad where

import Data.Set (Set)
import Control.Monad.State
import Control.Monad.Error

import AST
import qualified Symtab as S
import Logging

data GIState = GIState { gi_symtab :: S.Symtab,
                         gi_nextFreshVar :: Int,
                         gi_nextFreshTyVar :: Int }

data GIReason = VoidReturn
              | Unspecific
              deriving (Eq,Show)

data GIError = GIError GIReason [String]

newtype GI a = GI { unGI :: ErrorT GIError (StateT GIState IO) a }
    deriving (Monad,MonadIO)

instance Error GIError where
    noMsg = GIError Unspecific []
    strMsg s = GIError Unspecific [s]

getState :: GI GIState
getState = 
    GI $ lift get

setState :: GIState -> GI ()
setState s =
    GI (lift $ put s)

getSymtab :: GI S.Symtab
getSymtab =
    do s <- getState
       return (gi_symtab s)

throwGI :: GIReason -> String -> GI a
throwGI reason s = GI (throwError (GIError reason [s]))

failGI :: String -> GI a
failGI s = throwGI Unspecific s

-- catches an error matching the predicate given and clears the traceback
catchGI :: (GIReason -> Bool) -> GI a -> GI a -> GI a
catchGI pred (GI gi) handler =
    GI $ catchError gi (\ e@(GIError r _) -> if pred r
                                                then unGI handler
                                                else throwError e)

chooseGI :: GI a -> GI a -> GI a
chooseGI gi handler = catchGI (\_ -> True) gi handler

chooseGIs :: [GI a] -> GI a
chooseGIs [] = bug ("GIMonad.chooseGIs: empty list")
chooseGIs (gi:[]) = gi
chooseGIs (gi:gis) = chooseGI gi (chooseGIs gis)

addTrace :: String -> GI a -> GI a
addTrace s (GI gi) =
    GI $ catchError gi (\ (GIError r trace) -> throwError (GIError r (s:trace)))

freshTyvar :: GI TyvarName
freshTyvar = 
    do s <- getState
       let n = gi_nextFreshTyVar s
       setState (s { gi_nextFreshTyVar = (n-1) })
       return (TyvarName n)

freshTyvars :: Int -> GI [TyvarName]
freshTyvars i =
    mapM (\_ -> freshTyvar) [1..i]

freshVar :: GI VarName
freshVar =
    do s <- getState
       let n = gi_nextFreshVar s
       setState (s { gi_nextFreshVar = (n-1) })
       return (VarName n)

mkGIState :: S.Symtab -> GIState
mkGIState s = GIState { gi_symtab = s,
                        gi_nextFreshVar = -1,
                        gi_nextFreshTyVar = -1 }

withSymtab :: S.Symtab -> GI a -> IO (Either String a)
withSymtab sym (GI gi) = 
    do (res,_) <- runStateT (runErrorT gi) (mkGIState sym)
       case res of
         Right y -> return $ Right y
         Left (GIError _ trace) -> return $ Left ((unlines . reverse) trace)

onSymtab :: (S.Symtab -> a) -> GI a
onSymtab f = 
    do tab <- getSymtab
       return $ f tab

onSymtab' :: (S.Symtab -> Maybe a) -> String -> GI a
onSymtab' f err = 
    do tab <- getSymtab
       case f tab of
         Nothing -> fatal err
         Just x -> return x

findClass :: ClassName -> GI Class
findClass name =
    onSymtab' (S.findClass name) ("Unknown class: " ++ show name)

hasClass :: ClassName -> GI Bool
hasClass name = onSymtab (S.hasClass name)

allClasses :: GI [Class]
allClasses = onSymtab S.allClasses

findMDef :: ClassName -> MethodName -> GI (Maybe MethodDef)
findMDef c m = onSymtab (S.findMDef c m)

findField :: ClassName -> FieldName -> GI (Maybe FieldDef)
findField c f = onSymtab (S.findField c f)

findIface :: IfaceName -> GI Interface
findIface name = onSymtab' (S.findIface name) 
                           ("Unknown interface: " ++ show name)
allInterfaces :: GI [Interface]
allInterfaces = onSymtab S.allInterfaces

findMSig :: MethodName -> GI (Maybe (Interface, MethodSig, S.RecvIx))
findMSig m = onSymtab (S.findMSig m)

findStaticMSig :: IfaceName -> MethodName -> GI (Maybe MethodSig)
findStaticMSig i m = onSymtab (S.findStaticMSig i m)

findAllImpls :: IfaceName -> GI [Implementation]
findAllImpls i = onSymtab (S.findAllImpls i)

allImpls :: GI [Implementation]
allImpls = onSymtab S.allImpls

findAllFieldNames :: ClassName -> GI (Set FieldName)
findAllFieldNames c = onSymtab (S.findAllFieldNames c)
