{-# LANGUAGE ScopedTypeVariables #-}

module Graph (

  LDiGraph, DiGraph, Path, buildGraph, buildUnlabelledGraph, enumAllCycles

) where

import Data.Maybe
import Data.List
import Data.Map (Map)
import qualified Data.Map as Map

import Logging

type DiGraph n = LDiGraph n ()

newtype LDiGraph n l = LDiGraph (Map n [(n,l)])

type Path n l = [(n,l,n)]

buildUnlabelledGraph :: Ord n => [n] -> [(n,n)] -> DiGraph n
buildUnlabelledGraph nodes edges =
    buildGraph nodes (map (\ (from,to) -> (from, (), to)) edges)

buildGraph :: Ord n => [n] -> [(n,l,n)] -> LDiGraph n l
buildGraph nodes edges =
    if not (all (\ (from,_,to) -> from `elem` nodes && to `elem` nodes) edges)
       then bug ("Graph.buildGraph: not all edges present in nodes")
       else let emptyGraph = foldl (\m n -> Map.insert n [] m) Map.empty nodes
            in LDiGraph $ foldl (\m (from,l,to) -> 
                                     Map.insertWith (++) from [(to,l)] m)
                                emptyGraph edges

(!) :: Ord k => Map k v -> k -> v
(!) m k = case Map.lookup k m of
            Just v -> v
            Nothing -> bug ("Graph.!: node not present in graph")

-- Algorithm from the paper "A new way to enumerate cycles in graph" by Liu et al.
-- (AICT/ICIW 2006) 
enumAllCycles :: forall n l . Ord n => LDiGraph n l -> [Path n l]
enumAllCycles (LDiGraph graph) = selfLoops ++ loop initialQueue []
    where 
      selfLoops :: [Path n l]
      selfLoops =
          let nodes = Map.keys graph
          in concatMap (\n -> mapMaybe (\ (succ,l) -> 
                                            if succ == n 
                                               then Just [(n,l,n)]
                                               else Nothing)
                                       (graph ! n))
                       nodes
      initialQueue :: [(Path n l, n)]
      initialQueue =
          let nodes = Map.keys graph
          in -- find all nodes with outgoing edges
             concatMap (\head -> mapMaybe (\ (succ,l) -> 
                                               if succ > head 
                                                  then Just ([(head,l,succ)], head)
                                                  else Nothing)
                                          (graph ! head))
                       nodes
      loop :: [(Path n l, n)] {- The queue. Every path of the queue is in reverse
                                 order; that is, the real path can be obtained
                                 by reversing the list representing the path.
                                 Paths are never empty.
                                 The 2nd component of every queue element
                                 is the head of the path stored in the 1st component.
                               -}
           -> [Path n l] -> [Path n l]
      loop [] acc = acc
      loop ((rpath,head) : rest) acc =
          let (_,_,tail) : _ = rpath
              rest' = newOpenPaths (rpath,head) tail ++ rest
          in case find (\ (n,_) -> head == n) (graph ! tail) of
               Just (_,l) -> 
                   -- found a cycle
                   loop rest' (reverse ((tail,l,head):rpath) : acc)
               Nothing -> loop rest' acc
      newOpenPaths (rpath,head) tail =
          let adjs = graph ! tail
          in mapMaybe (\ (n,l) -> if not (n `pathElem` rpath) && n > head
                                     then Just ((tail,l,n):rpath, head)
                                     else Nothing) adjs

pathElem :: Eq n => n -> Path n l -> Bool
pathElem n p =
    isJust $ find (\ (s,_,e) -> n == s || n == e) p
         

{-

*Graph> let g = buildGraph [1::Int,2,3] [(1,13::Int,3),(1,12,2),(2,21,1),(2,22,2),(3,32,2),(2,222,2),(1,133,3)]
*Graph> enumAllCycles g
[[(2,222,2)],[(2,22,2)],[(1,13,3),(3,32,2),(2,21,1)],[(1,12,2),(2,21,1)],[(1,133,3),(3,32,2),(2,21,1)]]

-}