{-# LANGUAGE DeriveDataTypeable,GeneralizedNewtypeDeriving,TypeFamilies #-}

module Java (translateToJava) where

import Data.Typeable (Typeable)
import Data.Generics (Data)
import Data.Maybe 
import Data.List
import Data.Map (Map)
import qualified Data.Map as Map
import Text.PrettyPrint
import System.IO
import System.IO.Error
import Control.Monad.State
import Control.Monad.Writer
import System.FilePath hiding ( (<.>) )

import GIMonad
import qualified AST
import Pretty hiding (ppConstructor)
import Logging
import Subtyping hiding (lift)
import Environment
import MethodTypes
import Utils
import Builtins
import WellFormedness

--
-- Abstract syntax
--

data Type = TyClass AST.ClassName
          | TyPrim AST.TyPrim
    deriving (Eq,Show,Data,Typeable,Ord)
            
data Expr = 
      Var AST.VarName 
    | This
    | FAccess Expr AST.FieldName
    | MCall Expr AST.MethodName [Expr]
    | RetroCall Type           {- type of receiver -}
                [Type]         {- type and arguments -}
                AST.IfaceName  {- declaring interface -}
                Expr AST.MethodName [Expr]
    | IfaceCoercion AST.IfaceName Expr
    | New AST.ClassName [Expr]
    | Cast Type Expr
    | SynCast Type Expr
    | PrimOp AST.Op [Expr]
    | Lit AST.Lit
    deriving (Eq,Show,Data,Typeable)

data Stmt = 
      StmtExpr Expr
    | DefineLocal Type AST.VarName Expr
    | AssignLocal AST.VarName Expr
    | AssignField Expr AST.FieldName Expr
    | If Expr [Stmt]
    | IfElse Expr [Stmt] [Stmt]
    | While Expr [Stmt]
    | Return (Maybe Expr)
    | Skip
    deriving (Eq,Show,Data,Typeable)

data Class = Class { class_name :: AST.ClassName,
                     class_super :: Maybe AST.ClassName,
                     class_implements :: [AST.IfaceName],
                     class_constructor :: Maybe Constructor,
                     class_fields :: [FieldDef],
                     class_methods :: [MethodDef] }
    deriving (Eq,Show,Data,Typeable)

newtype MethodArg = MethodArg { unMethodArg :: (Type, AST.VarName) }
    deriving (Eq,Show,Data,Typeable)

argType :: MethodArg -> Type
argType (MethodArg (t, _)) = t

argName :: MethodArg -> AST.VarName
argName (MethodArg (_, x)) = x

data Constructor = Constructor { con_args :: [MethodArg],
                                 con_body :: [Stmt] }
    deriving (Eq,Show,Data,Typeable)

data FieldDef = FieldDef { fd_type :: Type,
                           fd_name :: AST.FieldName }
    deriving (Eq,Show,Data,Typeable)

data MethodSig = MethodSig { ms_return :: Maybe Type,
                             ms_name :: AST.MethodName,
                             ms_args :: [MethodArg] }
    deriving (Eq,Show,Data,Typeable)

data MethodDef = MethodDef { md_sig :: MethodSig,
                             md_body :: [Stmt] }
    deriving (Eq,Show,Data,Typeable)

data Implementation = Implementation { impl_iface :: AST.ClassName,
                                       impl_classes :: [AST.ClassName],
                                       impl_methods :: [MethodDef] }
    deriving (Eq,Show,Data,Typeable)

data DispatchMarker = NoDispatch 
                    | Dispatch Int {- dispatch index, starts at 0 -}
    deriving (Eq,Show,Data,Typeable)

data DispatchSig = DispatchSig { 
      disp_name :: AST.MethodName,
      disp_markers :: [DispatchMarker] {- including the receiver argument -} 
}    deriving (Eq,Show,Data,Typeable)

data Interface = Interface { iface_name :: AST.ClassName,
                             iface_sigs :: [MethodSig],
                             iface_dispatch :: [DispatchSig] }
    deriving (Eq,Show,Data,Typeable)

data Def = ImplDef Implementation | IfaceDef Interface | ClassDef Class
    deriving (Eq,Show,Data,Typeable)

data Program = Program { unProgram :: [Def] }
    deriving (Eq,Show,Data,Typeable)

--
-- Term substitution
--

-- assumptions:
-- * all expressions in the range of a substitution are closed
-- * substitution is never applied to a statement that writes
--   one of the variable in the domain of the substitution
newtype ExprSubst = ExprSubst { unExprSubst :: (Map AST.VarName Expr) }

instance Show ExprSubst where
    showsPrec i (ExprSubst m) = showsPrec i m

{-
instance Pretty ExprSubst where
    pp (ExprSubst m) =
        text "ExprSubst" <> braces (sep (punctuate comma 
          (map (\ (x,t) -> pp x <> text ": " <> pp t) (Map.toList m))))
-}

mkSubst :: [AST.VarName] -> [Expr] -> ExprSubst
mkSubst xs ts = 
    if length xs /= length ts 
       then bug ("Java.mkSubst: argument lists with different lengths: " ++
                 show xs ++ "; " ++ show ts)
       else
           ExprSubst (Map.fromList (zip xs ts))

isInDomain :: AST.VarName -> ExprSubst -> Bool
isInDomain x subst = isJust $ Map.lookup x (unExprSubst subst)
                       
class Subst t where
    apply :: ExprSubst -> t -> t

instance Subst t => Subst [t] where 
    apply s l = map (apply s) l

instance Subst t => Subst (Maybe t) where
    apply _ Nothing = Nothing
    apply s (Just x) = Just (apply s x)

instance Subst Expr where
    apply subst expr = 
        case expr of
          Var x -> case Map.lookup x (unExprSubst subst) of
                     Just e -> e
                     Nothing -> Var x
          This -> This
          FAccess e f -> FAccess (apply subst e) f
          MCall e m es -> MCall (apply subst e) m (apply subst es)
          RetroCall t ts i e m es -> 
              RetroCall t ts i (apply subst e) m (apply subst es)
          IfaceCoercion i e ->
              IfaceCoercion i (apply subst e)
          New c es -> New c (apply subst es)
          Cast t e -> Cast t (apply subst e)
          SynCast t e -> SynCast t (apply subst e)
          PrimOp op es -> PrimOp op (apply subst es)
          Lit l -> Lit l

instance Subst Stmt where
    apply subst stmt = 
        case stmt of
          StmtExpr e -> StmtExpr (apply subst e)
          DefineLocal t x e ->
              if isInDomain x subst
                then bug ("cannot apply substitution " ++ show subst ++ 
                          " to statement " ++ show stmt)
                else DefineLocal t x (apply subst e)
          AssignLocal x e ->
              if isInDomain x subst
                then bug ("cannot apply substitution " ++ show subst ++ 
                          " to statement " ++ show stmt)
                else AssignLocal x (apply subst e)
          AssignField e1 f e2 ->
              AssignField (apply subst e1) f (apply subst e2)
          If e stmts ->
              If (apply subst e) (apply subst stmts)
          IfElse e stmts1 stmts2 ->
              IfElse (apply subst e) (apply subst stmts1) (apply subst stmts2)
          While e stmts ->
              While (apply subst e) (apply subst stmts)
          Return e ->
              Return (apply subst e)
          Skip -> Skip

substThis :: Subst t => AST.VarName -> t -> t
substThis x t = apply (mkSubst [x] [This]) t

--
-- Translation from JavaGI
--

debugTrans :: MonadIO m => String -> m ()
debugTrans = debug

class FromAST a where
    type Result a
    fromAST :: TEnv -> a -> GI (Result a)

instance FromAST a => FromAST [a] where
    type Result [a] = [Result a]
    fromAST tenv = mapM (fromAST tenv)

instance FromAST AST.TyN where
    type Result AST.TyN = AST.ClassName
    fromAST _ cls = return $ AST.tyc_name cls

instance FromAST AST.TyK where
    type Result AST.TyK = AST.ClassName
    fromAST _ cls = return $ AST.tyi_name cls

instance FromAST AST.Type where
    type Result AST.Type = Type
    fromAST tenv t = 
        do -- debugTrans ("Translating type " ++ showPP t)
           case t of
             AST.TyG (AST.TyN cls) -> do c <- fromAST tenv cls
                                         return $ TyClass c
             AST.TyPrim prim -> return $ TyPrim prim
             AST.TyK iface -> return $ TyClass (AST.tyi_name iface)
             AST.TyG (AST.TyVar x) ->
                 do (AST.TyN n) <- gminimal tenv $ 
                           AST.tyObjectG : [AST.TyN n | AST.TyG (AST.TyN n) 
                                                        <- transBounds x tenv]
                    c <- fromAST tenv n
                    return $ TyClass c
             _ -> bug ("unexpected type: " ++ show t)

mkCast :: TEnv -> AST.Ref AST.Type -> Expr -> GI Expr
mkCast tenv r expr = 
    do t <- AST.readRef r
       t' <- fromAST tenv t
       return $ SynCast t' expr

insertCoercion :: Maybe AST.TyK -> Expr -> Expr
insertCoercion Nothing e = e
insertCoercion (Just (AST.TyIface iface _)) e = IfaceCoercion iface e

-- cast optimization missing here
instance FromAST AST.Expr where
    type Result AST.Expr = Expr
    fromAST tenv expr = 
        do debugTrans ("Translating expression " ++ show expr)
           res <- case expr of
             AST.Var x -> return $ Var x
             AST.FAccess e f r -> 
                 do e' <- fromAST tenv e 
                    mkCast tenv r (FAccess e' f)
             AST.MCall e m _ es r ->
                 do e' <- fromAST tenv e
                    es' <- fromAST tenv es
                    (mt, mRetroCall, argCoercions) <- AST.readRef r
                    callLhs <- case mRetroCall  of
                                 Just (u, us, i) -> 
                                     do u' <- fromAST tenv u 
                                        us' <- fromAST tenv us
                                        return $ RetroCall u' us' i
                                 Nothing -> return MCall
                    let call = callLhs e' m (map2 insertCoercion 
                                                  argCoercions es')
                    case mt of
                      Nothing -> return call
                      Just t ->
                            do t' <- fromAST tenv t
                               return $ SynCast t' call
             AST.SMCall _ _ _ _ _ _ -> 
                 fail ("cannot translate calls of static interface methods")
             AST.New n es r -> 
                 do argCoercions <- AST.readRef r
                    es' <- fromAST tenv es
                    n' <- fromAST tenv n
                    return $ New n' (map2 insertCoercion argCoercions es')
             AST.Cast n e ->
                 do n' <- fromAST tenv n
                    e' <- fromAST tenv e
                    return $ Cast (TyClass n') e'
             AST.PrimOp op es ->
                 do es' <- fromAST tenv es
                    return $ PrimOp op es'
             AST.Lit lit -> return $ Lit lit
           debugTrans ("Result of translating " ++ show expr ++ ": " ++ 
                                                show res)
           return res

instance FromAST AST.Stmt where
    type Result AST.Stmt = Stmt
    fromAST tenv stmt =
        do debugTrans ("Translating statement " ++ showPP stmt)
           case stmt of
             AST.StmtExpr e ->
                 do e' <- fromAST tenv e
                    return (StmtExpr e')
             AST.DefineLocal t x e r ->
                 do coercion <- AST.readRef r
                    t' <- fromAST tenv t
                    e' <- fromAST tenv e
                    return (DefineLocal t' x (insertCoercion coercion e'))
             AST.AssignLocal x e r ->
                 do coercion <- AST.readRef r
                    e' <- fromAST tenv e
                    return (AssignLocal x (insertCoercion coercion e'))
             AST.AssignField e1 f e2 r ->
                 do coercion <- AST.readRef r
                    e1' <- fromAST tenv e1
                    e2' <- fromAST tenv e2
                    return (AssignField e1' f (insertCoercion coercion e2'))
             AST.If e stmts ->
                 do e' <- fromAST tenv e
                    stmts' <- fromAST tenv stmts
                    return (If e' stmts')
             AST.IfElse e stmts1 stmts2 ->
                 do e' <- fromAST tenv e
                    stmts1' <- fromAST tenv stmts1
                    stmts2' <- fromAST tenv stmts2
                    return $ IfElse e' stmts1' stmts2'
             AST.While e stmts ->
                 do e' <- fromAST tenv e
                    stmts' <- fromAST tenv stmts
                    return (While e' stmts')
             AST.Return Nothing _ -> return $ Return Nothing
             AST.Return (Just e) r -> 
                 do coercion <- AST.readRef r
                    e' <- fromAST tenv e
                    return $ Return (Just (insertCoercion coercion e'))
             AST.Skip -> return Skip

instance FromAST a => FromAST (Maybe a) where
    type Result (Maybe a) = Maybe (Result a)
    fromAST _ Nothing = return $ Nothing
    fromAST tenv (Just x) = 
        do x' <- fromAST tenv x
           return $ Just x'

instance FromAST (AST.MethodSigGeneric a) where
    type Result (AST.MethodSigGeneric a) = MethodSig
    fromAST tenv msig = 
        do ret <- fromAST tenv (AST.ms_return msig)
           args <- fromAST tenv (AST.ms_args msig)
           return $ MethodSig { ms_return = ret,
                                ms_name = AST.ms_name msig,
                                ms_args = args }

instance FromAST AST.MethodArg where
    type Result AST.MethodArg = MethodArg
    fromAST tenv (AST.MethodArg (t,x)) = do t' <- fromAST tenv t
                                            return $ MethodArg (t', x)

instance FromAST AST.Constructor where
    type Result AST.Constructor = Constructor
    fromAST tenv con = 
        do args <- fromAST tenv (AST.con_args con)
           body <- fromAST tenv (AST.con_body con)
           return $ Constructor { con_args = args,
                                  con_body = substThis (AST.con_this con) body }

instance FromAST AST.FieldDef where
    type Result AST.FieldDef = FieldDef
    fromAST tenv field = 
        do debug ("Translating field " ++ AST.fd_name field)
           ty <- fromAST tenv (AST.fd_type field)
           return $ FieldDef { fd_type = ty,
                               fd_name = AST.fd_name field }

castArguments :: [Type] -> [MethodArg] -> GI ([Stmt], [MethodArg])
castArguments shouldBeTypes currentArgs = 
    if length shouldBeTypes /= length currentArgs
       then bug ("castArguments: should-be-types " ++ show shouldBeTypes ++ 
                 " and method arguments " ++ show currentArgs ++ 
                 " do not have the same length")
       else do l <- mapM castArgument (zip shouldBeTypes currentArgs)
               let (mstmts, args) = unzip l
               return (catMaybes mstmts, args)
    where
      castArgument (shouldBe, currentArg) = 
              let (cur, x) = unMethodArg currentArg
              in if shouldBe == cur
                    then return (Nothing, currentArg)
                    else do fresh <- freshVar
                            return (Just (DefineLocal cur x 
                                                 (SynCast cur (Var fresh))),
                                    MethodArg (shouldBe, fresh))

fromASTMDef :: TEnv -> AST.ClassName -> AST.MethodDef -> GI MethodDef
fromASTMDef tenv' cls md = 
    do debug ("Translating method " ++ AST.mdefName md ++ " of class " ++ cls)
       tenv <- extendTEnvWithMSig tenv' (AST.md_sig md)
       sig <- fromAST tenv (AST.md_sig md)
       body <- fromAST tenv (AST.md_body md)
       (argTypes, ret) <- mtypeMax tenv' cls (AST.mdefName md)
       (convAssigns, newArgs) <- castArguments argTypes (ms_args sig)
       let newSig = MethodSig { ms_return = ret,
                                ms_name = ms_name sig,
                                ms_args = newArgs }
       return $ MethodDef { md_sig = newSig,
                            md_body = substThis (AST.md_this md) 
                                                (convAssigns ++ body) }

instance FromAST AST.Class where
    type Result AST.Class = Class
    fromAST _ cls = 
        do debug ("Translating class " ++ AST.class_name cls)
           tenv <- mkTEnv (AST.class_cons cls) (AST.class_tyargs cls)
           super <- fromAST tenv (AST.class_super cls)
           con <- fromAST tenv (AST.class_constructor cls)
           fs <- fromAST tenv (AST.class_fields cls)
           let name = AST.class_name cls
           ms <- mapM (fromASTMDef tenv name) (AST.class_methods cls)
           return $ Class { class_name = name,
                            class_super = super,
                            class_implements = [],
                            class_constructor = con,
                            class_fields = fs,
                            class_methods = ms }

fromASTImplMDef :: TEnv -> AST.ClassName -> Interface -> AST.MethodDef 
                -> GI MethodDef
fromASTImplMDef tenv' recv iface md = 
    do debug ("Translating method " ++ AST.mdefName md ++ " of receiver " ++ 
              recv ++ " in implementation for " ++ iface_name iface)
       let sigInIface = case find (\sig -> ms_name sig == AST.mdefName md) 
                                  (iface_sigs iface) 
                        of Just s -> s
                           Nothing -> bug ("method " ++ AST.mdefName md ++ 
                                           " not found in interface " ++ 
                                           show iface)
       tenv <- extendTEnvWithMSig tenv' (AST.md_sig md)
       sig <- fromAST tenv (AST.md_sig md)
       let mdArgs = MethodArg (TyClass recv, AST.md_this md) :
                    ms_args sig
       (convAssigns, newArgs) <- castArguments (map argType $ 
                                                    ms_args sigInIface)
                                               mdArgs
       body <- fromAST tenv (AST.md_body md)
       let newSig = MethodSig { ms_return = ms_return sigInIface,
                                ms_name = ms_name sigInIface,
                                ms_args =  newArgs }
       return $ MethodDef { md_sig = newSig,
                            md_body = convAssigns ++ body }

fromASTImpl :: AST.Implementation -> Interface -> GI Implementation
fromASTImpl impl iface =
        do debug ("Translating implementation " ++ 
                  showPPImplHeader (AST.impl_iface impl) 
                                   (AST.impl_impl_types impl))
           tenv <- mkTEnv (AST.impl_cons impl) (AST.impl_tyargs impl)
           ifaceName <- fromAST tenv (AST.impl_iface impl)
           classes <- fromAST tenv (AST.impl_impl_types impl)
           methods <- mapM (\ (cls, rd) -> 
                                mapM (fromASTImplMDef tenv cls iface)
                                     (AST.rcdef_defs rd))
                           (zip classes (AST.impl_receiver_defs impl))
           return $ Implementation { impl_iface = ifaceName,
                                     impl_classes = classes,
                                     impl_methods = concat methods }


fromASTReceiverSig :: TEnv -> [AST.TyvarName] -> AST.ReceiverSig 
                   -> GI ([MethodSig], [DispatchSig])
fromASTReceiverSig tenv implVars recv =
    do debug ("Translating receiver signature " ++ showPP (AST.rcsig_name recv))
       let this = AST.TyG (AST.TyVar (AST.rcsig_name recv))
           recvIndex = case implIndex this of
                         NoDispatch -> 
                             bug ("fromASTReceiverSig: receiver name not " ++
                                  "among implementing types")
                         x -> x
       sigs <- mapM (\sig -> do sig' <- fromAST tenv sig
                                x <- freshVar
                                t <- fromAST tenv this
                                return $ MethodSig { 
                                             ms_return = ms_return sig',
                                             ms_name = ms_name sig',
                                             ms_args = MethodArg (t, x) :
                                                       ms_args sig' 
                                         })
                    (AST.rcsig_sigs recv)
       let disps = map (\sig -> DispatchSig { 
                                  disp_name = AST.ms_name sig,
                                  disp_markers = recvIndex : 
                                                 map (implIndex . AST.argType)
                                                     (AST.ms_args sig) 
                                })
                       (AST.rcsig_sigs recv)
       return (sigs, disps)
    where implIndex (AST.TyG (AST.TyVar x)) =
              case elemIndex x implVars of
                Just i -> Dispatch i
                Nothing -> NoDispatch
          implIndex _ = NoDispatch

fromASTMethodSigForInterfaceType :: TEnv -> AST.IfaceName -> AST.TyvarName 
                                 -> AST.MethodSig -> GI MethodSig
fromASTMethodSigForInterfaceType tenv iface implVar msig =
    let msigFixed = 
            -- if implVar is the result type of a method then
            -- we can replace it with the interface type
            case AST.ms_return msig of
              Just (AST.TyG (AST.TyVar x))
                | x == implVar ->
                    msig { AST.ms_return = 
                               Just $ AST.TyK (AST.TyIface iface []) }
                | otherwise ->
                    msig
              Just _ -> msig
              Nothing -> msig
    in fromAST tenv msigFixed

generateWrapperMethod :: TEnv -> AST.IfaceName -> AST.TyvarName 
                      -> AST.MethodSig -> GI MethodDef
generateWrapperMethod tenv iface implVar msig = 
    do sig <- fromASTMethodSigForInterfaceType tenv iface implVar msig
       let argTypes = map argType (ms_args sig)
           argValues = map (Var . argName) (ms_args sig)
           call = RetroCall (TyClass "Object") argTypes iface 
                            (FAccess This "wrapped") (ms_name sig) argValues
           stmt = case ms_return sig of
                    Nothing -> StmtExpr call
                    -- cast optimization ignored here
                    Just t -> Return (Just (SynCast t call))
       return $ MethodDef sig [stmt]

-- At this point, we completely ignore interface inheritance
instance FromAST AST.Interface where
    type Result AST.Interface = ([Interface], [Class])
    fromAST _ iface = 
        do debug ("Translating interface " ++ AST.iface_name iface)
           tenv <- mkTEnv (AST.iface_cons iface ++ 
                              map AST.ConR (AST.iface_impl_cons iface))
                          (AST.iface_tyargs iface ++ AST.iface_impl_types iface)
           l <- mapM (fromASTReceiverSig tenv (AST.iface_impl_types iface)) 
                     (AST.iface_receiver_sigs iface)
           let methodSigs = concatMap fst l
               dispatchSigs = concatMap snd l
               implIface = Interface { 
                                 iface_name = implIfaceName $ 
                                              AST.iface_name iface,
                                 iface_sigs = methodSigs,
                                 iface_dispatch = dispatchSigs 
                               }
           (typeIface, wrapperClass) <-
               case AST.iface_impl_types iface of
                 [x] -> 
                     do sigs <- mapM
                                  (fromASTMethodSigForInterfaceType
                                     tenv
                                     (AST.iface_name iface)
                                     x)
                                  (AST.rcsig_sigs $ 
                                      head (AST.iface_receiver_sigs iface))
                        methods <- mapM
                                     (generateWrapperMethod
                                       tenv
                                       (AST.iface_name iface)
                                       x)
                                    (AST.rcsig_sigs $ 
                                        head (AST.iface_receiver_sigs iface))
                        return ([Interface {
                                   iface_name = AST.iface_name iface,
                                   iface_sigs = sigs,
                                   iface_dispatch = []
                                 }],
                                [Class { 
                                   class_name = ifaceWrapperName $ 
                                                AST.iface_name iface,
                                   class_super = Nothing,
                                   class_implements = [AST.iface_name iface],
                                   class_constructor = 
                                     Just $ Constructor { 
                                       con_args = [MethodArg (TyClass "Object", 
                                                              AST.VarName 1)],
                                       con_body = [AssignField 
                                                     This 
                                                     "wrapped" 
                                                     (Var (AST.VarName 1))] 
                                     },
                                   class_fields = [FieldDef (TyClass "Object") 
                                                            "wrapped"],
                                   class_methods = methods }])
                 _ -> return ([], [])
           return (implIface : typeIface, wrapperClass)
          
translateProgram :: AST.Program -> GI Program
translateProgram (AST.Program defs) = 
    do debug ("Translating program")
       let (ifaces, rest) = partition (\def -> case def of
                                                 AST.IfaceDef _ -> True
                                                 _              -> False) defs
       ifacesTrans' <- mapMaybeM translateIface ifaces
       let (ifacesTrans, wrappers) = (concat (map fst ifacesTrans'),
                                      concat (map snd ifacesTrans'))
       restTrans <- mapMaybeM (translateDef ifacesTrans) rest
       return (Program (map IfaceDef ifacesTrans ++ map ClassDef wrappers ++
                        restTrans))
    where translateIface (AST.IfaceDef iface) = 
              if isBuiltinClass (AST.iface_name iface) 
                 then return Nothing
                 else do i <- fromAST emptyTEnv iface
                         return (Just i)
          translateIface d = bug ("translatedIface: unexpected definition: " 
                                  ++ show d)
          translateDef ifaces (AST.ImplDef impl) = 
              do let i = (AST.tyi_name . AST.impl_iface) impl
                 d <- fromASTImpl impl (fromJust $ 
                                        find (\iface -> iface_name iface == 
                                                        implIfaceName i) 
                                             ifaces)
                 return $ Just (ImplDef d)
          translateDef _ (AST.ClassDef cls) = 
              if isBuiltinClass (AST.class_name cls)
                 then return Nothing
                 else do d <- fromAST emptyTEnv cls
                         return $ Just (ClassDef d)
          translateDef _ d = bug ("translatedDef: unexpected definition: " ++ 
                                  show d)

--
-- Auxiliaries for translation
--

mtypeMax :: TEnv -> AST.ClassName -> AST.MethodName -> GI ([Type], Maybe Type)
mtypeMax tenv c m =
    do debug ("searching for highest definition of method " ++ m ++
              " starting in class " ++ c)
       cls <- findClass c
       msuper <- 
           case AST.class_super cls of
             Nothing -> return Nothing
             Just super@(AST.TyClass d _) ->
                 do definedInSuper <- (mtypeClass m super >> return True)
                                      `chooseGI` return False
                    if definedInSuper
                       then return $ Just d
                       else return Nothing
       case msuper of
         Just d -> mtypeMax tenv d m
         Nothing ->
             do msig <- mtypeClass m (AST.TyClass c 
                                             (map (AST.TyG . AST.TyVar) 
                                                      (AST.class_tyargs cls)))
                args <- fromAST tenv (map AST.argType (AST.ms_args msig))
                ret <- fromAST tenv (AST.ms_return msig)
                return (args, ret)

extendTEnvWithMSig :: TEnv -> AST.MethodSig -> GI TEnv
extendTEnvWithMSig tenv msig = 
    do tenv' <- extendTEnv tenv (AST.ms_cons msig)
       return $ addTyvars tenv' (AST.ms_tyargs msig)

--
-- Pretty printing of Java code
--

instance Pretty Type where
    pp (TyClass c) = transClassName c
    pp (TyPrim AST.TyInt) = text "int"
    pp (TyPrim AST.TyBool) = text "boolean"

data JTState = JTState { t_fresh :: Int,
                         t_vars :: [(Type,String)] }

initJTState :: JTState
initJTState = JTState { t_fresh = 0, t_vars = [] }

type JT = State JTState

freshVars :: Int -> JT [String]
freshVars n =
    do i <- gets t_fresh
       modify (\s -> s { t_fresh = i + n })
       return $ map (\i -> "fresh__" ++ show i) [i..(i+n-1)]

declareVars :: [Type] -> [String] -> JT ()
declareVars ts xs = 
    if length ts /= length xs
       then fail ("Java.declareVars: arity mismatch between types " ++ show ts
                  ++ " and variables " ++ show xs)
       else do l <- gets t_vars
               modify (\s -> s { t_vars = l ++ zip ts xs })

rtClass :: String
rtClass = "javagi.runtime.CoreRT"

dispatchVector :: AST.MethodName -> String
dispatchVector m = m ++ "__JavaGI_DispatchVector"

implIfaceName :: AST.IfaceName -> AST.IfaceName
implIfaceName i = i ++ "__JavaGI_Impl"

implClass :: AST.IfaceName -> [AST.ClassName] -> AST.ClassName
implClass i cs = i ++ (foldr (++) "__" cs) ++ "__JavaGI_ImplClass"

ifaceWrapperName :: AST.IfaceName -> AST.ClassName
ifaceWrapperName i = i ++ "__JavaGI_Wrapper"

mkJavaVar :: AST.VarName -> Doc
mkJavaVar (AST.VarName i) = 
    if i >= 0 then text $ "x__" ++ show i
              else text $ "y__" ++ show (-i)

transClassNameStr :: AST.ClassName -> String
transClassNameStr c = 
    if isBuiltinClass c && c /= "Object" then "GI" ++ c else c

transClassName :: AST.ClassName -> Doc
transClassName = text . transClassNameStr

exprToJava :: Expr -> JT Doc
exprToJava (Var i) = return $ mkJavaVar i
exprToJava This = return $ text "this"
exprToJava (FAccess e f) = 
    do e' <- exprToJava e
       return $ e' <.> text f
exprToJava (MCall e m args) = 
    do e' <- exprToJava e
       args' <- mapM exprToJava args
       return $ e' <.> text m <> parens (pp args')
exprToJava (RetroCall t ts i e m es) = 
    do -- Create fresh variables x0,...,xn for receiver and arguments
       -- A possible optimzation is to generate variables only for
       -- non-values.
       vars <- freshVars (length es + 1)
        -- We must declare these variables somewhere outside
       declareVars (t:ts) vars
       -- Generate the argument to the dispatch array.
       -- For an argument e of primitive type, this is
       -- "javagi.runtime.CoreRT.intDescriptor(x = e)", where
       -- x is the fresh variable of the argument. For a
       -- reference argument, this is "x = e".
       args <- mapM exprToJava (e:es)
       let arrayArgs = map (\ (t,x,e) -> case t of
                                            TyPrim AST.TyInt -> 
                                                intDescriptor x e
                                            TyPrim AST.TyBool -> 
                                                boolDescriptor x e
                                            _ -> text x <=> pp e)
                           (zip3 (t:ts) vars args)
           dispatchArray = parens (text "new Object[]" <> 
                                   braces (pp arrayArgs))
           impl = text rtClass <.> text "getMethods" <>
                  parens (text (implIfaceName i) <.> text "class" <> comma <+>
                          text (implIfaceName i) <.> 
                          text (dispatchVector m) <> comma <+>
                          dispatchArray)
           expr = parens (parens (text (implIfaceName i)) <+> impl) <.>
                  text m <> parens (pp (map text vars))
       return expr
    where
      intDescriptor = genDescriptor "intDescriptor"
      boolDescriptor = genDescriptor "boolDescriptor"
      genDescriptor :: String -> String -> Doc -> Doc
      genDescriptor s x e = text rtClass <.> text s <> parens (text x <+> e)
exprToJava (IfaceCoercion i e) =
    do e' <- exprToJava e
       return $ parens $ text "new" <+> text (ifaceWrapperName i) <> parens e'
exprToJava (New n args) =
    do args' <- mapM exprToJava args
       return $ parens $ text "new" <+> transClassName n <> parens (pp args')
exprToJava (Cast n e) =
    do e' <- exprToJava e
       return $ parens (parens (pp n) <+> e')
exprToJava (SynCast n e) =
    do e' <- exprToJava e
       return $ parens (parens (pp n) <+> e' 
                        <+> text "/* synthetic cast */")
exprToJava e@(PrimOp op args) = 
    do args' <- mapM exprToJava args
       case (op, args') of
         (AST.UnMinus, [e']) -> return $ parens (char '-' <> e')
         (AST.Not, [e']) -> return $ parens (char '!' <> e')
         (_, [e1,e2]) -> return $ parens (e1 <+> pp op <+> e2)
         _ -> bug ("invalid use of primitive operator: " ++ show e)
exprToJava (Lit l) =
    return $ case l of
               AST.LitInt i -> text (show i)
               AST.LitBool True -> text "true"
               AST.LitBool False -> text "true"
               AST.LitString s -> text ("(new GIString(" ++ show s ++ "))")
               AST.LitNull -> text "null"

stmtToJava :: Stmt -> JT Doc
stmtToJava (StmtExpr e) = 
    do e' <- exprToJava e
       return $ e' <> semi
stmtToJava (DefineLocal t x e) = 
    do e' <- exprToJava e
       return $ pp t <+> mkJavaVar x <=> e' <> semi
stmtToJava (AssignLocal x e) = 
    do e' <- exprToJava e
       return $ mkJavaVar x <=> e' <> semi
stmtToJava (AssignField e1 f e2) = 
    do e1' <- exprToJava e1
       e2' <- exprToJava e2
       return $ e1' <.> text f <=> e2' <> semi
stmtToJava (If e l) = 
    do e' <- exprToJava e
       l' <- mapM stmtToJava l
       return $ text "if" <+> e' <+> lbrace 
                $+$ nest indent (ppListNewline l')
                $+$ rbrace
stmtToJava (IfElse e l1 l2) = 
    do e' <- exprToJava e
       l1' <- mapM stmtToJava l1
       l2' <- mapM stmtToJava l2
       return $ text "if" <+> e' <+> lbrace 
                $+$ nest indent (ppListNewline l1')
                $+$ rbrace <+> text "else" <+> lbrace
                $+$ nest indent (ppListNewline l2')
                $+$ rbrace
stmtToJava (While e l) = 
    do e' <- exprToJava e
       l' <- mapM stmtToJava l
       return $ text "while" <+> e' <+> lbrace 
                $+$ nest indent (ppListNewline l')
                $+$ rbrace
stmtToJava (Return Nothing) = return $ text "return"
stmtToJava (Return (Just e)) =
    do e' <- exprToJava e
       return $ text "return" <+> e' <> semi
stmtToJava Skip = return semi

instance Pretty MethodArg where
    pp (MethodArg (t, x)) = pp t <+> mkJavaVar x

public :: Doc
public = text "public"

instance Pretty MethodSig where
    pp sig = 
        public <+>
        (case ms_return sig of
           Nothing -> text "void"
           Just t -> pp t) <+> 
        text (ms_name sig) <> parens (pp (ms_args sig))

ppMethodBody :: [Stmt] -> Doc
ppMethodBody l =
    let (body, state) = runState (mapM stmtToJava l) initJTState
        declCode = map (\ (t,x) -> pp t <+> text x <> semi) (t_vars state)
    in ppListNewline (declCode ++ body)

instance Pretty MethodDef where
    pp m = pp (md_sig m) <+> lbrace
           $+$ nest indent (ppMethodBody (md_body m))
           $+$ rbrace

ppConstructor :: AST.ClassName -> Maybe Constructor -> Doc
ppConstructor _ Nothing = empty
ppConstructor name (Just c) = 
    public <+> text name <> parens (pp (con_args c)) <+> lbrace
    $+$ nest indent (ppMethodBody (con_body c))
    $+$ rbrace

instance Pretty FieldDef where
    pp fd = public <+> pp (fd_type fd) <+> text (fd_name fd) <> semi

instance Pretty Class where
    pp c = text "class" <+> transClassName (class_name c) <> 
           (case class_super c of
              Nothing -> text " extends Object"
              Just n -> text " extends" <+> transClassName n)  <+>
           (case class_implements c of
              [] -> empty
              l -> text " implements" <+> pp (map text l))
           <+> lbrace
           $+$ nest indent (ppListNewline (class_fields c))
           $+$ nest indent (ppConstructor (transClassNameStr (class_name c)) 
                                          (class_constructor c))
           $+$ nest indent (ppListNewline (class_methods c))
           $+$ nest indent toString
           $+$ rbrace
      where
        toString = text "public String toString()" <+> lbrace $+$
                   nest indent (text "return \"" <> 
                                transClassName (class_name c) <>
                                parens (pp (map ppField (class_fields c))) <> 
                                text "\";") $+$
                   rbrace
        ppField f = text (fd_name f) <> (text "=\\\"\" +") <+> text (fd_name f)
                    <+> text "+ \"\\\""
                                        

instance Pretty Implementation where
    pp impl = text "class" <+> 
              text (implClass (impl_iface impl) (impl_classes impl)) <+>
              text "implements" <+> text (implIfaceName (impl_iface impl))
              <+> lbrace
              $+$ nest indent (ppListNewline (impl_methods impl))
              $+$ rbrace

instance Pretty DispatchSig where
    pp sig = text "public int[]" <+> text (dispatchVector (disp_name sig)) <=>
             text "new int[]" <> braces (ppMarkers (disp_markers sig)) <> semi
      where ppMarkers l = pp $ flatten $ 
                          filter (\p -> isDispatch (fst p)) $ 
                          zip l [(0::Int)..]
            flatten = concatMap (\ (Dispatch i, y) -> [i,y])
            isDispatch NoDispatch = False
            isDispatch _ = True

instance Pretty Interface where
    pp iface = text "interface" <+> text (iface_name iface) 
               <+> lbrace
               $+$ nest indent (ppListNewline (map (\s -> pp s <> semi)
                                                   (iface_sigs iface)))
               $+$ nest indent (ppListNewline (iface_dispatch iface))
               $+$ rbrace

instance Pretty Def where
    pp (IfaceDef i) = pp i
    pp (ClassDef i) = pp i
    pp (ImplDef i) = pp i

instance Pretty Program where
    pp (Program defs) = ppListNewline defs

--
-- Main entry point into this module
--
translateToJava :: AST.Program -> FilePath -> GI ()
translateToJava prog out =
    do checkMainOK "translate"
       p <- translateProgram prog
       debug ("Translated program: " ++ show p)
       let impls = mapMaybe asImplementation (unProgram p)
           code = headerComment ++ "\n\n" ++ showPP p ++ "\n\n" ++ 
                  mainClass impls ++ builtinJavaCode
       writeOk <- liftIO checkWriteOk
       unless writeOk $ 
          fatal ("Refusing to write to existing file " ++ out ++
                 ": the file has not been generated by core-javagic")
       debug ("About to write java code to " ++ out)
       liftIO $ writeFile out code
       info ("Wrote java code to " ++ out)
    where
      asImplementation :: Def -> Maybe Implementation
      asImplementation (ImplDef i) = Just i
      asImplementation _ = Nothing
      mainClass impls = 
          let name = takeBaseName out
          in "class " ++ name ++ " {\n" ++
             "  public static void main(String[] __JavaGI_args) {\n" ++
             "    javagi.runtime.CoreRT.init(" ++ implArray impls ++ ");\n" ++
             "    new Main().main();\n" ++
             "  }\n" ++
             "}\n"
      implArray impls = 
          "new javagi.runtime.Implementation[]{" ++
          intercalate "," (map implArrayElement impls) ++ "}"
      implArrayElement impl = 
          "new javagi.runtime.Implementation(" ++ 
          (implIfaceName (impl_iface impl)) ++ ".class, " ++
          "new Class<?>[]{" ++ 
          intercalate "," (map (\s -> transClassNameStr s ++ ".class") 
                               (impl_classes impl)) ++
          "}, new " ++ implClass (impl_iface impl) (impl_classes impl) ++ "())"
      headerComment = "// Generated by core-javagic. DO NOT EDIT!!"
      checkWriteOk = 
          withFile out ReadMode (\h -> do s <- hGetLine h
                                          return (s == headerComment))
          `catch` (\e -> if (isDoesNotExistError e) || (isEOFError e)
                            then return True
                            else ioError e)