{

{-# OPTIONS -fno-warn-name-shadowing     #-}
{-# OPTIONS -fno-warn-missing-signatures #-}
{-# OPTIONS -fno-warn-unused-binds       #-}
{-# OPTIONS -fno-warn-unused-matches     #-}

-- ^ don't want to see all the warns alex templates produce

module Lexer ( 
    scan, scanFile,
    Token(..), Tkn(..), Posn(..)
) where

import Logging

}

$digit      = [0-9]
$letter     = [a-z A-Z _]
$special    = [\(\)\,\;\[\]\{\}\.=]

@id         = $letter ($letter|$digit)*
@id_tyarg   = @id "<"
@decimal    = $digit+
@string     = \" [^\"]* \"
@keyword    = class | interface | implementation |
              extends | implements | mono | where |
              static | void | receiver | new | return |
              if | else | while | true | false | null |
              int | bool
@reservedop = "+" | "-" | "*" | "/" | "%" |
              "&&" | "||" | "!" | 
              "<=" | "<" | ">=" | ">" |
              "==" | "!="
@comment1   = "//" [^\n]*
@comment2   =   "/*" ([^\*] | \* [^\/] | \n )*  "*/"
@ignore     = $white+ | @comment1 | @comment2

javagi :-
  @ignore                     ;
  $special                    { \p s -> T p (SpecialT (head s)) }
  @keyword                    { \p s -> T p (KeywordT s) }
  @reservedop                 { \p s -> T p (OpT s) }
  @decimal                    { \p s -> T p (NumT (read s)) }
  @string                     { \p s -> T p (StringT (stripQuotes s)) }
  @id_tyarg                   { handleIdTyarg }
  @id                         { \p s -> T p (VarT s) }

{

data Token = T Posn Tkn | EOF
    deriving (Show)

data Tkn
    = VarT        String
    | VarTyargT   String
    | KeywordT    String
    | SpecialT    Char
    | OpT         String
    | NumT        Int
    | StringT     String
    deriving (Show)

data Posn = Posn !FilePath  -- file name
                 !Int       -- line number
                 !Int       -- column number

instance Show Posn where
    showsPrec _ (Posn f l c) = showString f . showString ":" . 
                               shows l . showString ":" . shows c

mkPosn :: FilePath -> AlexPosn -> Posn
mkPosn fname (AlexPn _ l c) = Posn fname l c

stripQuotes :: String -> String
stripQuotes [] = []
stripQuotes (x:xs) = stripLast xs

stripLast :: String -> String
stripLast (_:[]) = []
stripLast (x:xs) = x : stripLast xs
stripLast []     = []

handleIdTyarg p s =
    let s' = stripLast s
    in if s' == "implementation" 
          then T p (KeywordT s)
          else T p (VarTyargT s')

startPosn :: AlexPosn
startPosn = AlexPn 0 1 0

scan :: FilePath -> String -> [Token]
scan fname str = go (startPosn, '\n', str)
    where 
      go inp@(pos,_,str) = 
           case alexScan inp 0 of                         
             AlexEOF -> [EOF]
             AlexError (p,_,s) -> 
                 fatal ("Lexer error: " ++ 
                        show (mkPosn fname p) ++ ", next character: `" ++ 
                              [head s] ++ "'")
             AlexSkip  inp' len     -> go inp'        
             AlexToken inp' len act -> 
               act (mkPosn fname pos) (take len str) : go inp'

scanFile :: FilePath -> IO [Token]
scanFile fname = 
    do s <- readFile fname
       return (scan fname s)
    
-- We cannot use the "posn" wrapper distributed with Alex because
-- this wrapper does not provide filenames in positions. Instead,
-- we slightly modified the code of the "posn" wrapper and copied
-- it into our lexer. (In fact, we only removed the function
-- "alexScanTokens")

-- Start code copied from "AlexWrapper-posn"
-- -----------------------------------------------------------------------------
-- Alex wrapper code.
--
-- This code is in the PUBLIC DOMAIN; you may copy it freely and use
-- it for any purpose whatsoever.

-- -----------------------------------------------------------------------------
-- The input type


type AlexInput = (AlexPosn, 	-- current position,
		  Char,		-- previous char
		  String)	-- current input string

alexInputPrevChar :: AlexInput -> Char
alexInputPrevChar (p,c,s) = c

alexGetChar :: AlexInput -> Maybe (Char,AlexInput)
alexGetChar (p,c,[]) = Nothing
alexGetChar (p,_,(c:s))  = let p' = alexMove p c in p' `seq`
				Just (c, (p', c, s))

-- -----------------------------------------------------------------------------
-- Token positions

-- `Posn' records the location of a token in the input text.  It has three
-- fields: the address (number of chacaters preceding the token), line number
-- and column of a token within the file. `start_pos' gives the position of the
-- start of the file and `eof_pos' a standard encoding for the end of file.
-- `move_pos' calculates the new position after traversing a given character,
-- assuming the usual eight character tab stops.

data AlexPosn = AlexPn !Int !Int !Int
	deriving (Eq,Show)

alexStartPos :: AlexPosn
alexStartPos = AlexPn 0 1 1

alexMove :: AlexPosn -> Char -> AlexPosn
alexMove (AlexPn a l c) '\t' = AlexPn (a+1)  l     (((c+7) `div` 8)*8+1)
alexMove (AlexPn a l c) '\n' = AlexPn (a+1) (l+1)   1
alexMove (AlexPn a l c) _    = AlexPn (a+1)  l     (c+1)

-- End code copied from "AlexWrapper-posn"

}
