module Logging (

  LogLevel(..), LogTarget(..),

  initLogging, enterPhase, leavePhase,

  debug, info, warn, fatal, bug

) where

import Control.Monad
import Control.Monad.Trans
-- import System.Time
import System.IO
import System.IO.Unsafe
import Data.IORef
import System.Exit
-- import Text.Printf

data LogLevel = Debug | Info | Warn | Fatal deriving (Eq,Ord,Enum,Bounded)

instance Show LogLevel where
    show Debug = "[DEBUG]"
    show Info  = "[INFO] "
    show Warn  = "[WARN] "
    show Fatal = "[FATAL]"

data LogTarget = File FilePath
               | Stderr
               deriving (Eq,Show)

data LogState = LogState { l_level  :: LogLevel
                         , l_target :: LogTargetIntern
                         , l_phases :: [String] }

data LogTargetIntern = IFile Handle
                     | IStderr
                       
logState :: IORef LogState
logState = unsafePerformIO $ do newIORef (LogState Info IStderr [])
{-# NOINLINE logState #-}

initLogging :: MonadIO m => LogLevel -> LogTarget -> m ()
initLogging level target = liftIO $
    do target' <- case target of
                    File fp ->
                        do h <- openFile fp AppendMode
                           return $ IFile h
                    Stderr -> return IStderr
       writeIORef logState (LogState level target' [])

enterPhase :: MonadIO m => String -> m ()
enterPhase name = liftIO $ 
    do info ("ENTERING PHASE " ++ name)
       s <- readIORef logState
       writeIORef logState (s { l_phases = name : l_phases s })

leavePhase :: MonadIO m => m ()
leavePhase = liftIO $
    do s <- readIORef logState
       if null (l_phases s)
          then warn "leavePhase called although no phase active!"
          else do info ("LEAVING PHASE " ++ head (l_phases s))
                  writeIORef logState (s { l_phases = tail (l_phases s) })

doLog :: MonadIO m => LogLevel -> String -> m ()
doLog level msg = liftIO $
    do ls <- readIORef logState
       let curLevel = l_level ls
       when (level >= curLevel) $ do
         pref <- standardPrefix (l_phases ls) level
         let msg' = pref ++ msg
             h = case l_target ls of
                   IStderr -> stderr
                   IFile h' -> h'
         hPutStrLn h msg'
         hFlush h

{-
formatTime :: IO String
formatTime =
    do clockT <- getClockTime
       CalendarTime year mon day hour minutes sec picoSec
           _ _ _ _ _ <- toCalendarTime clockT
       let msec = picoSec `div` 1000000000
       return (printf "%4d-%02d-%02d %02d:%02d:%02d:%03d"
                      year (fromEnum mon + 1) day hour minutes sec msec)

standardPrefix :: [String] -> LogLevel -> IO String
standardPrefix phases ll =
    do time <- formatTime
       let phase = 
               case phases of
                 [] -> ""
                 (x:_) -> x ++ " "
       return ("[" ++ show ll ++ " " ++ phase ++ time ++ "] ")
-}

standardPrefix :: [String] -> LogLevel -> IO String
standardPrefix _ ll = 
    return (show ll ++ " ")

debug,info,warn :: MonadIO m => String -> m ()
debug msg = doLog Debug msg
info msg = doLog Info msg
warn msg = doLog Warn msg

fatal :: String -> a
fatal msg = unsafePerformIO $
    do hPutStrLn stderr msg
       exitWith (ExitFailure 1)

bug :: String -> a
bug msg = fatal ("BUG!!\n" ++ msg)