import System.Console.GetOpt
import System.Exit
import System.Environment
import Control.Monad

import AST
import Parser
import Logging
import Lexer
import Pretty
import Symtab
import GIMonad
import ProgramTyping
import Execution

import AllTests

version :: String
version = "development version"

data Flag   
    = Version
    | Help 
    | LogLevel LogLevel
    | LogFile FilePath
    | Dump Phase
    | DontRun
    | RunTests
  deriving Eq

data Phase 
    = Lex
    | Parse
    | RawParse
  deriving Eq

programName :: String
programName = "core-javagi"

-- ---------------------------------------------------------------------
-- Handling command line flags

--
-- associate concrete flags with their abstract representation
--
opts :: [OptDescr Flag]
opts = 
    [ Option ['v'] ["version"] (NoArg Version) "version string"
    , Option ['?'] ["help"] (NoArg Help) "this message"
    , Option []    ["debug"] (NoArg (LogLevel Debug)) "enable debug messages"
    , Option []    ["quiet"] (NoArg (LogLevel Warn)) 
                   "display only warnings and errors"
    , Option []    ["tests"] (NoArg RunTests) "run test cases"
    , Option []    ["log-file"] (ReqArg LogFile "FILE") "log file"
    , Option []    ["dont-run"] (NoArg DontRun) 
            "don't run the program"
    , Option []    ["dump-tokens"]   (NoArg (Dump Lex)) 
            "dump lexed tokens"
    , Option []    ["dump-ast"]  (NoArg (Dump Parse)) 
            "print abstract syntax tree"
    , Option []    ["dump-rawast"]   (NoArg (Dump RawParse)) 
            "dump raw data type of abstract syntax tree"
    ]


--
-- parse the args.
--
-- we also short circuit here, for the known cases of --version or --help
--
parseArgs :: [String] -> IO ([Flag], [FilePath])
parseArgs argv = 
    case (getOpt Permute opts argv) of
      (flags,fs,[]) -> 
          case () of 
            _| Version `elem` flags -> putStrLn version >> exitWith ExitSuccess
             | Help `elem` flags -> putStrLn usage   >> exitWith ExitSuccess
             | otherwise         -> return (flags, fs)
      (_,_,err) -> fatal $ concat err ++ usage

usage :: String
usage = usageInfo ("Usage: " ++ programName  ++ " [OPTION...] FILES") opts


logLevel :: [Flag] -> LogLevel
logLevel [] = Info
logLevel (LogLevel l:_) = l
logLevel (_:rest) = logLevel rest

logTarget :: [Flag] -> LogTarget
logTarget [] = Stderr
logTarget (LogFile f:_) = File f
logTarget (_:rest) = logTarget rest

parse :: [Flag] -> FilePath -> IO Program
parse flags f = 
    do toks <- scanFile f
       when (Dump Lex `elem` flags) $
            info ("Tokens from " ++ f ++ ":\n" ++ show toks)
       prog <- parseProgram toks
       when (Dump Parse `elem` flags) $
            info ("AST from " ++ f ++ ":\n" ++ showPP prog)
       when (Dump RawParse `elem` flags) $
            info ("Raw AST from " ++ f ++ ":\n" ++ show prog)
       return prog

mergeProgs :: [Program] -> Program
mergeProgs l = Program (concat (map unProgram l))

main :: IO ()
main =
    do args <- getArgs
       (flags,files) <- parseArgs args
       initLogging (logLevel flags) (logTarget flags)
       if RunTests `elem` flags
          then AllTests.run
          else if null files
                  then fatal usage
                  else do enterPhase "lex+parse"
                          progs <- mapM (parse flags) files
                          let prog = mergeProgs progs
                              symtab = buildSymtab prog
                          leavePhase
                          enterPhase "typechecking"
                          tycheckRes <- withSymtab symtab (assertOKProg prog)
                          case tycheckRes of
                            Right _ -> info "program type checks!"
                            Left s ->
                                fatal ("Program is ill-typed!\n" ++ s)
                          leavePhase
                          enterPhase "execution"
                          execRes <- withSymtab symtab execProgram
                          case execRes of
                            Right _ -> return ()
                            Left s ->
                                fatal ("Runtime error!\n" ++ s)
                          leavePhase
                          return ()
