{-# LANGUAGE PatternSignatures #-}

module MethodDefinitions ( getMDef, getSMDef, Method(..) ) where

import Data.Maybe
import Data.List

import AST
import GIMonad
import Inheritance
import Types
import Logging
import Pretty
import Utils
import Subtyping
import Environment
import Builtins

data Method = Builtin BuiltinId | NonBuiltin MethodDefG

getMDefClass :: MethodName -> TyN -> GI Method
getMDefClass m n@(TyClass c us) =
    do cls <- findClass c
       case builtinId c m of
         Just id -> return (Builtin id)
         Nothing ->
             case find (\md -> mdefName md == m) (class_methods cls) of
               Nothing ->
                   do super <- upOneClass n
                      case super of
                        Nothing -> failGI ("Runtime error: method " ++ m ++ 
                                           " not found in class " ++ c)
                        Just n' -> getMDefClass m n'
               Just mdef ->
                   let subst = mkSubst (class_tyargs cls) us
                   in do debug ("Found definition of " ++ m ++ " in class " ++
                                c)
                         return (NonBuiltin $ apply subst mdef)

getMDefIface :: MethodName -> TyN -> [Maybe TyN] -> GI Method
getMDefIface m n ns =
    do res <- findMSig m
       case res of
         Nothing -> 
             failGI ("Runtime error: interface method " ++ m ++ " not found")
         Just (iface,msig,j) ->
             do debug ("getMDefIface, found method " ++ m ++ " in interface " ++ 
                       iface_name iface ++ " at dispatch position " ++ show j)
                let zs = iface_impl_types iface
                    ts = map (fst . unMethodArg) (ms_args msig)
                    zj = zs !! (j-1)
                    mQuest i = if i /= j then contrib m (zs!!(i-1)) ts ns
                                         else contrib m zj (TyG (TyVar zj) : ts) 
                                                      (Just n:ns)
                impls <- findAllImpls (iface_name iface)
                minimplArgs <- mapMaybeM (matchImpl mQuest) impls
                debug ("getMDefIface, minimplArgs = " ++ showPP minimplArgs)
                (subst, impl) <- minimpl minimplArgs
                let rcdef = impl_receiver_defs impl !! (j-1)
                    mdef = case find (\md -> mdefName md == m) (rcdef_defs rcdef) of
                             Just md -> md
                             Nothing -> bug ("Execution.getMDefIface: Implementation "
                                             ++ showPP impl ++
                                             " does not define method " ++ m)
                debug ("getMDefIface, Found definition of " ++ m ++ 
                       " with receiver " ++
                       showPPQ n ++ " and arguments " ++ showPPQ ns ++ 
                       " in implementation " ++
                       showPPImplHeader (impl_iface impl) (impl_impl_types impl))
                return (NonBuiltin (apply subst mdef))
    where
      matchImpl :: (Int -> GI (Maybe TyN)) -> Implementation 
                -> GI (Maybe (TySubst, Implementation))
      matchImpl mQuest impl@(Implementation xs _ ms _ _ _ _) =
          do let l = length ms
             pairs <- mapMaybeM (\i -> do mm <- mQuest i
                                          debug ("getMDefIface, mQuest " ++ show i
                                                 ++ " = " ++
                                                 showPP mm)
                                          return $ do m <- mm
                                                      return (TyG (TyN m), 
                                                              TyG (TyN (ms!!(i-1)))))
                                [1..l]
             msubst <- unifyModSub emptyTEnv xs pairs
             return $ do subst <- msubst
                         return (subst, impl)

minimpl :: [(TySubst, Implementation)] -> GI (TySubst, Implementation)
minimpl [] = bug ("Execution.minimpl: empty argument list")
minimpl l =
    let nss :: [[TyN]] = map (\ (subst,impl) -> apply subst (impl_impl_types impl)) l
    in do minIndex <- findMin nss (zip [0..] nss)
          return (l !! minIndex)
    where
      findMin _ [] = bug ("Execution.minImpl.findMin: no minimum found")
      findMin all ((i,ns):rest) =
          do b <- allM (\ns' -> ns `extendsClassMulti` ns') all
             if b then return i else findMin all rest
               
extendsClassMulti :: [TyN] -> [TyN] -> GI Bool
extendsClassMulti ns1 ns2 = 
    allM (\ (n1,n2) -> n1 `extendsClass` n2) (zip ns1 ns2)

contrib :: MethodName -> TyvarName -> [Type] -> [Maybe TyN] -> GI (Maybe TyN)
contrib m x ts ns = 
    let n = length ts
        cands = [ ns !! (i-1) | i <- [1..n], ts !! (i-1) == TyG (TyVar x) ]
    in do debug ("contrib, x = " ++ showPP x ++ ", ts = " ++ showPPQ ts ++ 
                 ", ns = " ++
                 showPPQ ns ++ ", cands = " ++ showPPQ cands)
          if null cands then do debug ("contrib, result: nil")
                                return Nothing 
             else if Nothing `elem` cands 
                     then failGI ("NullPointerException on call of " ++ m)
                     else do n <- lubs (map fromJust cands)
                             debug ("contrib, result: " ++ showPPQ n)
                             return (Just n)

lubs :: [TyN] -> GI TyN
lubs [] = bug ("Execution.lubs: empty argument list")
lubs [n] = return n
lubs (n:ns) =
    do m <- lubs ns
       n `lub` m

lub :: TyN -> TyN -> GI TyN
lub n m =
    do mres <- firstJustM 
                 [do b <- n `extendsClass` m
                     if b then return (Just m) else return Nothing
                 ,do b <- m `extendsClass` n
                     if b then return (Just n) else return Nothing
                 ,do mn' <- upOneClass n
                     case mn' of
                       Nothing -> return Nothing
                       Just n' ->
                           do res <- n' `lub` m
                              return (Just res)]
       case mres of
         Just res -> return res
         Nothing -> bug ("Execution.lub: failed on arguments " ++ showPPQ n ++ 
                         " and " ++ showPPQ m)

getMDef :: MethodName -> Maybe TyN -> [Maybe TyN] -> GI Method
getMDef m Nothing _ = failGI ("NullPointerException on call of " ++ m)
getMDef m (Just n) ns =
    chooseGIs [getMDefClass m n,
               getMDefIface m n ns,
               bug ("Execution.getMDef: no method " ++ m ++ " found for receiver "
                    ++ showPPQ n ++ " and arguments " ++ showPPQ ns)]

getSMDef :: MethodName -> TyK -> [Type] -> GI MethodDefG
getSMDef m (TyIface i _) ws =
    do impls <- findAllImpls i
       minimplArg <- mapMaybeM matchImpl impls
       (subst, impl) <- minimpl minimplArg
       let mdef = case find (\md -> mdefName md == m) (impl_static_defs impl) of
                    Just md -> md
                    Nothing -> bug ("Execution.getSMDefIface: Implementation "
                                    ++ showPP impl ++
                                    " does not define static method " ++ m)
       return (apply subst mdef)
    where
      matchImpl impl@(Implementation xs _ ms _ _ _ _) =
          do let l = length ms
                 pairs = mapMaybe (\i -> let m = ms !! (i-1)
                                         in if m == tyObjectN then Nothing
                                               else case ws !! (i-1) of
                                                      TyG (TyN n) -> Just (n,m)
                                                      _ -> Nothing)
                                  [1..l]
             msubst <- unifyModSub emptyTEnv xs (map (\ (n,m) -> (TyG (TyN n),
                                                                  TyG (TyN m))) 
                                                     pairs)
             return $ do subst <- msubst
                         return (subst, impl)