module ParserUtils where

import Control.Monad.State
import Data.List (sort)
import qualified Data.Set as Set
import Data.Generics as Gen

import qualified NameEnv as NE
import Logging
import AST
import Lexer

type TermEnv = NE.T String VarName

type TypeEnv = NE.T String TyvarName

data ParseState = ParseState { termEnv :: TermEnv,
                               typeEnv :: TypeEnv,
                               curClassName :: Maybe String,
                               implId :: Int }

mkParseState :: [String] -> [String] -> ParseState
mkParseState vars tyvars = 
    let empty = emptyParseState
    in empty { termEnv = NE.push vars [0..] (termEnv empty),
               typeEnv = NE.push tyvars [0..] (typeEnv empty)
             }

emptyParseState :: ParseState
emptyParseState =
    let e1 = NE.empty
        e2 = NE.empty
        in ParseState e1 e2 Nothing 0

type PM = StateT ParseState IO

getTermEnv :: PM TermEnv
getTermEnv = gets termEnv

getTypeEnv :: PM TypeEnv
getTypeEnv = gets typeEnv

setTermEnv :: TermEnv -> PM ()
setTermEnv env = modify (\s -> s { termEnv = env } )

setTypeEnv :: TypeEnv -> PM ()
setTypeEnv env = modify (\s -> s { typeEnv = env } )

freshImplId :: PM ImplId
freshImplId = 
    do state <- get
       let i = implId state
       put (state { implId = i+1 })
       return i

stringOfVar :: Token -> String
stringOfVar (T _ t) = stringOfVar' t
stringOfVar EOF = 
    bug ("ParserUtils.stringOfVar called with illegal argument EOF")

stringOfVar' :: Tkn -> String
stringOfVar' (VarT s) = s
stringOfVar' (VarTyargT s) = s
stringOfVar' x = bug ("ParserUtils.stringOfVar' called with illegal argument: " 
                      ++ show x)

lookupTHIS :: PM TyvarName
lookupTHIS =
    do env <- getTypeEnv
       case NE.lookup "This" env of
         Nothing -> 
           bug ("variable \"This\" not found in type environment during parsing")
         Just x -> return x

lookupThis :: PM VarName
lookupThis =
    do env <- getTermEnv
       case NE.lookup "this" env of
         Nothing -> 
           bug ("variable \"this\" not found in term environment during parsing")
         Just x -> return x

handleExprVar :: String -> PM Expr
handleExprVar s =
    do env <- getTermEnv
       case NE.lookup s env of
         Nothing -> 
             do this <- lookupThis
                t <- newRef
                return $ FAccess (Var this) s t
         Just x -> return $ (Var x)
       
handleType :: Token -> [Type] -> PM Type
handleType EOF _ = bug ("ParserUtils.handleType called with EOF")
handleType (T p t) l =
    do let s = stringOfVar' t
       env <- getTypeEnv
       case NE.lookup s env of
         Nothing -> return $ TyParse (TyClassIface s l)
         Just x ->
             return $ if null l 
                         then TyG (TyVar x)
                         else fatal ("Parse error: type variable " ++ show s ++ 
                                     " at " ++ show p ++ 
                                     " does not take arguments")

handleTypeVar :: Token -> PM TyvarName
handleTypeVar EOF = bug "ParserUtils.handleTypeVar called with EOF"
handleTypeVar (T p t) =
    do let s = stringOfVar' t
       env <- getTypeEnv
       case NE.lookup s env of
         Nothing -> fatal ("Parse error: type variable expected at " ++ show p
                           ++ ", got: ``" ++ show t ++ "''")
         Just x -> return x

handleAssign :: String -> Expr -> PM Stmt
handleAssign varName expr =
    do env <- getTermEnv
       case NE.lookup varName env of
         Nothing -> 
             do this <- lookupThis
                r <- newRef
                return $ AssignField (Var this) varName expr r
         Just x -> do r <- newRef
                      return $ AssignLocal x expr r

handleMCall :: String -> [Type] -> [Expr] -> PM Expr
handleMCall name ts es = 
    do this <- lookupThis
       r <- newRef
       case name of
         "print" -> newUtils >>= \e -> return (MCall e name ts es r)
         "printInt" -> newUtils >>= \e -> return (MCall e name ts es r)
         _ -> return (MCall (Var this) name ts es r)
    where newUtils = 
              do r <- newRef
                 return (New (TyClass "Utils" []) [] r)

asIfaceType :: TyNK -> TyK
asIfaceType t = TyIface (tyci_name t) (tyci_args t)

asClassType :: TyNK -> TyN
asClassType t = TyClass (tyci_name t) (tyci_args t)

asTyG :: Posn -> Type -> TyG
asTyG _ (TyG g) = g
asTyG _ TyBot = bug ("TyBot detected during parsing")
asTyG p (TyPrim t) = fatal ("Parse error: expected G-type at " ++ show p
                               ++ ", got primitive type " ++ show t)
asTyG p (TyK i) = fatal ("Parse error: expected G-type at " ++ show p
                            ++ ", got interface type " ++ show i)
asTyG _ (TyParse t) = TyN (asClassType t)

firstUnused :: (Ord b, Num b) => NE.T a b -> b
firstUnused env =
    case reverse $ sort (NE.elems env) of
      [] -> 0
      x:_ -> x+1

defineParams :: (Num a, Ord a, Enum a) => [String] -> PM (NE.T String a) 
             -> (NE.T String a -> PM ()) -> PM [a]
defineParams names getFun setFun =
    do env <- getFun
       let n = firstUnused env
           vars = take (length names) [n..]
           env' = NE.push names vars env
       setFun env'
       return vars

undefineParams :: PM (NE.T k v) -> (NE.T k v -> PM a) -> PM a
undefineParams getFun setFun =
    do env <- getFun
       let env' = NE.pop env
       setFun env'

defineTypeParams :: String -> [String] -> PM [TyvarName]
defineTypeParams x names = 
    do res <- defineParams names getTypeEnv setTypeEnv
       debug ("Defining new slot for type parameters " ++ show names ++ 
              " of " ++ x ++ " as " ++ show res)
       return res

undefineTypeParams :: String -> PM ()
undefineTypeParams x = 
    do debug ("Undefining most recent slot of type parameters of " ++ x)
       undefineParams getTypeEnv setTypeEnv

defineExprParams :: String -> [String] -> PM [VarName]
defineExprParams x names = 
    do res <- defineParams names getTermEnv setTermEnv
       debug ("Defining new slot for expression parameters " ++ show names ++ 
              " of " ++ x ++ " as " ++ show res)
       return res

undefineExprParams :: String -> PM ()
undefineExprParams x = 
    do debug ("Undefining most recent slot of expression parameters of " ++ x)
       undefineParams getTermEnv setTermEnv

defineCurrentClassName :: String -> PM ()
defineCurrentClassName name = 
    modify (\s -> s { curClassName = Just name })
    
undefineCurrentClassName :: PM ()
undefineCurrentClassName =
    modify (\s -> s { curClassName = Nothing })

getCurrentClassName :: PM String
getCurrentClassName = 
    do x <- gets curClassName
       case x of
         Just name -> return name
         Nothing -> 
           bug ("name of current class not available but nevertheless accessed")

handleClassDef :: Token -> (String, [TyvarName]) -> TyN -> [ConP] 
            -> [Either FieldDef (Either Constructor MethodDef)] -> PM Class
handleClassDef tok (name, tyargs) super cons fcm =
    let (fdefs, x) = splitEither fcm
        (constrs, mdefs) = splitEither x
        constr = case constrs of
                   [] -> Nothing
                   [c] -> Just c
                   _ -> fatal ("Parse error: class at " ++ show (posn tok) ++
                               " has more than one constructor")
    in do undefineTypeParams "class"
          undefineCurrentClassName
          return (Class name tyargs (Just super) cons constr fdefs mdefs)

handleInterfaceDef :: (String, [TyvarName]) -> ([TyvarName], [ConR])
                   -> [ConP] -> [MethodSig] -> Maybe [ReceiverSig] 
                   -> PM Interface
handleInterfaceDef (name, tyargs) (implTypes, implCons) cons staticSigs msigs =
    do undefineTypeParams "interface"
       undefineTypeParams "interface"
       case msigs of
         Just sigs ->
             return (Interface name tyargs implTypes implCons cons 
                               staticSigs sigs)
         Nothing ->
             let sigs = map (\x -> ReceiverSig x []) implTypes
             in return (Interface name tyargs implTypes implCons cons 
                                  staticSigs sigs)

handleLocalDef :: Type -> String -> Expr -> PM Stmt
handleLocalDef ty name e = 
    do env <- getTermEnv
       let n = firstUnused env
       debug ("Defining local variable " ++ show name ++ " as " ++ show n)
       setTermEnv (NE.insert name n env)
       r <- newRef
       return $ DefineLocal ty n e r

posn :: Token -> Posn
posn EOF = bug ("ParserUtils.posn called with EOF")
posn (T p _) = p

splitEither :: [Either a b] -> ([a], [b])
splitEither l =
    foldr f ([],[]) l
    where f :: Either a b -> ([a], [b]) -> ([a], [b])
          f (Left x) (l1,l2) = (x:l1, l2)
          f (Right x) (l1,l2) = (l1, x:l2)

parseError :: [Token] -> PM a
parseError []          = fatal ("Parse error: unexpected end of file")
parseError (EOF : _)   = fatal ("Parse error: unexpected end of file")
parseError (T p t : _) = fatal ("Parse error: unexpected token ``" ++ show t ++
                                "'' at " ++ show p)

fixAST :: Program -> Program
fixAST prog = fixAST' prog prog

fixAST' :: (Data a, Typeable a) => Program -> a -> a
fixAST' (Program defs) x = Gen.everywhere (Gen.mkT fixType) x
    where add (ImplDef _) x = x
          add (IfaceDef i) (s1, s2) = (s1, Set.insert (iface_name i) s2)
          add (ClassDef c) (s1, s2) = (Set.insert (class_name c) s1, s2)
          (classes, ifaces) = foldr add (Set.empty, Set.empty) defs
          fixType :: Type -> Type
          fixType (TyParse (TyClassIface name args)) =
              if name `Set.member` classes
                 then TyG (TyN (TyClass name args))
                 else if name `Set.member` ifaces
                         then TyK (TyIface name args)
                         else fatal ("Parse error: class or interface " ++ 
                                     show name ++ " not found")
          fixType t = t

genericParse :: (Typeable a, Data a) =>
                ([Token] -> PM a) -> [Token] -> [String] -> Program -> IO a
genericParse f toks tyvars prog =
    do let state = mkParseState [] tyvars
       res <- runStateT (f toks) state
       let x = fst res
       return (fixAST' prog x)
