{-# LANGUAGE FlexibleInstances #-}

module Pretty where

import Text.PrettyPrint hiding (render)
import qualified Text.PrettyPrint as PP

import AST
import Logging
import Utils

class Pretty a where
    pp :: a -> Doc

varAsString :: VarName -> String
varAsString (VarName 0) = "this"
varAsString (VarName i) = "x_" ++ show i

tyvarAsString :: TyvarName -> String
tyvarAsString (TyvarName i) = 
    if i >= 0 then "Y_" ++ show i else "Z_" ++ tail (show i)

dot :: Doc
dot = char '.'

langle, rangle :: Doc
langle = char '<'
rangle = char '>'

angles :: Doc -> Doc
angles d = langle <> d <> rangle


infixr 5 <.>,<=>

(<.>), (<=>) :: Doc -> Doc -> Doc
d1 <.> d2 = d1 <> dot <> d2
d1 <=> d2 = d1 <+> char '=' <+> d2

indent :: Int
indent = 4

instance Pretty Doc where
    pp d = d

instance Pretty VarName where
    pp x = text (varAsString x)

instance Pretty TyvarName where
    pp x = text (tyvarAsString x)

instance (Pretty a, Pretty b) => Pretty (a,b) where
    pp (a,b) = parens (pp a <> comma <+> pp b)

instance Pretty a => Pretty (Maybe a) where
    pp Nothing = text "Nothing"
    pp (Just x) = text "Just" <+> parens (pp x)

instance Pretty Int where
    pp i = text (show i)
    
instance Pretty TyPrim where
    pp TyInt = text "int"
    pp TyBool = text "bool"

instance Pretty TyG where
    pp (TyN n) = pp n
    pp (TyVar i) = pp i

ppTyargs :: [Type] -> Doc
ppTyargs [] = empty
ppTyargs l = angles (pp l)

instance Pretty TyN where
    pp n = text (tyc_name n) <> ppTyargs (tyc_args n)

instance Pretty TyK where
    pp n = text (tyi_name n) <> ppTyargs (tyi_args n)

instance Pretty TyNK where
    pp n = text (tyci_name n) <> ppTyargs (tyci_args n)

instance Pretty Type where
    pp (TyG g) = pp g
    pp (TyPrim p) = pp p
    pp (TyK k) = pp k
    pp (TyParse nk) = pp nk
    pp TyBot = text "BOT"

instance Pretty ConR where
    pp (Impl gs k) = 
        (case gs of
          [g] -> pp g
          _ -> brackets (pp gs))
        <+> text "implements" <+> pp k

instance Pretty ConGR where
    pp (ImplG ts k) =
        (case ts of
          [t] -> pp t
          _ -> brackets (pp ts))
        <+> text "implements" <+> pp k

instance Pretty ConP where
    pp (ConR c) = pp c
    pp (Ext x t) = pp x <+> text "extends" <+> pp t
    pp (Mono x) = pp x <+> text "mono"

instance Pretty ConGP where
    pp (ConGR c) = pp c
    pp (ExtG t u) = pp t <+> text "extends" <+> pp u
    pp (MonoG t) = pp t <+> text "mono"

instance Pretty ConNil where
    pp (ImplNil tsNil i vsNil) = 
        (case tsNil of
          [tNil] -> pp tNil
          _ -> brackets (pp tsNil))
        <+> text "implements" <+> text i <>
        (case vsNil of
           [] -> empty
           _ -> angles (pp vsNil))

instance Pretty Expr where
    pp (Var i) = pp i
    pp (FAccess e f _) = pp e <> dot <> text f
    pp (MCall e m tyargs args _) = pp e <> dot <> text m <> ppTyargs tyargs <>
                                   parens (pp args)
    pp (SMCall k implTypes m tyargs args _) =
        pp k <> brackets (pp implTypes) <> dot <> text m <> ppTyargs tyargs <>
        parens (pp args)
    pp (New n args _) = parens $ text "new" <+> pp n <> parens (pp args)
    pp (Cast n e) = parens $ parens (pp n) <+> pp e
    pp e@(PrimOp op args) =
        case (op, args) of
          (UnMinus, [e']) -> parens (char '-' <> pp e')
          (Not, [e']) -> parens (char '!' <> pp e')
          (_, [e1,e2]) -> parens (pp e1 <+> pp op <+> pp e2)
          _ -> bug ("invalid use of primitive operator: " ++ show e)
    pp (Lit l) = pp l

instance Pretty Lit where
    pp (LitInt i) = text (show i)
    pp (LitBool True) = text "true"
    pp (LitBool False) = text "false"
    pp (LitString s) = text (show s)
    pp LitNull = text "null"

instance Pretty Op where
    pp op = text (case op of
                    UnMinus -> "-"
                    Add -> "+"
                    Sub -> "-"
                    Mult -> "*"
                    Div -> "/"
                    Mod -> "%"
                    And -> "&&"
                    Or -> "||"
                    Not -> "!"
                    Leq -> "<="
                    Lt -> "<"
                    Geq -> ">="
                    Gt -> ">"
                    Equal -> "=="
                    NotEqual -> "!=")

instance Pretty a => Pretty [a] where
    pp l = hsep $ punctuate comma (map pp l)

ppListNewline :: Pretty a => [a] -> Doc
ppListNewline l = foldr ($+$) empty (map pp l)

instance Pretty Stmt where
    pp (StmtExpr e) = pp e <> semi
    pp (DefineLocal t x e _) = pp t <+> pp x <+> equals <+> pp e <> semi
    pp (AssignLocal x e _) = pp x <+> equals <+> pp e <> semi
    pp (AssignField e f e' _) = 
        pp e <> dot <> text f <+> equals <+> pp e' <> semi
    pp (If e l) = text "if" <+> pp e <+> lbrace 
                  $+$ nest indent (ppListNewline l)
                  $+$ rbrace
    pp (IfElse e l1 l2) = text "if" <+> pp e <+> lbrace 
                          $+$ nest indent (ppListNewline l1)
                          $+$ rbrace <+> text "else" <+> lbrace
                          $+$ nest indent (ppListNewline l2)
                          $+$ rbrace
    pp (While e l) = text "while" <+> pp e <+> lbrace 
                     $+$ nest indent (ppListNewline l)
                     $+$ rbrace
    pp (Return e _) = text "return" <+> pp e <> semi
    pp Skip = semi

instance Pretty FieldDef where
    pp f = pp (fd_type f) <+> text (fd_name f) <> semi

instance Pretty MethodArg where
    pp (MethodArg (t,x)) = pp t <+> pp x

ppConstrClause :: Pretty a => [a] -> Doc
ppConstrClause [] = empty
ppConstrClause l = text " where" <+> pp l

ppTyvarList :: [TyvarName] -> Doc
ppTyvarList l = hsep $ punctuate comma (map pp l)

ppTyparamList :: [TyvarName] -> Doc
ppTyparamList [] = empty
ppTyparamList l = angles (ppTyvarList l)

instance Pretty a => Pretty (MethodSigGeneric a) where
    pp m = ppTyparamList (ms_tyargs m)
           <+>
           (case ms_return m of
              Nothing -> text "void"
              Just t -> pp t)
           <+>
           text (ms_name m)
           <>
           parens (pp (ms_args m))
           <>
           ppConstrClause (ms_cons m)

instance Pretty a => Pretty (MethodDefGeneric a) where
    pp m = pp (md_sig m) <+> lbrace
           $+$ nest indent (ppListNewline (md_body m))
           $+$ rbrace

ppConstructor :: ClassName -> Maybe Constructor -> Doc
ppConstructor _ Nothing = empty
ppConstructor name (Just c) = 
    text name <> parens (pp (con_args c)) <+> lbrace
    $+$ nest indent (ppListNewline (con_body c))
    $+$ rbrace

instance Pretty Class where
    pp c = text "class" <+> text (class_name c) <> 
           ppTyparamList (class_tyargs c)
           <> 
           (case class_super c of
              Nothing -> empty
              Just n -> text " extends" <+> pp n) 
           <> ppConstrClause (class_cons c)
           <+> lbrace
           $+$ nest indent (ppListNewline (class_fields c))
           $+$ nest indent (ppConstructor (class_name c) (class_constructor c))
           $+$ nest indent (ppListNewline (class_methods c))
           $+$ rbrace

ppStaticMsig :: MethodSig -> Doc
ppStaticMsig m = text "static" <+> pp m

instance Pretty Interface where
    pp i = text "interface" <+> text (iface_name i) 
           <> ppTyparamList (iface_tyargs i)
           <+> brackets (ppTyvarList (iface_impl_types i) 
                         <> ppConstrClause (iface_impl_cons i))
           <+> ppConstrClause (iface_cons i)
           <+> lbrace
           $+$ nest indent (foldr ($+$) empty 
                                        (map ppStaticMsig 
                                             (iface_static_sigs i)))
           $+$ nest indent (ppListNewline (iface_receiver_sigs i))
           $+$ rbrace

instance Pretty ReceiverSig where
    pp r = text "receiver" <+> pp (rcsig_name r) <+> lbrace
           $+$ nest indent (ppListNewline (rcsig_sigs r))
           $+$ rbrace

ppStaticMdef :: MethodDef -> Doc
ppStaticMdef m = text "static" <+> pp m

instance Pretty Implementation where
    pp i = text "implementation" <> ppTyparamList (impl_tyargs i)
           <+> pp (impl_iface i) 
           <+> brackets (pp (impl_impl_types i))
           <> ppConstrClause (impl_cons i)
           <+> lbrace
           $+$ nest indent (foldr ($+$) empty 
                                      (map ppStaticMdef (impl_static_defs i)))
           $+$ nest indent (ppListNewline (impl_receiver_defs i))
           $+$ rbrace

instance Pretty ReceiverDef where
    pp r = text "receiver" <+> lbrace
           $+$ nest indent (ppListNewline (rcdef_defs r))
           $+$ rbrace

instance Pretty Def where
    pp (ImplDef i) = pp i
    pp (IfaceDef i) = pp i
    pp (ClassDef c) = pp c

instance Pretty Program where
    pp (Program l) = ppListNewline l

ppList :: Pretty a => [a] -> Doc
ppList l = brackets (pp l)

render :: Doc -> String
render = PP.render

showPP :: Pretty a => a -> String
showPP = render . pp

showPPQ :: Pretty a => a -> String
showPPQ x = quote (showPP x)

showPPList :: Pretty a => [a] -> String
showPPList = render . ppList

ppImplHeader :: TyK -> [TyN] -> Doc
ppImplHeader k ns = pp k <> brackets (pp ns)

showPPImplHeader :: TyK -> [TyN] -> String
showPPImplHeader k ns = render (ppImplHeader k ns)