{-# LANGUAGE PatternSignatures #-}

module ProgramTyping where

import Control.Monad
import Data.List

import AST
import Subtyping
import Types
import GIMonad
import Pretty
import Environment
import Inheritance
import MethodTypes
import Logging
import WellFormedness
import StatementTyping
import Utils
import Restrictions

assertSubMSig :: TEnv -> MethodSig -> MethodSigG -> GI ()
assertSubMSig tenv msig1 msig2 =
    do let xs1 = ms_tyargs msig1
           xs2 = ms_tyargs msig2
       unless (length xs1 == length xs2) $
              failGI ("Type error: method signatures " ++ showPP msig1 ++ " and "
                      ++ showPP msig2 ++ 
                      " have not the same number of type arguments")
       ys <- freshTyvars (length xs1)
       let subst1 = mkSubst xs1 (map (\x -> TyG (TyVar x)) ys)
           subst2 = mkSubst xs2 (map (\x -> TyG (TyVar x)) ys)
           ts1 :: [Type] = apply subst1 $ map argType (ms_args msig1)
           mt1 = apply subst1 $ ms_return msig1
           ps1 = apply subst1 $ ms_cons msig1
           ts2 = apply subst2 $ map argType (ms_args msig2)
           mt2 = apply subst2 $ ms_return msig2
           ps2 = apply subst2 $ ms_cons msig2
       unless (ts1 == ts2) $
              failGI ("Type error: method signatures " ++ showPPQ msig1 ++ 
                      " and " ++ showPPQ msig2 ++ 
                      " have mismatched arguments")
       unless (ps1 == ps2) $
              failGI ("Type error: method signatures " ++ showPPQ msig1 ++ 
                      " and " ++ showPPQ msig2 ++ 
                      " have mismatched constraints")
       case (mt1,mt2) of
         (Nothing,Nothing) -> return ()
         (Just t1, Just t2) ->
             do tenv' <- extendTEnv tenv (map asConP ps1)
                b <- sub tenv' t1 t2
                unless b $
                       failGI ("Type error: method signatures " ++ 
                               showPPQ msig1 ++ 
                               " and " ++ showPPQ msig2 ++ 
                               " have mismatched return types")
         _ -> 
             failGI ("Type error: method signatures " ++ showPPQ msig1 ++ 
                     " and " ++ showPPQ msig2 ++ 
                     " have mismatched return types")
    where
      asConP :: ConGP -> ConP
      asConP (ConGR (ImplG ts k)) = ConR (Impl (map asGType ts) k)
      asConP (ExtG t u) = Ext (asTyvar t) u
      asConP (MonoG t) = Mono (asTyvar t)
      asGType (TyG g) = g
      asGType t = bug ("ProgramTyping.assertSubMSig: cannot convert " ++ showPP t
                       ++ " to a G-type")
      asTyvar (TyG (TyVar x)) = x
      asTyvar t = bug ("ProgramTyping.assertSubMSig: cannot convert " ++ showPP t
                       ++ " to a type variable")

assertOKOverride :: TEnv -> MethodName -> MethodSig -> TyN -> GI ()
assertOKOverride tenv m msig n =
    do sigs <- allSuperSigs n
       mapM_ (assertSubMSig tenv msig) sigs
    where
      allSuperSigs :: TyN -> GI [MethodSigG]
      allSuperSigs n =
          do mn' <- upOneClass n
             case mn' of
               Nothing -> return []
               Just n'@(TyClass c _) ->
                   do mmdef <- findMDef c m
                      case mmdef of
                        Nothing -> allSuperSigs n'
                        Just _ ->
                            do msig <- mtypeClass m n'
                               msigs <- allSuperSigs n'
                               return (msig : msigs)

assertOKMSig :: TEnv -> MethodSig -> GI ()
assertOKMSig tenv msig =
    let xs = ms_tyargs msig
        ts = map argType (ms_args msig)
        mt = ms_return msig
        ps = ms_cons msig
    in do tenv' <- extendTEnv (addTyvars tenv xs) ps
          b1 <- typesOK tenv' ts
          b2 <- case mt of
                  Nothing -> return True
                  Just t -> typeOK tenv' t
          b3 <- constraintsOK tenv' (map asConGP ps)
          unless (b1 && b2 && b3) $ 
                 failGI ("Type error: method signature " ++ showPP msig 
                         ++ " not well-formed: " ++ show b1 ++ ", " ++
                        show b2 ++ ", " ++ show b3)

asConGP :: ConP -> ConGP
asConGP (ConR (Impl gs k)) = ConGR (ImplG (map TyG gs) k)
asConGP (Ext x t) = ExtG (TyG (TyVar x)) t
asConGP (Mono x) = MonoG (TyG (TyVar x))

assertOKMDef :: TEnv -> Maybe Type -> MethodDef -> GI ()
assertOKMDef tenv thisType mdef =
    do let msig = md_sig mdef
           xs = ms_tyargs msig
           ts = map argType (ms_args msig)
           ys = map argName (ms_args msig)
           mt = ms_return msig
           ps = ms_cons msig
           venv = case thisType of
                    Nothing -> mkVEnv ys ts
                    Just t -> mkVEnv (md_this mdef : ys) (t : ts)
       tenv' <- extendTEnv (addTyvars tenv xs) ps
       assertOKMSig tenv msig
       (_,returns) <- stmtsTypes tenv' venv mt (md_body mdef)
       case (mt, returns) of
         (Just _, False) ->
             failGI ("Type error: body of method " ++ showPPQ mdef ++ 
                     " does not necessarily return")
         _ -> return ()

assertOKConstructor :: TEnv -> Constructor -> TyN -> GI ()
assertOKConstructor tenv constr n =
    do let ts = map argType (con_args constr)
           ys = map argName (con_args constr)
           this = con_this constr
           venv = mkVEnv (this:ys) (TyG (TyN n) : ts)
       b1 <- typesOK tenv ts
       unless b1 $ failGI ("Type error: arguments of constructor of " ++ 
                           showPP n ++ " are not well-formed")
       stmtsTypes tenv venv Nothing (con_body constr)
       return ()

assertOKMDefInClass :: TEnv -> MethodDef -> TyN -> GI ()
assertOKMDefInClass tenv mdef n =
    do addTrace ("Type error: checking method definition\n" ++ showPP mdef 
                 ++ "\nin class " ++ showPP n ++ " failed")
                (assertOKMDef tenv (Just $ TyG (TyN n)) mdef)
       addTrace ("Type error: method\n" ++ showPP mdef ++ "\nin class " ++ 
                 showPP n
                 ++ " does not correctly override methods in superclasses")
                (assertOKOverride tenv (ms_name (md_sig mdef)) (md_sig mdef) n)

assertOKRCSig :: TEnv -> ReceiverSig -> GI ()
assertOKRCSig tenv rcsig =
    do addTrace ("Type error: checking signatures of receiver " ++ 
                 showPP (rcsig_name rcsig) ++ " failed")
                (mapM_ (assertOKMSig tenv) (rcsig_sigs rcsig))

assertImplMethod :: TEnv -> Maybe Type -> MethodDef -> MethodSigG -> GI ()
assertImplMethod tenv thisType mdef msig = 
    do assertOKMDef tenv thisType mdef
       assertSubMSig tenv (md_sig mdef) msig

assertImplMethods :: TEnv -> Maybe Type -> [MethodDef] -> [MethodSigG] -> GI ()
assertImplMethods tenv thisTy mdefs msigs =
    do let mdefs' = sortBy (\m1 m2 -> compare (mdefName m1) (mdefName m2)) mdefs
           msigs' = sortBy (\s1 s2 -> compare (ms_name s1) (ms_name s2)) msigs
       unless (map mdefName mdefs == map ms_name msigs) $
              fail ("Type error: methods " ++ show (map mdefName mdefs) ++ 
                    " and " ++ show (map ms_name msigs) ++ " do not match")
       mapM_ (\ (mdef,msig) -> assertImplMethod tenv thisTy mdef msig)
                 (zip mdefs' msigs')

assertOKClass :: Class -> GI ()
assertOKClass cls =
    do info ("Checking class " ++ (class_name cls))
       let c = class_name cls
           xs = class_tyargs cls
           mn = class_super cls
           ps = class_cons cls
           ts = map fd_type (class_fields cls)
           mdefs = class_methods cls
           mconstr = class_constructor cls
           cx = TyClass c (map (TyG . TyVar) xs)
       checkClassDef c
       tenv <- mkTEnv ps xs
       b1 <- case mn of
               Just n -> 
                   do debug "Checking well-formedness of superclass" 
                      typeOK tenv (TyG (TyN n))
               Nothing -> return True
       debug ("Checking well-formedness of field types")
       b2 <- typesOK tenv ts
       debug ("Checking well-formedness of constraints")
       b3 <- constraintsOK tenv (map asConGP ps)
       debug ("b1 = " ++ show b1 ++ ", b2 = " ++ show b2 ++ ", b3 = " ++ show b3)
       unless (b1 && b2 && b3) $
              failGI ("Type error: types/constraints used in definition of " ++
                      c ++ " not well-formed")
       debug ("Checking well-formedness of method definitions")
       mapM_ (\mdef -> assertOKMDefInClass tenv mdef cx) mdefs
       case mconstr of
         Nothing -> return ()
         Just constr -> assertOKConstructor tenv constr cx

assertOKIface :: Interface -> GI ()
assertOKIface (Interface i xs ys rs ps staticSigs recvSigs) =
    do info ("Checking interface " ++ i)
       checkIfaceDef i
       let rsps = (map ConR rs ++ ps)
       tenv <- mkTEnv rsps (xs ++ ys)
       b1 <- constraintsOK tenv (map asConGP rsps)
       unless b1 $
              failGI ("Type error: constraints used in definition of " ++ i ++
                      " not well-formed")
       mapM_ (assertOKMSig tenv) staticSigs
       mapM_ (assertOKRCSig tenv) recvSigs

assertOKImpl :: Implementation -> GI ()
assertOKImpl impl@(Implementation xs k@(TyIface i ts) 
                                  ns ps staticDefs recvDefs _) =
    do let id = showPP k ++ "[" ++ showPP ns ++ "]"
       info ("Checking implementation " ++ id)
       checkImplDef impl
       (Interface _ ys zs _ _ staticSigs recvSigs) <- findIface i
       tenv <- mkTEnv ps xs
       let nsType = map (TyG . TyN) ns
           subst = mkSubst (ys ++ zs) (ts ++ nsType)
           cons = ConGR $ ImplG nsType k
       b1 <- constraintsOK tenv (map asConGP ps)
       unless b1 $
              failGI ("Type error: constraints used in implementation " ++ id
                      ++ " not well-formed")
       b2 <- constraintOK tenv cons
       unless b2 $ 
              failGI ("Type error: constraint " ++ showPP cons ++ 
                      " not well-formed")
       assertImplMethods tenv Nothing staticDefs (apply subst staticSigs)
       sortedRecvSigs <- sortRecvSigs zs recvSigs
       mapM_ (\ (recvDef, recvSig, i) -> 
                  let thisType = nsType !! (i-1)
                  in assertImplMethods tenv (Just thisType)
                         (rcdef_defs recvDef)
                         (apply subst (rcsig_sigs recvSig)))
             (zip3 recvDefs sortedRecvSigs [1..])
    where
      sortRecvSigs [] [] = return []
      sortRecvSigs [] _ = failGI ("Type error: to many receiver signatures")
      sortRecvSigs (z:zs) recvSigs = 
          case findAndDelete (\rsig -> z == rcsig_name rsig) recvSigs of
            Nothing ->
                failGI ("Type error: no method signatures for implementing type "
                        ++ show z)
            Just (sig, recvSigs') ->
                do rest <- sortRecvSigs zs recvSigs'
                   return (sig:rest)
          
assertOKProg :: Program -> GI ()
assertOKProg (Program defs) = 
    do checkProgram
       mapM_ assertOKDef defs
    where
      assertOKDef (ImplDef impl) = assertOKImpl impl
      assertOKDef (IfaceDef iface) = assertOKIface iface
      assertOKDef (ClassDef cls) = assertOKClass cls