{-# LANGUAGE PatternSignatures #-}

module Restrictions where

import Data.Maybe
import Control.Monad
import Data.Set (Set)
import qualified Data.Set as Set
import Data.List

import AST
import GIMonad
import Inheritance
import Pretty
import Types
import Utils
import Environment
import Subtyping
import Logging
import Unification
import Graph
import Positions
import Dispatch

checkClassDef :: ClassName -> GI ()
checkClassDef clsName =
    do -- R-Class-1 
       supers <- allSuperClasses clsName
       mapM_ checkFieldNamesUnique supers
       -- R-Class-2 checked when constructing symtab
    where
      checkFieldNamesUnique clsName' = 
          do fs <- findAllFieldNames clsName
             fs' <- findAllFieldNames clsName'
             if Set.null (fs `Set.intersection` fs')
                then return ()
                else failGI ("Class " ++ clsName ++ " shadows some field of "
                             ++ "class " ++ clsName')

isGTyvarFrom :: Set TyvarName -> TyG -> Bool
isGTyvarFrom ys (TyVar x) = x `Set.member` ys
isGTyvarFrom _ _ = False

atTop :: Set TyvarName -> Type -> Bool
atTop _ (TyG (TyVar _)) = True
atTop ys t = ys `disjointFrom` free t

checkIfaceDef :: IfaceName -> GI ()
checkIfaceDef ifaceName =
    do -- R-Iface-1 checked while constructing symtab
       iface <- findIface ifaceName
       let ys = Set.fromList (iface_impl_types iface)
       -- R-Iface-2, R-Iface-3
       mapM (checkConstraint ys) (iface_impl_cons iface)
       -- R-Iface-4
       mapM_ (\rcsig -> mapM_ (checkSig ys) (rcsig_sigs rcsig))
             (iface_receiver_sigs iface)
       -- R-Prog-7, R-Prog-8
       case iface_impl_types iface of
         [_] -> do supers <- allSuperInterfaces 
                               (TyIface ifaceName (map (TyG . TyVar) 
                                                       (iface_tyargs iface)))
                   -- supers contains no duplicates, so it cannot contain
                   -- more than one type using the same interface name
                   let superNames = map tyi_name supers
                   if any (\n -> count n superNames >= 2) superNames
                      then failGI ("Interface " ++ ifaceName ++ 
                                   " violates restriction R-Prog-7")
                      else return ()
                   plus <- posPlus ifaceName 1
                   minus <- posMinus ifaceName 1
                   if not plus && not minus
                      then do b <- allM (\ (k1,k2) -> 
                                             do b1 <- k1 `extendsIface` k2
                                                b2 <- k2 `extendsIface` k1
                                                return (b1 || b2))
                                        [(k1,k2) | k1 <- supers, k2 <- supers]
                              unless b $ 
                                     failGI ("Interface " ++ ifaceName ++ 
                                             " violates restriction R-Prog-8")
                      else return ()
         _ -> return ()
    where
      checkConstraint ys r@(Impl gs k) =
          if (ys `disjointFrom` free k) && uniqueElems gs &&
             all (isGTyvarFrom ys) gs then return ()
             else failGI ("Constraint " ++ showPPQ r ++ " of interface " ++ 
                          ifaceName ++ " violates restriction R-Iface-2 or " ++
                          "R-Iface-3")
      checkSig ys (MethodSig _ mu m args qs) =
          if (ys `disjointFrom` free qs) && 
             (all (\ (MethodArg (t,_)) -> atTop ys t) args) &&
             (case mu of
               Just u -> atTop ys u
               Nothing -> True)
             then return ()
             else failGI ("Method signature of " ++ m ++ " of interface " ++
                          ifaceName ++ " violates restriction R-Iface-4")

checkImplDef :: Implementation -> GI ()
checkImplDef (Implementation xs k@(TyIface i _) ns ps _ _ _) = 
    do -- R-Impl-1
       disps <- dispSet i
       let dispNs = [ n | (n,j) <- (zip ns [1..]), j `Set.member` disps ]
           b1 = Set.fromList xs `Set.isSubsetOf` free dispNs
       unless b1 $ failGI ("Implementation " ++ showPPImplHeader k ns  
                           ++ " violates restriction R-Impl-1")
       -- R-Impl-2 already checked when checking well-formedness of
       -- constraint "ns implements i<vs>"
       -- R-Impl-3
       let b2 = all (\p -> case p of
                             (ConR (Impl gs _)) ->
                                 all (\g -> case g of
                                              TyVar _ -> True
                                              _ -> False) gs
                             _ -> True) ps
       unless b2 $ failGI ("Implementation " ++ showPPImplHeader k ns  
                           ++ " violates restriction R-Impl-3")

allDisjointImplementations :: GI [(Implementation, Implementation)]
allDisjointImplementations = 
    do impls <- allImpls
       let pairs = [(i1,i2) | i1 <- impls, i2 <- impls, i1 /= i2, 
                              getName i1 == getName i2 ]
           getName = tyi_name . impl_iface
       return $
         map (\ (i1, Implementation xs' k' ms' ps' staticDefs' recvDefs' id) ->
                   let xs = take (length xs') 
                              [(case impl_tyargs i1 of
                                  [] -> 0
                                  l -> maximum l + 1) ..]
                       renaming = mkRenaming xs' xs
                       k = apply renaming k'
                       ms = apply renaming ms'
                       ps = map asConP' (apply renaming ps')
                       staticDefs = map asMDef (apply renaming staticDefs')
                       recvDefs = map (\rcdef -> ReceiverDef $
                                           map asMDef (apply renaming 
                                                       (rcdef_defs rcdef)))
                                      recvDefs'
                   in (i1, Implementation xs k ms ps staticDefs recvDefs id)) 
             pairs
    where
      asConP' :: ConGP -> ConP
      asConP' gp = case asConP gp of
                     Just p -> p
                     Nothing -> bug ("Restrictions.allDisjointImplementations: "
                                     ++ "cannot convert " ++ showPPQ gp
                                     ++ " to a P-constraint")
      asMDef :: MethodDefG -> MethodDef
      asMDef def = 
          let sig = md_sig def
              sig' =  sig { ms_cons = map asConP' (ms_cons sig) }
          in def { md_sig = sig' }

data TyParamDepGraphLabel = Expansive | NonExpansive
                          deriving (Eq,Show)

checkProgram :: GI ()
checkProgram =
    do checkRProg5
       checkRTEnv3
       -- R-Prog-1 already checked when constructing symtab
       implPairs <- allDisjointImplementations
       mapM_ checkRProg2 implPairs
       mapM_ checkRProg3 implPairs
       mapM_ checkRProg4 implPairs
       mapM_ checkRProg6 implPairs
    where
      checkRProg2 (Implementation xs k@(TyIface i ts) ms _ _ _ _,
                   Implementation ys k'@(TyIface _ us) ns _ _ _ _) =
          do debug ("Checking restriction R-Prog-2")
             disps <- dispSet i
             let l = [ (TyN m, TyN n) | (m,n,j) <- (zip3 ms ns [1..]), 
                                        j `Set.member` disps ]
             msubst <- unifyModGLB emptyTEnv (xs++ys) l
             case msubst of
               Nothing -> return ()
               Just subst ->
                   do let l' = [ (m, n) | (m,n,j) <- (zip3 ms ns [1..]), 
                                          not (j `Set.member` disps) ]
                          b1 = all (\ (m,n) -> (apply subst m :: TyN)
                                               == apply subst n) l'
                          b2 = (apply subst ts :: [Type]) == apply subst us
                      unless (b1 && b2) $ 
                             failGI ("Implementations " ++ 
                                     showPPImplHeader k ms ++ " and " ++
                                     showPPImplHeader k' ns ++ 
                                     " violate restriction R-Prog-2")
      checkRProg3 (Implementation xs k@(TyIface i _) ms _ _ _ _,
                   Implementation ys k'@(TyIface _ _) ns _ _ _ _) =
          do debug ("Checking restriction R-Prog-3")
             msubst <- unifyModGLB emptyTEnv (xs++ys)
                                   (zip (map TyN ms) (map TyN ns))
             case msubst of 
               Nothing -> return ()
               Just subst ->
                   do mglbsMsNs <- glbs emptyTEnv (map (TyG . TyN )
                                                       (apply subst ms))
                                                  (map (TyG . TyN )
                                                       (apply subst ns))
                      case mglbsMsNs of
                        Nothing -> bug ("Restrictions.checkProgram: glbs " ++
                                        "returned Nothing")
                        Just glbsMsNs ->
                            do impls <- findAllImpls i
                               let b = any (implMatches (map asTyN glbsMsNs)) 
                                           impls
                               unless b $
                                   failGI ("Implementations " ++ 
                                           showPPImplHeader k ms ++ " and " ++
                                           showPPImplHeader k' ns ++ 
                                           " violate restriction R-Prog-3")
          where
            implMatches ns (Implementation xs _ ms _ _ _ _) =
                case unify' (\x -> x `elem` xs) (zip (map (TyG . TyN) ns) 
                                                     (map (TyG . TyN) ms)) of
                  Nothing -> False
                  _ -> True
            asTyN :: Type -> TyN
            asTyN (TyG (TyN n@(TyClass _ _))) = n
            asTyN t = bug ("Restrictions.asTyN: cannot convert " ++ showPPQ t ++ 
                           " to a class type")
      checkRProg4 (Implementation xs k@(TyIface _ _) ms ps _ _ _,
                   Implementation ys k'@(TyIface _ _) ns qs _ _ _) =
          do debug ("Checking restriction R-Prog-4")
             msubst <- unifyModSub emptyTEnv (xs++ys)
                                   (zip (map (TyG . TyN) ms) 
                                        (map (TyG . TyN) ns))
             case msubst of
               Nothing -> return ()
               Just subst ->
                   do b <- allM (consHolds (apply subst qs :: [ConGP])) 
                                (apply subst ps :: [ConGP])
                      unless b $
                             failGI ("Implementations " ++ 
                                     showPPImplHeader k ms ++ " and " ++
                                     showPPImplHeader k' ns ++ 
                                     " violate restriction R-Prog-4")
          where
            consHolds qs p =
                do let goodQs = catMaybes (map asConP qs)
                   tenv <- mkTEnv goodQs []
                   b <- entails tenv p
                   if b then return True
                      else case p of
                             ConGR gr -> 
                                 do supQs <- sups (mapMaybe asConGR qs)
                                    return (gr `Set.member` supQs)
                             ExtG t u ->
                                 anyM (\q -> case q of
                                               ExtG t' u'| t == t' -> 
                                                   sub' tenv u' u
                                               _ -> return False) qs
                             MonoG _ -> return (p `elem` qs)
            asConGR :: ConGP -> Maybe ConGR
            asConGR (ConGR gr) = Just gr
            asConGR _ = Nothing
      checkRProg5 =
          do cg <- buildClassGraph
             unless (null $ enumAllCycles cg) $
                    failGI ("There is a cycle in the class hierarchy")
             ig <- buildIfaceGraph
             unless (null $ enumAllCycles ig) $
                    failGI ("There is a cycle in the interface hierarchy")
          where
            buildClassGraph :: GI (DiGraph ClassName)
            buildClassGraph =
                do classes <- allClasses
                   let nodes = map class_name classes
                       edges = mapMaybe (\cd -> 
                                             do (TyClass c _) <- class_super cd
                                                return (class_name cd, c))
                                        classes
                   return (buildUnlabelledGraph nodes edges)
            buildIfaceGraph :: GI (DiGraph IfaceName)
            buildIfaceGraph =
                do interfaces <- allInterfaces
                   let nodes = map iface_name interfaces
                       edges = concatMap (\id -> map (\ (Impl _ (TyIface i _)) ->
                                                          (iface_name id, i))
                                                     (iface_impl_cons id))
                                         interfaces
                   return (buildUnlabelledGraph nodes edges)
      checkRProg6 (Implementation xs k ms _ _ _ _,
                   Implementation ys k' ns _ _ _ _) =
          case unify' (\x -> x `elem` xs || x `elem` ys) 
                      (zip (map (TyG . TyN) ms) (map (TyG . TyN) ns)) of
            Nothing -> return ()
            Just _ -> failGI ("Implementations " ++ 
                              showPPImplHeader k ms ++ " and " ++
                              showPPImplHeader k' ns ++ 
                              " overlap (R-Prog-6)")
      checkRTEnv3 =
          do classes <- allClasses
             let nodes = concatMap (\cls -> map (\i -> (class_name cls, i))
                                              [1 .. (length $ class_tyargs cls)])
                                   classes
                 edges = concatMap tyParamDepGraphEdges classes
                 tyParamDepGraph = buildGraph nodes edges
                 cycles = enumAllCycles tyParamDepGraph
             if (Expansive `elem` (map (\ (_,l,_) -> l) (concat cycles)))
                then failGI ("Type parameter dependency graph is expansive " ++
                             "(T-TEnv-3)")
                else return ()
          where
            tyParamDepGraphEdges (Class _ _ Nothing _ _ _ _) = []
            tyParamDepGraphEdges (Class c xs (Just n) _ _ _ _) =
                loop c (zip xs [1..]) [] [] (TyG (TyN n))
            loop :: ClassName
                 -> [(TyvarName,Int)]       -- Type variables and their index
                 -> [(ClassName, Int)]      -- Direct parent subterm, the Int
                                            -- component is the index of the
                                            -- child. Contains either 0 or 1
                                            -- element.
                 -> [(ClassName,Int)]       -- Ancestor subterms, excluding
                                            -- direct parents.
                 -> Type                    -- current subterm
                 -> [((ClassName,Int), TyParamDepGraphLabel, (ClassName,Int))]
            loop name vars parents ancestors ty =
                case ty of
                  TyG (TyVar y) ->
                      case find (\ (x,_) -> x == y) vars of
                        Just (_, i) ->
                            map (\to -> ((name,i), NonExpansive, to)) parents
                            ++
                            map (\to -> ((name,i), Expansive, to)) ancestors
                        Nothing -> []
                  TyG (TyN (TyClass d ts)) ->
                      concatMap (\ (t,i) -> loop name vars [(d,i)] 
                                                 (parents++ancestors) t)
                                (zip ts [1..])
                  TyK (TyIface _ ts) ->
                      concatMap (\ t -> loop name vars [] (parents++ancestors) t)
                                ts
                  _ -> []
                 
asConP :: ConGP -> Maybe ConP
asConP (ConGR (ImplG ts k)) = 
    do gs <- mapM asGType ts
       return (ConR (Impl gs k))
asConP (ExtG t u) = 
    do x <- asTyvar t
       return (Ext x u)
asConP (MonoG t) = 
    do x <- asTyvar t
       return (Mono x)

asGType :: Type -> Maybe TyG
asGType (TyG g) = Just g
asGType _ = Nothing

asTyvar :: Type -> Maybe TyvarName
asTyvar (TyG (TyVar x)) = Just x
asTyvar _ = Nothing

asTyN :: Type -> TyN
asTyN (TyG (TyN n@(TyClass _ _))) = n
asTyN t = bug ("Restrictions.asTyN: cannot convert " ++ showPPQ t ++ 
               " to a class type")
