module Subtyping where

import Data.Maybe
import Data.Set (Set)
import qualified Data.Set as Set

import GIMonad
import AST
import {-# SOURCE #-} Environment
import Types
import Inheritance
import Utils
import Logging
import Unification
import Positions
import Pretty

sub' :: TEnv -> Type -> Type -> GI Bool
sub' tenv t u = 
    do debug (showPP tenv ++ " |-' " ++ showPP t ++ " <= " ++ showPP u ++ ": ?")
       res <- sub'' tenv t u
       debug (showPP tenv ++ " |-' " ++ showPP t ++ " <= " ++ showPP u ++ ": " ++
              show res)
       return res

sub'' :: TEnv -> Type -> Type -> GI Bool
sub'' tenv t u = 
    if (isObject u || isBot t) && not (isPrim t)
       then return True
       else case t of
              TyPrim _ ->
                  return (t == u)
              TyG (TyVar x) ->
                  if t == u then return True
                  else anyM (\v -> sub'' tenv v u) (upperBounds x tenv)
              TyG (TyN n) ->
                  case u of
                    TyG (TyN n') -> n `extendsClass` n'
                    _ -> return False
              TyK k ->
                  case u of
                    TyK k' -> k `extendsIface` k'
                    _ -> return False
              t -> bug ("Subtyping.sub'': unexpected type " ++ show t)

sub :: TEnv -> Type -> Type -> GI Bool
sub tenv t u = 
    do debug (showPP tenv ++ " |- " ++ showPP t ++ " <= " ++ showPP u ++ ": ?")
       res <- subAux tenv Set.empty t u
       debug (showPP tenv ++ " |- " ++ showPP t ++ " <= " ++ showPP u ++ ": " ++
              show res)
       return res

data SubWithCoercion = NoSubtype 
                     | SubtypeWithoutCoercion 
                     | SubtypeWithCoercion TyK
                     deriving (Eq,Show)

subWithCoercion :: TEnv -> Type -> Type -> GI SubWithCoercion
subWithCoercion tenv t u =
    do debug (showPP tenv ++ " |- " ++ showPP t ++ " <= " ++ showPP u ++ ": ?")
       res <- subWithCoercionAux tenv Set.empty t u
       debug (showPP tenv ++ " |- " ++ showPP t ++ " <= " ++ showPP u ++ ": " ++
              show res)
       return res

subAux :: TEnv -> GoalCache -> Type -> Type -> GI Bool
subAux tenv cache t u =
    do b <- sub'' tenv t u
       if b then return True
          else case u of
                 TyK k -> entailsAux tenv cache True (ConGR (ImplG [t] k))
                 _ -> return False

subWithCoercionAux :: TEnv -> GoalCache -> Type -> Type -> GI SubWithCoercion
subWithCoercionAux tenv cache t u =
    do b <- sub'' tenv t u
       if b then return SubtypeWithoutCoercion
          else case u of
                 TyK k -> 
                     do b' <- entailsAux tenv cache True (ConGR (ImplG [t] k))
                        return $ if b' then SubtypeWithCoercion k
                                       else NoSubtype
                 _ -> return NoSubtype

entails :: TEnv -> ConGP -> GI Bool
entails tenv p = 
    do debug (showPP tenv ++ " ||- " ++ showPP p ++ ": ?")
       res <- entailsAux tenv Set.empty False p
       debug (showPP tenv ++ " ||- " ++ showPP p ++ ": " ++ show res)
       return res

entailsAll :: TEnv -> [ConGP] -> GI Bool
entailsAll tenv ps = andM (map (entails tenv) ps)

-- Given a TEnv and a non-empty list L of G-types, returns a type H such that 
-- there exists no H' /= H in L with "TEnv |-a' H' <= H".
gminimal :: TEnv -> [TyG] -> GI TyG
gminimal _ [] = bug ("Subtyping.gminimal: empty list")
gminimal tenv (x:xs) = iter x xs
    where iter g [] = return g
          iter g (h:hs) = 
              do useH <- sub'' tenv (TyG h) (TyG g)
                 if useH then iter h hs else iter g hs

type UPSub = [(Type,Type)]

unifyModSub :: TEnv -> [TyvarName] -> UPSub -> GI (Maybe TySubst)
unifyModSub tenv xs up =
    do res <- unifyModSub' 
       debug ("Unify<=(" ++ showPP tenv ++ ", [" ++ showPP xs ++ "], " 
              ++ showPP up ++ ") returned " ++ showPP res)
       return res
    where
      unifyModSub' :: GI (Maybe TySubst)
      unifyModSub' = 
          do l <- reduce [] up
             return $ findFirstJust (\up' -> unify' (\x -> x `elem` xs) up') l
      reduce :: UPSub -> UPSub -> GI [UPSub]
      reduce done [] = return [done]
      reduce done (first : rest) =
          case first of
            (TyBot, _) -> reduce done rest
            -- U-IFACE1
            (TyK _, u@(TyG _)) -> reduce done ((tyObject, u) : rest)
            -- U-IFACE2, U-IFACE3
            (TyK k@(TyIface i _), u@(TyK (TyIface j _)))
                | i == j -> reduce (first:done) rest
                | otherwise ->
                    do l <- upOneIface k
                       if null l
                          then reduce done ((tyObject, u) : rest)
                          else concatMapM 
                                   (\k' -> reduce done ((TyK k', u) : rest)) l
            -- U-CLASS
            ((TyG (TyN n1@(TyClass c _))), u@(TyG (TyN (TyClass d _))))
                | c == d -> reduce (first:done) rest
                | otherwise ->
                    do mn1' <- upOneClass n1
                       case mn1' of
                         Nothing -> return [] -- fail immediately
                         Just n1' -> reduce done ((TyG (TyN n1'), u) : rest)
            -- U-VAR-ENV, U-VAR-OBJECT
            (TyG (TyVar x), u) ->
                let bounds = upperBounds x tenv
                in case bounds of
                     [] -> reduce done ((tyObject, u) : rest)
                     l -> concatMapM (\t -> reduce done ((t,u) : rest)) l
            p -> reduce (p:done) rest

glb :: TEnv -> Type -> Type -> GI (Maybe Type)
glb tenv t u =
    do b1 <- sub tenv t u
       if b1 then return $ Just t
          else do b2 <- sub tenv u t
                  if b2 then return (Just u) else return Nothing

glbs :: TEnv -> [Type] -> [Type] -> GI (Maybe [Type])
glbs tenv ts us =
    do l <- mapM (\ (t,u) -> glb tenv t u) (zip ts us)
       if Nothing `elem` l then return Nothing
          else return (Just $ catMaybes l)


type UPGLB = [(TyG,TyG)]

unifyModGLB :: TEnv -> [TyvarName] -> UPGLB -> GI (Maybe TySubst)
unifyModGLB tenv xs up =
    let ups = allPossibilities up
    in findFirstJustM (\up -> unifyModSub tenv xs up) ups
    where allPossibilities :: UPGLB -> [UPSub]
          allPossibilities [] = [[]]
          allPossibilities ((t,u) : rest) = 
              let l = allPossibilities rest
                  p1 = (TyG t, TyG u)
                  p2 = (TyG u, TyG t)
              in concatMap (\up -> [p1:up, p2:up]) l

                    
type GoalCache = Set ConR

entailsAuxAll :: TEnv -> GoalCache -> Bool -> [ConGP] -> GI Bool
entailsAuxAll tenv cache beta ps = andM (map (entailsAux tenv cache beta) ps)
              
entailsAux :: TEnv -> GoalCache -> Bool -> ConGP -> GI Bool
entailsAux tenv cache beta p = 
    case p of
      ExtG t u -> subAux tenv cache t u
      MonoG t ->
          case t of
            TyG (TyVar x) -> return $ x `Set.member` monoVars tenv
            TyG (TyN _) -> return True
            _ -> return False
      ConGR (ImplG ts (TyIface i vs)) ->
          do x <- entailsAuxNil tenv cache beta 
                      (ImplNil (map Just ts) i (map Just vs)) 
             case x of
               Nothing -> return False
               _ -> return True

fillWithFreshTyvars :: [Maybe Type] -> GI ([TyvarName], [Type])
fillWithFreshTyvars tsNil = 
    do l <- mapM fill tsNil
       let (mtyvars, ts) = unzip l
       return (catMaybes mtyvars, ts)
    where
      fill (Just t) = return (Nothing, t)
      fill Nothing = do x <- freshTyvar
                        return (Just x, TyG (TyVar x))


entailsAuxNil :: TEnv -> GoalCache -> Bool -> ConNil -> GI (Maybe ConGR)
entailsAuxNil tenv cache beta p@(ImplNil tsNil i vsNil) =
    firstJustM 
        [-- ENT-NIL-ALG-ENV
         do let sups = supImplCons tenv
            findFirstJustM 
              (\ (ImplG us (TyIface i' vs)) -> 
                    if not (matches vsNil vs) || i /= i' then return Nothing
                       else do mts <- liftNil tenv beta i tsNil us
                               case mts of
                                 Nothing -> return Nothing
                                 Just ts -> 
                                     return (Just $ ImplG ts (TyIface i vs)))
              (Set.toList sups)
        ,-- ENT-NIL-ALG-IFACE1
         case tsNil of
           [Just t] -> do (xs, vs) <- fillWithFreshTyvars vsNil
                          msubst <- unifyModSub tenv xs [(t, TyK (TyIface i vs))]
                          case msubst of
                            Nothing -> return Nothing
                            Just subst ->
                                do let iv = TyIface i (apply subst vs)
                                   b1 <- lift tenv beta i [t] [TyK iv]
                                   b2 <- posPlus i 1
                                   if (b1 && b2) 
                                      then return $ Just (ImplG [t] iv) 
                                      else return Nothing
           _ -> return Nothing
        ,-- ENT-NIL-ALG-IFACE2
         case tsNil of
           [Just t@(TyK (TyIface j _))] ->
               do b <- posPlus j 1
                  if not b then return Nothing
                     else do (xs, vs) <- fillWithFreshTyvars vsNil
                             msubst <- unifyModSub tenv xs 
                                         [(t, TyK (TyIface i vs))]
                             case msubst of
                               Nothing -> return Nothing
                               Just subst ->
                                   let iv = TyIface i (apply subst vs)
                                   in return $ Just (ImplG [t] iv)
           _ -> return Nothing
        ,do impls <- findAllImpls i
            findFirstJustM (\impl ->
                let xs = impl_tyargs impl
                    ws = tyi_args (impl_iface impl)
                    ns = impl_impl_types impl
                    ps = impl_cons impl
                    up = mapMaybe (\x ->
                                        -- We know that x is not Nothing and
                                        -- not TyBot if x is a dispatch type
                                        -- The latter property is ensured by
                                        -- our definition of contrib in
                                        -- module MethodTypes. Moreover,
                                        -- we know that all types variables
                                        -- from x appear in the "dispatch part"
                                        -- of ns. Hence, solving the unification
                                        -- problem up determines instantiations
                                        -- for all types variables in xs.
                                       case x of
                                         (Nothing,_) -> Nothing
                                         (Just t, n) -> Just (t, TyG (TyN n)))
                                  (zip tsNil ns)
                in do debug ("Checking whether constraint " ++ showPP p ++
                             " holds by " ++ showPP impl)
                      msubst <- unifyModSub tenv xs up
                      debug ("After unifyModSub: " ++ showPP msubst)
                      case msubst of
                        Nothing -> return Nothing
                        Just subst ->
                            do mts <- liftNil tenv beta i tsNil
                                              (map (TyG . TyN) 
                                                   (apply subst ns))
                               case mts of
                                 Nothing -> return Nothing
                                 Just ts ->
                                  do let b1 = vsNil `matches` apply subst ws
                                         substI = TyIface i (apply subst ws)
                                         cons = Impl 
                                                  (map TyN (apply subst ns))
                                                  substI
                                         b2 = not (cons `Set.member` cache)
                                     debug ("b1 = " ++ show b1 ++ ", b2 = "
                                            ++ show b2)
                                     if not (b1 && b2) then return Nothing
                                        else do b <- entailsAuxAll tenv
                                                       (Set.insert cons cache)
                                                        False
                                                       (apply subst ps)
                                                debug ("Constraints of " ++
                                                       "definition hold: " ++
                                                       show b)
                                                let res = ImplG ts substI
                                                if b then return (Just res)
                                                   else return Nothing)
                 impls
        ]

entailsNil :: TEnv -> ConNil -> GI (Maybe ConGR)
entailsNil tenv con = entailsAuxNil tenv Set.empty False con

liftNil :: TEnv -> Bool -> IfaceName -> [Maybe Type] -> [Type] 
        -> GI (Maybe [Type])
liftNil tenv beta i tsNil us =
    do res <- mapM liftNilSingle (zip3 tsNil us [1..])
       let (ts, bs1, bs2) = unzip3 res
       if and bs1 && (beta || and bs2)
          then return (Just ts)
          else return Nothing
    where
      liftNilSingle (tNil, u, j) =
          do let t = case tNil of
                       Nothing -> u
                       Just t -> t
             b1 <- case tNil of
                     Nothing -> return True
                     Just t -> sub'' tenv t u
             b2 <- case tNil of
                     Nothing -> return True
                     Just t | t /= u    -> posMinus i j
                            | otherwise -> return True
             return (t,b1,b2)

matches :: [Maybe Type] -> [Type] -> Bool
matches l1 l2 = all matchesSingle (zip l1 l2)
    where matchesSingle (Nothing, _) = True
          matchesSingle (Just t1, t2) = t1 == t2

lift :: TEnv -> Bool -> IfaceName -> [Type] -> [Type] -> GI Bool
lift tenv beta i ts us =
    do allM (\ (t,u,j) -> do b1 <- sub'' tenv t u
                             b2 <- posMinus i j
                             return (b1 && (beta || t == u || b2)))
            (zip3 ts us [1..])

allSuperTypes :: TEnv -> Type -> GI [Type]
allSuperTypes _ TyBot = bug ("Subtyping.allSuperTypes: TyBot")
allSuperTypes tenv t = 
    do s <- allSuperTypes' tenv t 
       return $ Set.toList s

allSuperTypes' :: TEnv -> Type -> GI (Set Type)
allSuperTypes' tenv ty = 
    do res <- allSuperTypes'' tenv ty
       debug ("allSuperTypes', ty = " ++ showPPQ ty ++ ", res = " ++ 
              showPPList (Set.toList res))
       return res

allSuperTypes'' :: TEnv -> Type -> GI (Set Type)
allSuperTypes'' _ TyBot = bug ("Subtyping.allSuperTypes'': TyBot")
allSuperTypes'' tenv t@(TyG (TyN n)) =
    do msup <- upOneClass n
       case msup of
         Nothing -> return $ Set.singleton t
         Just n' -> do s <- allSuperTypes'' tenv (TyG (TyN n'))
                       return $ Set.insert t s
allSuperTypes'' tenv t@(TyK k) = 
    do sups <- upOneIface k
       ls <- mapM (allSuperTypes'' tenv) (map TyK sups)
       return $ Set.insert t (Set.insert tyObject (Set.unions ls))
allSuperTypes'' tenv t@(TyG (TyVar x)) = 
    do let bounds = upperBounds x tenv
       ls <- mapM (allSuperTypes'' tenv) bounds
       return $ Set.insert t (Set.insert tyObject (Set.unions ls))
allSuperTypes'' _ t = return $ Set.singleton t

