module SubtypingTests ( runTests ) where

import Control.Monad.Trans
import Data.List
import Data.Maybe

import Environment
import AST
import Lexer
import Parser
import GIMonad
import Subtyping
import Logging
import Symtab
import Utils
import Pretty
import Types

prog :: String
prog = "class A<X> {}\n" ++
       "class B<X,Y> extends A<X> {}\n" ++
       "interface NotMinus [This] { This bar(); }\n" ++
       "interface JSuper {}\n" ++
       "interface J [X where X implements JSuper] {}\n" ++
       "interface I<X> {}\n" ++
       "interface ISub<X> [This where This implements I<X>, This implements J] {}\n"++
       "interface NotPlus [This where This implements I<Object>] {\n" ++
       "  void foo(This x);\n" ++
       "}\n" ++
       "interface NotPlusSub [This where This implements NotPlus] {}\n" ++
       "implementation<X> I<X> [A<X>] where X implements J {}\n" ++
       "implementation NotMinus [Object] {}" 

tenv :: [String]
tenv = ["X implements J",
        "XSub extends X",
        "Y extends Z",
        "Z extends A<X>"]

tyvars :: [String]
tyvars = ["X", "XSub", "Y", "Z", "W"]

asRealTyvar :: String -> TyvarName
asRealTyvar x = case elemIndex x tyvars of
                  Nothing -> fatal ("SubtypingTests.asRealTyvar: unknown type " ++
                                    "variable: " ++ x)
                  Just i -> TyvarName i

unifyModSubTests :: [([String], String, String, Maybe [(String,String)])]
unifyModSubTests = [([], "B<String,Object>", "A<String>", Just []),
                    (["W"], "B<W,Object>", "A<String>", Just [("W", "String")]),
                    ([], "B<X,Object>", "A<X>", Just []),
                    ([], "B<W,Object>", "A<String>", Nothing),
                    ([], "Y", "A<X>", Just []),
                    (["W"], "ISub<Object>", "I<W>", Just [("W", "Object")])
                   ]

subTests :: [(String,String,Bool)]
subTests = [("B<Object,X>", "A<Object>", True),
            ("B<X,Y>", "A<Z>", False),
            ("X", "J", True),
            ("X", "I<Object>", False),
            ("Y", "A<X>", True),
            ("ISub<Z>", "I<Z>", True),
            ("NotPlus", "I<Object>", True)]

entailTests :: [(String,Bool)]
entailTests = [("ISub<Z> implements I<Z>", True),
               ("NotPlus implements I<Object>", True),
               ("NotPlus implements NotPlus", False),
               ("NotPlusSub implements NotPlus", False),
               ("X implements J", True),
               ("X implements JSuper", True),
               ("Object implements NotMinus", True),
               ("String implements NotMinus", False),
               ("XSub implements J", True),
               ("A<XSub> implements I<XSub>", True),
               ("A<XSub> implements I<X>", False),
               ("B<X,Object> implements I<X>", True)]

getProg :: IO Program
getProg =
    do let toks = scan "<string>" prog
       prog <- parseProgram toks
       return prog

getTEnv :: Program -> GI TEnv
getTEnv prog =
    do cons <- mapM parseCons tenv
       mkTEnv cons []
    where parseCons s = 
              do debug ("Parsing constraint " ++ show s)
                 let toks = scan "<string>" s
                 liftIO $ parseConstraint toks tyvars prog

parseTy :: String -> Program -> IO Type
parseTy s prog = do debug ("Parsing type " ++ show s)
                    let toks = scan "<string>" s
                    parseType toks tyvars prog
                 
runUnifyModSubTests :: Program -> TEnv -> GI ()
runUnifyModSubTests prog tenv =
    mapM_ run unifyModSubTests 
    where 
      run (vars, s1, s2, expSubst) =
          do let testName = "``Unify<=(" ++ s1 ++ ", " ++ s2 ++ ")''"
             debug ("Running test " ++ testName)
             t <- liftIO $ parseTy s1 prog
             u <- liftIO $ parseTy s2 prog
             let realVars = map asRealTyvar vars
             msubst <- unifyModSub tenv realVars [(t,u)]
             debug ("Result of unification " ++ showPP msubst)
             let reportFailure = fatal ("Failed (test " ++ testName ++ 
                                        "): should return " ++ show expSubst ++ 
                                        ", but returned " ++ showPP msubst)
             case (expSubst, msubst) of
               (Nothing,Nothing) ->
                   info ("OK (test " ++ testName ++ ")")
               (Just subst1, Just subst2) ->
                   do al1Unsorted <- mapM (\ (k,v) -> do let x = asRealTyvar k
                                                         t <- liftIO $ parseTy v prog
                                                         return (x,t))
                                          subst1
                      let al1 = sortBy (\ (k1,_) (k2,_) -> compare k1 k2) al1Unsorted
                          al2 = tySubstAsList subst2
                      if al1 == al2
                         then info ("OK (test " ++ testName ++ ")")
                         else reportFailure
               _ -> reportFailure
                                                
runSubTests :: Program -> TEnv -> GI ()
runSubTests prog tenv =
    mapM_ run subTests
    where
      run (s1,s2,expRes) =
          do debug ("Running test " ++ quote s1 ++ " <= " ++ quote s2)
             t <- liftIO $ parseTy s1 prog
             u <- liftIO $ parseTy s2 prog
             b <- sub tenv t u
             if b == expRes 
                then info ("OK (test " ++ quote s1 ++ " <= " ++ quote s2 ++ ")")
                else fatal ("Failed (test " ++ quote s1 ++ " <= " ++ quote s2 ++ 
                            "): should return " ++ show expRes ++ 
                            " but returned " ++ show b)

runEntailTests :: Program -> TEnv -> GI ()
runEntailTests prog tenv =
    mapM_ run entailTests
    where
      run (s,expRes) =
          do debug ("Running test " ++ quote s)
             cons <- liftIO $ parse s
             b <- entails tenv cons
             if b == expRes 
                then info ("OK (test " ++ quote s ++ ")")
                else fatal ("Failed (test " ++ quote s ++
                            "): should return " ++ show expRes ++ 
                            " but returned " ++ show b)        
      parse s = do debug ("Parsing generalized constrained " ++ show s)
                   let toks = scan "<string>" s
                   parseConstraintG toks tyvars prog
    
runTests :: IO ()
runTests =
    do prog <- getProg
       let symtab = buildSymtab prog
       testRes <- withSymtab symtab $ do
         tenv <- getTEnv prog
         runUnifyModSubTests prog tenv
         runSubTests prog tenv
         runEntailTests prog tenv
       case testRes of
         Right _ -> info ("All tests OK")
         Left s -> info ("At least one test failed:\n" ++ s)
               
