module Symtab where

import Data.Map (Map)
import qualified Data.Map as Map
import Data.Set (Set)
import qualified Data.Set as Set

import AST
import Logging

type RecvIx = Int

-- the strictness annotations are necessary to ensure that all checks
-- during symtab construction are really performed.
data Symtab = Symtab { st_classes :: !(Map ClassName Class),
                       st_ifaces :: !(Map IfaceName Interface),
                       st_impls :: !(Map IfaceName [Implementation]),
                       st_mdefs :: !(Map (ClassName,MethodName) MethodDef),
                       st_fields :: !(Map (ClassName,FieldName) FieldDef),
                       st_fieldNames :: !(Map ClassName (Set FieldName)),
                       st_msigs :: !(Map MethodName (Interface,MethodSig,RecvIx)),
                       st_static_msigs :: !(Map (IfaceName,MethodName) MethodSig)
                     }
            deriving Show

emptySymtab :: Symtab
emptySymtab = Symtab Map.empty Map.empty Map.empty Map.empty Map.empty Map.empty
                     Map.empty Map.empty

buildSymtab :: Program -> Symtab
buildSymtab (Program defs) =
    foldr add emptySymtab defs
    where
      add (ImplDef i) tab = 
          let (TyIface name _) = impl_iface i
              impls = findAllImpls name tab
          in tab { st_impls = Map.insert name (i:impls) (st_impls tab) }
      add (IfaceDef i) tab =
          let name = iface_name i
          in case findIface name tab of
               Just _ -> fatal ("Interface " ++ show name ++ 
                                " defined more than once")
               Nothing -> 
                   let newIfaceMap = Map.insert name i (st_ifaces tab)
                       newMSigsMap = 
                           foldr (\ (k,rsig) map -> 
                                  foldr
                                  (\sig map ->
                                       let m = ms_name sig
                                       in 
                                         case Map.lookup m map of
                                           Just _ -> 
                                               fatal ("Name of interface method "
                                                      ++ show m ++ " not unique")
                                           Nothing ->
                                               Map.insert m (i,sig,k) map)
                                  map
                                  (rcsig_sigs rsig))
                           (st_msigs tab)
                           (zip [1..] (iface_receiver_sigs i))
                       newStaticMSigsMap = addAll "Static interface method "
                                                  (\sig -> (name, ms_name sig))
                                                  (iface_static_sigs i)
                                                  (st_static_msigs tab)
                   in tab { st_ifaces = newIfaceMap,
                            st_msigs = newMSigsMap,
                            st_static_msigs = newStaticMSigsMap
                          }
      add (ClassDef c) tab =
          let name = class_name c
          in case findClass name tab of
               Just _ -> fatal ("Class " ++ show name ++ 
                                " defined more than once")
               Nothing -> 
                   let newClassMap = Map.insert name c (st_classes tab)
                       newMDefMap = addAll "Class method "
                                           (\def -> (name, ms_name (md_sig def)))
                                           (class_methods c)
                                           (st_mdefs tab)
                       newFieldMap = addAll "Class field "
                                            (\fd -> (name, fd_name fd))
                                            (class_fields c)
                                            (st_fields tab)
                       newFieldNamesMap = 
                           Map.insert name 
                                      (Set.fromList
                                              (map fd_name (class_fields c)))
                                      (st_fieldNames tab)
                   in tab { st_classes = newClassMap,
                            st_mdefs = newMDefMap,
                            st_fields = newFieldMap,
                            st_fieldNames = newFieldNamesMap
                          }
      addAll :: (Show k, Ord k) => String -> (v -> k) -> [v] -> Map k v 
             -> Map k v
      addAll prefix f l m =
          foldr (\x m -> let key = f x
                         in case Map.lookup key m of
                              Just _ -> fatal (prefix ++ show key ++ 
                                               " defined more than once")
                              Nothing -> Map.insert key x m) m l

--
-- Lookup routines
--

findClass :: ClassName -> Symtab -> Maybe Class
findClass name tab = Map.lookup name (st_classes tab)

hasClass :: ClassName -> Symtab -> Bool
hasClass name tab = Map.lookup name (st_classes tab) /= Nothing

allClasses :: Symtab -> [Class]
allClasses tab = Map.elems (st_classes tab)

findMDef :: ClassName -> MethodName -> Symtab -> Maybe MethodDef
findMDef c m tab = Map.lookup (c,m) (st_mdefs tab)

findField :: ClassName -> FieldName -> Symtab -> Maybe FieldDef
findField c f tab = Map.lookup (c,f) (st_fields tab)

findIface :: IfaceName -> Symtab -> Maybe Interface
findIface name tab = Map.lookup name (st_ifaces tab)

allInterfaces :: Symtab -> [Interface]
allInterfaces tab = Map.elems (st_ifaces tab)

findMSig :: MethodName -> Symtab -> Maybe (Interface, MethodSig, RecvIx)
findMSig name tab = Map.lookup name (st_msigs tab)

findStaticMSig :: IfaceName -> MethodName -> Symtab -> Maybe MethodSig
findStaticMSig i m tab = Map.lookup (i,m) (st_static_msigs tab)

findAllImpls :: IfaceName -> Symtab -> [Implementation]
findAllImpls name tab = 
    case Map.lookup name (st_impls tab) of
      Just l -> l
      Nothing -> []

allImpls :: Symtab -> [Implementation]
allImpls tab = concat (Map.elems (st_impls tab))

findAllFieldNames :: ClassName -> Symtab -> Set FieldName
findAllFieldNames c tab = 
    case Map.lookup c (st_fieldNames tab) of
      Just s -> s
      Nothing -> Set.empty
    