{-# LANGUAGE MultiParamTypeClasses,TypeSynonymInstances #-}

-- common operations on types
module Types (

    Free(..), Subst(..), TySubst, tySubstAsList,
    dom, composeSubst, mkRenaming, mkSubst, emptySubst, freeIn, substMSig

) where

import Data.Map (Map)
import qualified Data.Map as Map
import Data.Set (Set)
import qualified Data.Set as Set
import Text.PrettyPrint
-- import Debug.Trace

import AST
import Logging
import Pretty

newtype TySubst = TySubst (Map TyvarName Type)

instance Show TySubst where
    showsPrec i (TySubst m) = showsPrec i m

instance Pretty TySubst where
    pp (TySubst m) =
        text "TySubst" <> braces (sep (punctuate comma 
          (map (\ (x,t) -> pp x <> text ": " <> pp t) (Map.toList m))))

dom :: TySubst -> Set TyvarName
dom (TySubst m) = Map.keysSet m

-- the type variables are in ascending order
tySubstAsList :: TySubst -> [(TyvarName,Type)]
tySubstAsList (TySubst m) = Map.assocs m

composeSubst :: TySubst -> TySubst -> TySubst
composeSubst (TySubst last) (TySubst first) = TySubst $
    apply (TySubst last) first `Map.union` (last `Map.difference` first)

mkRenaming :: [TyvarName] -> [TyvarName] -> TySubst
mkRenaming xs ys = mkSubst xs (map (\x -> TyG (TyVar x)) ys)

mkSubst :: [TyvarName] -> [Type] -> TySubst
mkSubst xs ts = 
{-    if length xs /= length ts 
       then bug ("Types.mkSubst: argument lists with different lengths: " ++
                 show xs ++ "; " ++ show ts)
       else -} 
           TySubst (Map.fromList (zip xs ts))

emptySubst :: TySubst
emptySubst = TySubst Map.empty

class Free t where
    free  :: t -> Set TyvarName

freeIn :: Free t => TyvarName -> t -> Bool
freeIn x t = x `Set.member` (free t)

class Subst t t' where
    apply :: TySubst -> t -> t'

-- 
-- general instances
--

instance Free t => Free [t] where 
    free l = Set.unions (map free l)

instance Subst t t' => Subst [t] [t'] where 
    apply s l = map (apply s) l

instance Free t => Free (Maybe t) where
    free Nothing = Set.empty
    free (Just t) = free t

instance Subst t t' => Subst (Maybe t) (Maybe t') where
    apply _ Nothing = Nothing
    apply s (Just t) = Just (apply s t)

instance Free t => Free (Map k t) where
    free m = free (map snd (Map.toList m))

instance Free TySubst where
    free (TySubst m) = free m

instance Subst t t' => Subst (Map k t) (Map k t') where
    apply s m = Map.map (apply s) m

instance Subst TySubst TySubst where
    apply s (TySubst m) = TySubst $ apply s m

instance (Free t, Free t') => Free (t, t') where
    free (a,b) = free a `Set.union` free b

instance (Subst t1 t1', Subst t2 t2') => Subst (t1, t2) (t1', t2') where
    apply s (a,b) = (apply s a, apply s b)

--
-- instances for types
--

instance Subst TyvarName Type where
    apply (TySubst m) x = case Map.lookup x m of
                            Just t -> t
                            Nothing -> TyG (TyVar x)

instance Free TyPrim where
    free _    = Set.empty

instance Subst TyPrim TyPrim where
    apply _ t = t

instance Free TyN where
    free c = free (tyc_args c)

instance Subst TyN TyN where
    apply s (TyClass name args) = TyClass name (apply s args)

instance Free TyK where
    free i = free (tyi_args i)

instance Subst TyK TyK where
    apply s (TyIface name args) = TyIface name (apply s args)

instance Free TyG where
    free (TyN n) = free n
    free (TyVar x) = Set.singleton x

instance Subst TyG Type where
    apply s (TyN n) = TyG (TyN (apply s n))
    apply s (TyVar x) = apply s x

instance Free Type where
    free (TyG g) = free g
    free (TyK k) = free k
    free (TyPrim p) = free p
    free t = bug ("found type " ++ show t)

instance Subst Type Type where
    apply s (TyG g) = apply s g
    apply s (TyK k) = TyK (apply s k)
    apply s (TyPrim p) = TyPrim (apply s p)
    apply _ t = bug ("found type " ++ show t)

--
-- instances for types
--

instance Free ConR where
    free (Impl gs k) = Set.union (free gs) (free k)

instance Subst ConR ConGR where
    apply s (Impl gs k) = ImplG (apply s gs) (apply s k)

instance Free ConGR where
    free (ImplG ts k) = Set.union (free ts) (free k)

instance Subst ConGR ConGR where
    apply s (ImplG ts k) = ImplG (apply s ts) (apply s k)

instance Free ConP where
    free (ConR r) = free r
    free (Ext x t) = Set.insert x (free t)
    free (Mono x) = Set.singleton x

instance Subst ConP ConGP where
    apply s (ConR r) = ConGR (apply s r)
    apply s (Ext x t) = ExtG (apply s x) (apply s t)
    apply s (Mono x) = MonoG (apply s x)

instance Free ConGP where
    free (ConGR gr) = free gr
    free (ExtG t u) = Set.union (free t) (free u)
    free (MonoG t) = free t

instance Subst ConGP ConGP where
    apply s (ConGR gr) = ConGR (apply s gr)
    apply s (ExtG t u) = ExtG (apply s t) (apply s u)
    apply s (MonoG t) = MonoG (apply s t)

--
-- instances for expressions
--

instance Free VarName where
    free _ = Set.empty

instance Subst VarName VarName where
    apply _ x = x

instance Free Expr where
    free e = 
        case e of
          FAccess e _ _ -> free e
          MCall e _ ts es _ -> Set.unions [free e, free ts, free es]
          SMCall k ts _ ts' es _ -> Set.unions [free k, free ts, free ts', 
                                                free es]
          New n es _ -> Set.union (free n) (free es)
          Cast n es -> Set.union (free n) (free es)
          PrimOp _ es -> free es
          _ -> Set.empty

instance Subst Expr Expr where
    apply s e = 
        case e of
          FAccess e f t -> FAccess (apply s e) f t
          MCall e m ts es t -> MCall (apply s e) m (apply s ts) (apply s es) t
          SMCall k ts m ts' es t -> 
              SMCall (apply s k) (apply s ts) m (apply s ts') (apply s es) t
          New n es r -> New (apply s n) (apply s es) r
          Cast n es -> Cast (apply s n) (apply s es)
          PrimOp op es -> PrimOp op (apply s es)
          _ -> e

--
-- instances for statements
--

instance Free Stmt where
    free e = 
        case e of
          StmtExpr e -> free e
          DefineLocal t _ e _ -> Set.union (free t) (free e)
          AssignLocal _ e _ -> free e
          AssignField e _ e' _ -> Set.union (free e) (free e')
          If e stmts -> free e `Set.union` free stmts
          IfElse e stmts stmts' -> 
              free e `Set.union` free stmts `Set.union` free stmts'
          While e stmts -> free e `Set.union` free stmts
          Return e _ -> free e
          Skip -> Set.empty

instance Subst Stmt Stmt where
    apply s stmt =
        case stmt of
          StmtExpr e -> StmtExpr (apply s e)
          DefineLocal t x e r -> DefineLocal (apply s t) x (apply s e) r
          AssignLocal x e r -> AssignLocal x (apply s e) r
          AssignField e f e' r -> AssignField (apply s e) f (apply s e') r
          If e stmts -> If (apply s e) (apply s stmts)
          IfElse e stmts stmts' -> 
              IfElse (apply s e) (apply s stmts) (apply s stmts')
          While e stmts -> While (apply s e) (apply s stmts)
          Return e r -> Return (apply s e) r
          Skip -> Skip
        
--
-- instances for methods
--

instance Free MethodArg where
    free (MethodArg p) = free p

instance Subst MethodArg MethodArg where
    apply s (MethodArg p) = MethodArg (apply s p)

instance Free a => Free (MethodSigGeneric a) where
    free msig = Set.unions [free (ms_return msig),
                            free (ms_args msig),
                            free (ms_cons msig)]
                `Set.difference` 
                (Set.fromList $ ms_tyargs msig)

substMSig :: (Free a, Subst a b) =>
             TySubst -> MethodSigGeneric a -> MethodSigGeneric b
substMSig subst msig = snd $ substMSig' subst msig Set.empty

substMSig' :: (Free a, Subst a b) =>
              TySubst -> MethodSigGeneric a -> Set TyvarName 
           -> (TySubst, MethodSigGeneric b)
substMSig' s msig toAvoid =
    let xs = ms_tyargs msig
        toAvoid' = Set.unions [toAvoid,
                               free s,
                               dom s,
                               free (ms_return msig),
                               free (ms_args msig),
                               free (ms_cons msig)]
        fresh = take (length xs) [(Set.findMax toAvoid' + 1)..]
        renameSubst = mkRenaming xs fresh
        newSubst = s `composeSubst` renameSubst
    in (newSubst, MethodSig { ms_tyargs = fresh,
                              ms_return = apply newSubst (ms_return msig),
                              ms_name = ms_name msig,
                              ms_args = apply newSubst (ms_args msig),
                              ms_cons = apply newSubst (ms_cons msig) })

instance (Free a, Subst a b) => 
    Subst (MethodSigGeneric a) (MethodSigGeneric b) where
    apply s msig = substMSig s msig

instance Free a => Free (MethodDefGeneric a) where
    free mdef = free (md_sig mdef) `Set.union` free (md_body mdef)

instance (Free a, Subst a b) => 
    Subst (MethodDefGeneric a) (MethodDefGeneric b) where
    apply s mdef =
        let (s', msig) = substMSig' s (md_sig mdef) (free (md_body mdef)) 
        in MethodDef { md_sig = msig,
                       md_this = md_this mdef,
                       md_body = apply s' (md_body mdef) }
