module WellFormedness where

import qualified Data.Set as Set
import Data.List

import AST
import Subtyping
import Environment
import Utils
import GIMonad
import Types
import Logging
import Pretty

typesOK :: TEnv -> [Type] -> GI Bool
typesOK tenv ts = allM (typeOK tenv) ts


typeOK :: TEnv -> Type -> GI Bool
typeOK tenv t = 
    do debug (showPP tenv ++ " |- " ++ showPP t ++ " OK: ?")
       res <- typeOK' tenv t
       debug (showPP tenv ++ " |- " ++ showPP t ++ " OK: " ++ show res)
       return res

typeOK' :: TEnv -> Type -> GI Bool
typeOK' tenv t =
    case t of
      TyBot -> return True
      TyPrim _ -> return True
      TyParse _ -> bug ("TyParse detected: " ++ show t)
      TyG (TyVar x) ->
          return (x `Set.member` definedTyvars tenv)
      TyG (TyN (TyClass "Object" [])) -> return True
      TyG (TyN (TyClass name ts)) -> 
          do cdef <- findClass name
             let ps = class_cons cdef
                 xs = class_tyargs cdef
                 subst = mkSubst xs ts
             andM [return (length xs == length ts),
                   typesOK tenv ts, 
                   entailsAll tenv (apply subst ps)]
      TyK k@(TyIface name ts) ->
          do idef <- findIface name
             y <- freshTyvar
             let cons = ConR $ Impl [TyVar y] k
                 xs = iface_tyargs idef
                 ys = iface_impl_types idef
                 rs = iface_impl_cons idef
                 ps = iface_cons idef
                 subst = mkSubst (ys ++ xs) (TyG (TyVar y) : ts)
             tenv' <- extendTEnv tenv [cons]
             andM [return (length xs == length ts),
                   return (length ys == 1),
                   typesOK tenv ts, 
                   entailsAll tenv' (apply subst (ps ++ (map ConR rs)))]
          
constraintsOK :: TEnv -> [ConGP] -> GI Bool
constraintsOK tenv ps = allM (constraintOK tenv) ps

constraintOK :: TEnv -> ConGP -> GI Bool
constraintOK tenv p =
    do debug (showPP tenv ++ " |- " ++ showPP p ++ " OK: ?")
       res <- constraintOK' tenv p
       debug (showPP tenv ++ " |- " ++ showPP p ++ " OK: " ++ show res)
       return res

constraintOK' :: TEnv -> ConGP -> GI Bool
constraintOK' tenv p =
    case p of
      MonoG t -> typeOK tenv t
      ExtG t u -> typesOK tenv [t,u]
      ConGR (ImplG ts (TyIface i us)) ->
          do idef <- findIface i
             let xs = iface_tyargs idef
                 ys = iface_impl_types idef
                 ps = iface_cons idef
                 rs = iface_impl_cons idef
                 subst = mkSubst (xs++ys) (us++ts)
             debug ("xs = " ++ show xs ++ ", us = " ++ show us ++ 
                    ", ys = " ++ show ys ++ ", ts = " ++ show ts)
             andM [return (length xs == length us),
                   return (length ys == length ts),
                   typesOK tenv ts,
                   typesOK tenv us,
                   entailsAll tenv (apply subst (map ConR rs ++ ps))]

checkMainOK :: String -> GI MethodDef
checkMainOK s = 
    do b <- hasClass "Main"
       if not b
          then failGI ("Class \"Main\" not found, cannot " ++ s ++ " program")
          else do cls <- findClass "Main"
                  case (class_constructor cls,
                        find (\md -> mdefName md == "main") 
                             (class_methods cls)) of
                    (Just _, _) ->
                        failGI ("Class \"Main\" has a constructor, cannot " ++
                                s ++ " program")
                    (_, Nothing) -> 
                        failGI ("Class \"Main\" does not provide a method " ++
                                "\"main\", cannot " ++ s ++ " program")
                    (Nothing, Just md) ->
                        case () of
                         _| not (null (mdefArgs md)) ->
                              failGI ("Method \"main\" of class \"Main\" " ++
                                      "takes " ++
                                      "arguments, cannot " ++ s ++ " program")
                          | not (null (mdefTyargs md)) ->
                              failGI ("Method \"main\" of class \"Main\" " ++
                                      "takes type arguments, cannot " ++ s ++ 
                                      " program")
                          | mdefReturn md /= Nothing ->
                              failGI ("Method \"main\" of class \"Main\" " ++
                                      "does not return void, " ++
                                      "cannot " ++ s ++ " program")
                          | otherwise -> return md