package javagi.runtime;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class CoreRT {

    private static final Map<Class<?>, List<Implementation>> map = 
        new HashMap<Class<?>, List<Implementation>>();
    
    /*
     * The dispatch vector has the following format:
     * dispatchVector[2*i] is the implementing type to which
     * the argument at index dispatchVector[2*i + 1] contributes.
     */
    public static final Object getMethods(Class<?> iface,
                                          int[] dispatchVector,
                                          Object[] args) {
        List<Implementation> impls = map.get(iface);
        if (impls == null) {
            throw new CoreJavaGIError("No implementations available for interface " +
                                      iface.getName() + ". Internal state:\n" + internalStateAsString());
        }
        for (Implementation impl : impls) {
            if (matches(impl, dispatchVector, args)) {
                return impl.methods;
            }
        }
        throw new CoreJavaGIError("No matching implementation found for interface " +
                                  iface.getName() + ", dispatch vector " + 
                                  Arrays.toString(dispatchVector) + ", and argument classes " +
                                  argumentClassesAsString(args) + ". Internal state:\n" + internalStateAsString());
    }

    private static String internalStateAsString() {
        StringBuffer sb = new StringBuffer();
        for (Map.Entry<Class<?>, List<Implementation>> entry : map.entrySet()) {
            sb.append("  ");
            sb.append(entry.getKey().getName());
            if (entry.getValue().isEmpty()) {
                sb.append(" -> []\n");
            } else {
                sb.append(" ->\n");
                for (Implementation impl : entry.getValue()) {
                    sb.append("    * ");
                    sb.append(impl);
                    sb.append("\n");
                }
            }
        }
        return sb.toString();
    }

    private static String argumentClassesAsString(Object[] args) {
        StringBuffer sb = new StringBuffer();
        sb.append("[");
        for (int i = 0; i < args.length; i++) {
            if (args[i] == null) {
                sb.append("null");
            } else {
                sb.append(args[i].getClass().getName());
            }
            if (i < args.length - 1) {
                sb.append(",");
            }
        }
        sb.append("]");
        return sb.toString();
    }

    static boolean matches(Implementation impl, int[] dispatchVector, Object[] args) {
        for (int i = 1; i < dispatchVector.length; i = i+2) {
            int implType = dispatchVector[i-1];
            int argIndex = dispatchVector[i];
            if (args[argIndex] == null) {
                String msg;
                if (argIndex == 0) {
                    msg = "receiver of method call is null";
                } else {
                    msg = "dispatch argument " + argIndex + " of method call is null";
                }
                throw new NullPointerException(msg);
            }
            if (! Utils.isSubtype(args[argIndex].getClass(),
                                  impl.implementingTypes[implType])) {
                return false;
            }
        }
        return true;
    }

    public static final void init(Implementation[] implementations) {
        // build the implementation map
        for (Implementation impl : implementations) {
            List<Implementation> l = map.get(impl.iface);
            if (l == null) {
                l = new ArrayList<Implementation>();
                map.put(impl.iface, l);
            }
            l.add(impl);
        }
        // now sort the implementations
        for (Map.Entry<Class<?>, List<Implementation>> entry : map.entrySet()) {
            Collections.sort(entry.getValue());           
        }
    }
    
    /* 
     * Primitive values are not passed in the "args" argument of
     * "getImplementation". Instead, descriptors obtained via the
     * following methods are used.
     */
    private static class PrimitiveDescriptor {}
    public static final PrimitiveDescriptor primDescriptor = new PrimitiveDescriptor();
    public static Object intDescriptor(int i) {
        return primDescriptor;
    }
    public static Object boolDescriptor(boolean b) {
        return primDescriptor;
    }
}