package javagi.runtime;

import java.util.Arrays;

/*
 * 
 * A note an the implementation of the Comparable interface.
 * 
 * Only implementation definitions for identical interfaces and with an 
 * implementing types array of equal length can be compared. 
 * 
 * Given these preconditions hold, an implementation I is less than 
 * (greater than) an implementation J iff each implementing type of I is a 
 * subtype (supertype) of the corresponding implementing type of J. 
 * Otherwise, I and J are considered to be equal.
 * This implies that the natural ordering of this class is inconsistent 
 * with equals!
 * 
 * The ordering for implementations just describe allows us to sort a sequence
 * of implementations in such a way that a more specific implementation
 * of some interface is placed in front of a more general implementation. 
 */
public final class Implementation implements Comparable<Implementation> {

    public final Class<?> iface;
    public final Class<?>[] implementingTypes;
    public final Object methods;
    
    public Implementation(Class<?> iface, 
                          Class<?>[] implementingTypes,
                          Object methods) {
        this.iface = iface;
        this.implementingTypes = implementingTypes;
        this.methods = methods;
    }

    public int compareTo(Implementation that) {
        if (this == that) return 0;
        Class<?>[] thisImplTypes = this.implementingTypes;
        Class<?>[] thatImplTypes = that.implementingTypes;
        if (this.iface != that.iface ||
            thisImplTypes.length != thatImplTypes.length) {
            throw new IllegalArgumentException("Cannot compare " + this + " with " + that);
        }
        int res = 0;
        for (int i = 0; i < thisImplTypes.length; i++) {
            Class<?> thisClass = thisImplTypes[i];
            Class<?> thatClass = thatImplTypes[i];
            boolean thisThat = Utils.isSubtype(thisClass, thatClass);
            boolean thatThis = Utils.isSubtype(thatClass, thisClass);
            if (thisThat && thatThis) {
                // do nothing
            } else if (thisThat && (res == 0 || res == -1)) {
                res = -1;
            } else if (thatThis && (res == 0 || res == 1)) {
                res = 1;
            } else {
                return 0;
            }
        }
        return res;
    }
    
    public String toString() {
        return "Implementation{iface=" + this.iface + ", implementingTypes=" + 
               Arrays.toString(this.implementingTypes) + ", methods=" + methods;
    }
}
