#!/usr/bin/env python2.5

import glob
import os
import sys
import subprocess
import re

WIDTH = 50
LOG_FILE = 'tests.log'
TEST_DIRECTORIES = ['.', '../examples', 'translation']
IGNORE_FOR_JAVAGIC = ['parseable_ok1.gi', '../examples/parseable.gi']

FAIL_PAT = '.*fail[0-9]*\\.gi$'
OK_PAT = '.*\\.gi$'

log_file = open(LOG_FILE, 'w')

def log(s):
    log_file.write(s)
    log_file.write('\n')
    
def is_test(f, fail=False):
    if fail:
        return re.match(FAIL_PAT, f)
    else:
        return re.match(OK_PAT, f) and not re.match(FAIL_PAT, f)

def is_ignored_for_javagic(t):
    for x in IGNORE_FOR_JAVAGIC:
        if os.path.normpath(x) == os.path.normpath(t):
            return True
    return False

def get_tests():
    if len(sys.argv) > 1:
        l = sys.argv[1:]
    else:
        l = []
        for d in TEST_DIRECTORIES:
            l = l + (glob.glob(d + '/*.gi'))
    failures = []
    successes = []
    for x in l:
        if is_test(x, True): failures.append(x)
        elif is_test(x, False): successes.append(x)
        else:
            print 'file %s does not denote a test' % x
            sys.exit(1)
    return (successes, failures)

def run_javagi(f):
    p = subprocess.Popen(["../dist/build/core-javagi/core-javagi",
                          "--quiet", f],
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (out,err) = p.communicate()
    i = p.wait()
    return (i, out, err)

def run_javagic(f):
    d = os.path.dirname(f)
    if not d: d = '.'
    os.system('rm -f %s/*.class' % d)
    p = subprocess.Popen(["../scripts/core-javagic.sh",
                          "--quiet", f],
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (out,err) = p.communicate()
    i = p.wait()
    if i != 0:
        return (i, out, err)
    (b,_) = os.path.splitext(os.path.basename(f))
    p = subprocess.Popen(["java",
                          "-classpath", d + ':' + '../java/build', b],
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (out,err) = p.communicate()
    i = p.wait()
    return (i, out, err)

def get_initial_comment(f):
    lines = open(f).readlines()
    lines = map(lambda s: s.strip(), lines)
    if not lines or not lines[0].strip().startswith("/*"):
        return None
    out = []
    for l in lines[1:]:
        i = l.find("*/")
        if i == -1:
            out.append(l)
        else:
            out.append(l[:i])
            break
    return '\n'.join(out).strip()

def log_test(t, ecode, out, err):
    log('---\nTest %s\nexit code: %d\nstdout:\n%s\nstderr:\n%s' \
        % (t, ecode, out, err))

def start_test(t):
    sys.stdout.write(t)
    sys.stdout.flush()

def end_test(t, ok, comment=None, details=None):
    n = abs(WIDTH - len(t))
    if comment is not None:
        comment = ' (%s)' % comment
    else:
        comment = ''
    ws = n*' '
    if ok:
        print '%sok%s' % (ws,comment)
    else:
        print '%sFAIL%s' % (ws,comment)
    if details is not None:
        log(details)

def run_failing_tests(tests):
    res = []
    for t in tests:
        start_test(t)
        (i,out,err) = run_javagi(t)
        log_test(t, i, out, err)
        if i != 0:
            pat = get_initial_comment(t)
            if pat is None:
                pat = ''
            if err.find(pat) >= 0:
                end_test(t,True)
            else:
                end_test(t,False, 'wrong output',
                         details='error message pattern: %s' % pat)
                res.append(t)
        else:
            end_test(t,False, details='test should fail but succeeded')
            res.append(t)
    return res

def run_success_tests(tests):
    res = []
    for t in tests:
        start_test(t)
        (i,out,err) = run_javagi(t)
        log_test(t, i, out, err)
        if i == 0:
            exp = get_initial_comment(t) or ''
            if exp == out.strip():
                if is_ignored_for_javagic(t):
                    end_test(t,True)
                else:
                    (i,out,err) = run_javagic(t)
                    log('run of javagic on ' +
                        '%s\nexit code: %d\nstdout:\n%s\nstderr:\n%s' \
                        % (t, i, out, err))
                    if i == 0:
                        if exp == out.strip():
                            end_test(t,True)
                        else:
                            end_test(t,False, 'wrong output',
                                     details='expected output: %s' % exp)
                            res.append(t)
                    else:
                        end_test(t,False)
                        res.append(t)            
            else:
                end_test(t,False, 'wrong output',
                         details='expected output: %s' % exp)
                res.append(t)
        else:
            end_test(t,False)
            res.append(t)
    return res

def main():
    (ss,fs) = get_tests()
    f1 = run_failing_tests(fs)
    f2 = run_success_tests(ss)
    failed = f1 + f2
    print '---'
    print "%d tests, %d failure(s)" % (len(ss + fs), len(failed))
    if len(failed) != 0:
        print "Failed tests:"
        print ' * ' + '\n * '.join(failed)
        print '(For details see %s)' % LOG_FILE

if __name__ == '__main__':
    main()
